/*
 * Name:    d_cmpmtr.c
 *
 * Purpose: MX driver for the Compumotor 6000 series of motor controllers.
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "d_cmpmtr.h"

/* ============ Motor channels ============ */

MX_RECORD_FUNCTION_LIST mxd_compumotor_record_function_list = {
	mxd_compumotor_initialize_type,
	mxd_compumotor_create_record_structures,
	mxd_compumotor_finish_record_initialization,
	mxd_compumotor_delete_record,
	mxd_compumotor_print_structure,
	mxd_compumotor_read_parms_from_hardware,
	mxd_compumotor_write_parms_to_hardware,
	mxd_compumotor_open,
	mxd_compumotor_close,
	NULL,
	mxd_compumotor_resynchronize
};

MX_MOTOR_FUNCTION_LIST mxd_compumotor_motor_function_list = {
	mxd_compumotor_motor_is_busy,
	mxd_compumotor_move_absolute,
	mxd_compumotor_get_position,
	mxd_compumotor_set_position,
	mxd_compumotor_soft_abort,
	mxd_compumotor_immediate_abort,
	mxd_compumotor_positive_limit_hit,
	mxd_compumotor_negative_limit_hit,
	mxd_compumotor_find_home_position,
	mxd_compumotor_constant_velocity_move,
	mxd_compumotor_get_parameter,
	mxd_compumotor_set_parameter
};

/* Compumotor motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_compumotor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_STEPPER_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_COMPUMOTOR_STANDARD_FIELDS
};

long mxd_compumotor_num_record_fields
		= sizeof( mxd_compumotor_record_field_defaults )
			/ sizeof( mxd_compumotor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_compumotor_rfield_def_ptr
			= &mxd_compumotor_record_field_defaults[0];

#define COMPUMOTOR_DEBUG	FALSE

/* A private function for the use of the driver. */

static mx_status_type
mxd_compumotor_get_pointers( MX_MOTOR *motor,
			MX_COMPUMOTOR **compumotor,
			MX_COMPUMOTOR_INTERFACE **compumotor_interface,
			const char *calling_fname )
{
	const char fname[] = "mxd_compumotor_get_pointers()";

	MX_RECORD *compumotor_interface_record;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( compumotor == (MX_COMPUMOTOR **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_COMPUMOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_COMPUMOTOR_INTERFACE pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*compumotor = (MX_COMPUMOTOR *) (motor->record->record_type_struct);

	if ( *compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_COMPUMOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface_record =
			(*compumotor)->compumotor_interface_record;

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The compumotor_interface_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	*compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			compumotor_interface_record->record_type_struct;

	if ( *compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_COMPUMOTOR_INTERFACE pointer for record '%s' is NULL.",
			motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_create_record_structures()";

	MX_MOTOR *motor;
	MX_COMPUMOTOR *compumotor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	compumotor = (MX_COMPUMOTOR *) malloc( sizeof(MX_COMPUMOTOR) );

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_COMPUMOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = compumotor;
	record->class_specific_function_list
				= &mxd_compumotor_motor_function_list;

	motor->record = record;

	/* A Compumotor motor is treated as an stepper motor. */

	motor->subclass = MXC_MTR_STEPPER;

	/* The Compumotor reports accelerations in steps/sec**2 */

	motor->acceleration_type = MXF_MTR_ACCEL_DIST_PER_SECOND_SQUARED;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_finish_record_initialization()";

	MX_MOTOR *motor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	MX_RECORD **motor_array;
	int i, j, num_axes;
	mx_status_type status;

	motor = (MX_MOTOR *) record->record_class_struct;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( compumotor_interface->motor_array == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The motor_array pointer for the Compumotor interface '%s' is NULL.",
			compumotor_interface->record->name );
	}

	i = compumotor->controller_number - 1;
	j = compumotor->axis_number - 1;

	num_axes = compumotor_interface->num_axes[i];

	motor_array = compumotor_interface->motor_array[i];

	if ( motor_array == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The motor_array pointer for index %d on the "
			"Compumotor interface '%s' is NULL.",
			i, compumotor_interface->record->name );
	}

	motor_array[j] = record;

	status = mx_motor_finish_record_initialization( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_print_structure()";

	MX_MOTOR *motor;
	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);

	fprintf(file, "  Motor type        = COMPUMOTOR motor.\n\n");

	fprintf(file, "  name              = %s\n", record->name);
	fprintf(file, "  port name         = %s\n",
					compumotor_interface->record->name);
	fprintf(file, "  controller number = %d\n",
					compumotor->controller_number);
	fprintf(file, "  axis number       = %d\n", compumotor->axis_number);
	fprintf(file, "  use encoder       = %d\n", compumotor->use_encoder);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
		}
	
	fprintf(file, "  position          = %g %s  (%ld steps)\n",
			motor->position, motor->units,
			motor->raw_position.stepper );
	fprintf(file, "  scale             = %g %s per step.\n",
			motor->scale, motor->units);
	fprintf(file, "  offset            = %g %s.\n",
			motor->offset, motor->units);
	
	fprintf(file, "  backlash          = %g %s  (%ld steps)\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.stepper );
	
	fprintf(file, "  negative limit    = %g %s  (%ld steps)\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.stepper );

	fprintf(file, "  positive limit    = %g %s  (%ld steps)\n",
		motor->positive_limit, motor->units,
		motor->raw_positive_limit.stepper );

	deadband = motor->scale * (double) motor->deadband.stepper;

	fprintf(file, "  deadband          = %g %s  (%ld steps)\n\n",
		deadband, motor->units,
		motor->deadband.stepper );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_read_parms_from_hardware( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_read_parms_from_hardware()";

	MX_MOTOR *motor;
	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	double position;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the current position.  This has the side effect of
	 * updating the position value in the MX_MOTOR structure.
	 */

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* At present, the position is the only parameter we read out.
	 * We rely on the non-volatile memory of the controller
	 * for the rest of the controller parameters.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_open( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_open()";

	MX_MOTOR *motor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	char command[80];
	char response[80];
	int length;
	mx_status_type status;

	MX_DEBUG(2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Verify that the controller is present by asking for 
	 * its revision number.
	 */

	sprintf( command, "%d_!TREV", compumotor->controller_number );

	status = mxi_compumotor_command( compumotor_interface, command,
			response, sizeof response, COMPUMOTOR_DEBUG );

	switch( status.code ) {
	case MXE_SUCCESS:
		break;
	case MXE_INTERFACE_IO_ERROR:
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Cannot communicate with Compumotor controller %d on RS-232 port '%s'.  "
"Is it turned on?",
			compumotor->controller_number,
			compumotor_interface->rs232_record->name );
		break;
	default:
		return status;
		break;
	}

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to controller revision level command.");
	}

	MX_DEBUG(2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_resynchronize()";

	MX_COMPUMOTOR *compumotor;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed was NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) record->record_type_struct;

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	status = mxi_compumotor_resynchronize(
			compumotor->compumotor_interface_record );

	return status;
}

/* ============ Motor specific functions ============ */

MX_EXPORT mx_status_type
mxd_compumotor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_motor_is_busy()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	char response[80];
	int length;
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( command, "%d_!%dTAS", compumotor->controller_number,
					compumotor->axis_number );

	status = mxi_compumotor_command( compumotor_interface, command,
			response, sizeof response, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to motor status command.");
	}

	if ( response[0] == '1' ) {
		motor->busy = TRUE;
	} else {
		motor->busy = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_move_absolute()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	long motor_steps;
	int i, length, command_buffer_left;
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Change the Compumotor axis to preset mode. */

	status = mxd_compumotor_enable_continuous_mode( compumotor,
						compumotor_interface, FALSE );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Set the position to move to. */

	motor_steps = motor->raw_setpoint.stepper;

	sprintf( command, "%d_!%dD%ld", compumotor->controller_number,
			compumotor->axis_number, motor_steps );

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Construct the move command. */

	sprintf( command, "%d_!GO", compumotor->controller_number );

	for ( i = 0; i < MX_MAX_COMPUMOTOR_AXES; i++ ) {

		length = strlen( command );
		command_buffer_left = sizeof( command ) - length - 1;

		if ( (i+1) == compumotor->axis_number ) {
			strncat( command, "1", command_buffer_left );
		} else {
			strncat( command, "X", command_buffer_left );
		}
	}

	command[ length + MX_MAX_COMPUMOTOR_AXES ] = '\0';

	/* Command the move to start. */

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_get_position()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	char response[80];
	long motor_steps;
	int num_tokens;
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( compumotor->use_encoder ) {
		sprintf( command, "%d_!%dTPE",
			compumotor->controller_number,
			compumotor->axis_number );
	} else {
		sprintf( command, "%d_!%dTPM",
			compumotor->controller_number,
			compumotor->axis_number );
	}

	status = mxi_compumotor_command( compumotor_interface, command,
			response, sizeof response, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_tokens = sscanf( response, "%ld", &motor_steps );

	if ( num_tokens != 1 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"No position value seen in response to '%s' command.  "
		"Response seen = '%s'", command, response );
	}

	motor->raw_position.stepper = motor_steps;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_set_position()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_RECORD **motor_array;
	long motor_steps;
	long other_motor_steps;
	char command[100];
	char buffer[80];
	int i, j, num_axes, length, command_buffer_left;
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( compumotor_interface->motor_array == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The motor_array pointer for the Compumotor interface '%s' is NULL.",
			compumotor_interface->record->name );
	}

	i = compumotor->controller_number - 1;

	num_axes = compumotor_interface->num_axes[i];

	motor_array = compumotor_interface->motor_array[i];

	motor_steps = motor->raw_setpoint.stepper;

	/* Construct a PSET command. */

	sprintf( command, "%d_!PSET", compumotor->controller_number );

	for ( j = 0; j < num_axes; j++ ) {

		if ( j != 0 ) {
			length = strlen( command );
			command_buffer_left = sizeof(buffer) - length - 1;
			strncat( command, ",", command_buffer_left );
		}
		if ( j+1 == compumotor->axis_number ) {
			sprintf( buffer, "%ld", motor_steps );
		} else {
			if ( motor_array[j] == (MX_RECORD *) NULL ) {
				strcpy( buffer, "0" );
			} else {
				status = mx_motor_get_position_steps(
					motor_array[j], &other_motor_steps );

				if ( status.code != MXE_SUCCESS )
					return status;

				sprintf( buffer, "%ld", other_motor_steps );
			}
		}
		length = strlen( command );
		command_buffer_left = sizeof( command ) - length - 1;

		strncat( command, buffer, command_buffer_left );
	}

	/* Send the PSET command. */

	status = mxi_compumotor_command( compumotor_interface, command,
			NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_soft_abort()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	int i, length;
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Construct a stop command that only stops this axis. */

	sprintf( command, "%d_!S", compumotor->controller_number );

	length = strlen( command );

	for ( i = 0; i < MX_MAX_COMPUMOTOR_AXES; i++ ) {

		if ( (i+1) == compumotor->axis_number ) {
			command[ length + i ] = '1';
		} else {
			command[ length + i ] = '0';
		}
	}

	command[ length + MX_MAX_COMPUMOTOR_AXES ] = '\0';

	/* Send the stop command. */

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_immediate_abort()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Send a kill motion command that halts all axes attached to
	 * this controller.
	 */

	sprintf( command, "%d_!K", compumotor->controller_number );

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_positive_limit_hit()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	char response[80];
	int length;
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( command, "%d_!%dTAS", compumotor->controller_number,
					compumotor->axis_number );

	status = mxi_compumotor_command( compumotor_interface, command,
			response, sizeof response, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to motor status command.");
	}

	/* When constructing bit indices for the output of the TAS
	 * command, remember that the output of TAS also has
	 * embedded underscores in it that you must skip over too.
	 */

	if ( ( response[17] == '1' ) || ( response[20] == '1' ) ) {
		motor->positive_limit_hit = TRUE;
	} else {
		motor->positive_limit_hit = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_negative_limit_hit()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	char response[80];
	int length;
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( command, "%d_!%dTAS", compumotor->controller_number,
					compumotor->axis_number );


	status = mxi_compumotor_command( compumotor_interface, command,
			response, sizeof response, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to motor status command.");
	}

	/* When constructing bit indices for the output of the TAS
	 * command, remember that the output of TAS also has
	 * embedded underscores in it that you must skip over too.
	 */

	if ( ( response[18] == '1' ) || ( response[21] == '1' ) ) {
		motor->negative_limit_hit = TRUE;
	} else {
		motor->negative_limit_hit = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_find_home_position()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	int i, length;
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Construct the home search command. */

	sprintf( command, "%d_!HOM", compumotor->controller_number );

	length = strlen( command );

	for ( i = 0; i < MX_MAX_COMPUMOTOR_AXES; i++ ) {

		if ( (i+1) == compumotor->axis_number ) {
			switch( motor->home_search ) {
			case 1:
				command[ length + i ] = '0';
				break;
			case (-1):
				command[ length + i ] = '1';
				break;
			default:
				command[ length + i ] = 'X';
			}
		} else {
			command[ length + i ] = 'X';
		}
	}

	command[ length + MX_MAX_COMPUMOTOR_AXES ] = '\0';

	/* Command the home search to start. */

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_constant_velocity_move( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_constant_velocity_move()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	int i, length, command_buffer_left;
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Change the Compumotor axis to continuous mode. */

	status = mxd_compumotor_enable_continuous_mode( compumotor,
						compumotor_interface, TRUE );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Set the direction of the move. */

	if ( motor->constant_velocity_move >= 0 ) {
		sprintf( command, "%d_!%dD+", compumotor->controller_number,
						compumotor->axis_number );
	} else {
		sprintf( command, "%d_!%dD-", compumotor->controller_number,
						compumotor->axis_number );
	}
		
	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Construct the move command. */

	sprintf( command, "%d_!GO", compumotor->controller_number );

	for ( i = 0; i < MX_MAX_COMPUMOTOR_AXES; i++ ) {

		length = strlen( command );
		command_buffer_left = sizeof( command ) - length - 1;

		if ( (i+1) == compumotor->axis_number ) {
			strncat( command, "1", command_buffer_left );
		} else {
			strncat( command, "X", command_buffer_left );
		}
	}

	command[ length + MX_MAX_COMPUMOTOR_AXES ] = '\0';

	/* Command the move to start. */

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_get_parameter()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	char response[80];
	int num_items;
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->parameter_type == MXLV_MTR_SPEED ) {

		sprintf( command, "%d_!%dV", compumotor->controller_number,
						compumotor->axis_number );

		status = mxi_compumotor_command( compumotor_interface, command,
				response, sizeof(response), COMPUMOTOR_DEBUG );

		if ( status.code != MXE_SUCCESS )
			return status;

		num_items = sscanf( response, "V%lg", &(motor->raw_speed) );

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Unable to parse returned speed for motor '%s'.  Response = '%s'",
				motor->record->name, response );
		}

	} else if ( motor->parameter_type == MXLV_MTR_BASE_SPEED ) {

		sprintf( command, "%d_!%dSSV", compumotor->controller_number,
						compumotor->axis_number );

		status = mxi_compumotor_command( compumotor_interface, command,
				response, sizeof(response), COMPUMOTOR_DEBUG );

		if ( status.code != MXE_SUCCESS )
			return status;

		num_items = sscanf( response, "SSV%lg",
						&(motor->raw_base_speed) );

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Unable to parse returned base speed for motor '%s'.  Response = '%s'",
				motor->record->name, response );
		}

	} else if ( motor->parameter_type == MXLV_MTR_MAXIMUM_SPEED ) {

		motor->raw_maximum_speed = 1600000;  /* steps per second */

	} else if ( motor->parameter_type == MXLV_MTR_ACCELERATION_PARAMETERS){

		sprintf( command, "%d_!%dA", compumotor->controller_number,
						compumotor->axis_number );

		status = mxi_compumotor_command( compumotor_interface, command,
				response, sizeof(response), COMPUMOTOR_DEBUG );

		if ( status.code != MXE_SUCCESS )
			return status;

		num_items = sscanf( response, "A%lg",
					&(motor->acceleration_parameters[0]) );

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Unable to parse returned speed for motor '%s'.  Response = '%s'",
				motor->record->name, response );
		}

		motor->acceleration_parameters[1] = 0.0;
		motor->acceleration_parameters[2] = 0.0;
		motor->acceleration_parameters[3] = 0.0;

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			motor->parameter_type );
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_set_parameter()";

	MX_COMPUMOTOR *compumotor;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	char command[80];
	mx_status_type status;

	status = mxd_compumotor_get_pointers( motor, &compumotor,
						&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Format the appropriate parameter setting command. */

	if ( motor->parameter_type == MXLV_MTR_SPEED ) {

		sprintf( command, "%d_!%dV%g", compumotor->controller_number,
						compumotor->axis_number,
						motor->raw_speed );

	} else if ( motor->parameter_type == MXLV_MTR_BASE_SPEED ) {

		sprintf( command, "%d_!%dSSV%g", compumotor->controller_number,
						compumotor->axis_number,
						motor->raw_base_speed );

	} else if ( motor->parameter_type == MXLV_MTR_MAXIMUM_SPEED ) {

		motor->raw_maximum_speed = 1600000;  /* steps per second */

		return MX_SUCCESSFUL_RESULT;

	} else if ( motor->parameter_type == MXLV_MTR_ACCELERATION_PARAMETERS){

		/* First make sure that deceleration equals acceleration. */

		sprintf( command, "%d_!%dAD0", compumotor->controller_number,
					compumotor->axis_number );
		
		status = mxi_compumotor_command( compumotor_interface,
				command, NULL, 0, COMPUMOTOR_DEBUG );

		if ( status.code != MXE_SUCCESS )
			return status;

		/* Then, format the set acceleration command. */

		sprintf( command, "%d_!%dA%g", compumotor->controller_number,
					compumotor->axis_number,
					motor->acceleration_parameters[0] );

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			motor->parameter_type );
	}

	/* Send the command to the Compumotor controller. */

	status = mxi_compumotor_command( compumotor_interface,
				command, NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

/*======================================================================*/

MX_EXPORT mx_status_type
mxd_compumotor_enable_continuous_mode( MX_COMPUMOTOR *compumotor,
				MX_COMPUMOTOR_INTERFACE *compumotor_interface,
				int enable_flag )
{
	const char fname[] = "mxd_compumotor_enable_continuous_mode()";

	MX_RECORD **motor_array;
	char command[80];
	int i, j, num_axes, length, command_buffer_left;
	mx_status_type status;

	/* Are we being commanded to change to the positioner mode
	 * that we are already in?  If so, return without sending any
	 * commands out.
	 */

	if ( compumotor->continuous_mode_enabled == enable_flag )
		return MX_SUCCESSFUL_RESULT;

	/* Construct the MC command that we will use to change the mode. */

	if ( compumotor_interface->motor_array == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The motor_array pointer for the Compumotor interface '%s' is NULL.",
			compumotor_interface->record->name );
	}

	i = compumotor->controller_number - 1;

	num_axes = compumotor_interface->num_axes[i];

	motor_array = compumotor_interface->motor_array[i];

	sprintf( command, "%d_!MC", compumotor->controller_number );

	for ( j = 0; j < num_axes; j++ ) {

		length = strlen( command );
		command_buffer_left = sizeof( command ) - length - 1;

		if ( j+1 == compumotor->axis_number ) {
			strncat( command, "1", command_buffer_left );
		} else {
			strncat( command, "X", command_buffer_left );
		}
	}

	/* Send the MC command. */

	status = mxi_compumotor_command( compumotor_interface, command,
			NULL, 0, COMPUMOTOR_DEBUG );

	if ( status.code == MXE_SUCCESS ) {
		compumotor->continuous_mode_enabled = enable_flag;
	}
	return status;
}

