/*
 * Name:    d_eptime.c
 *
 * Purpose: MX timer driver for the EPICS APS-XFD developed scaler record.
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_meas.h"
#include "mx_timer.h"
#include "d_eptime.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the timer driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_epics_timer_record_function_list = {
	mxd_epics_timer_initialize_type,
	mxd_epics_timer_create_record_structures,
	mxd_epics_timer_finish_record_initialization,
	mxd_epics_timer_delete_record,
	NULL,
	mxd_epics_timer_read_parms_from_hardware,
	mxd_epics_timer_write_parms_to_hardware,
	mxd_epics_timer_open,
	mxd_epics_timer_close
};

MX_TIMER_FUNCTION_LIST mxd_epics_timer_timer_function_list = {
	mxd_epics_timer_is_busy,
	mxd_epics_timer_start,
	mxd_epics_timer_stop,
	mxd_epics_timer_clear,
	mxd_epics_timer_read,
	mxd_epics_timer_get_mode,
	mxd_epics_timer_set_mode,
	mxd_epics_timer_set_modes_of_associated_counters
};

/* EPICS timer data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_epics_timer_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_TIMER_STANDARD_FIELDS,
	MXD_EPICS_TIMER_STANDARD_FIELDS
};

long mxd_epics_timer_num_record_fields
		= sizeof( mxd_epics_timer_record_field_defaults )
		  / sizeof( mxd_epics_timer_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_epics_timer_rfield_def_ptr
			= &mxd_epics_timer_record_field_defaults[0];

MX_EXPORT mx_status_type
mxd_epics_timer_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_timer_create_record_structures()";

	MX_TIMER *timer;
	MX_EPICS_TIMER *epics_timer;

	/* Allocate memory for the necessary structures. */

	timer = (MX_TIMER *) malloc( sizeof(MX_TIMER) );

	if ( timer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_TIMER structure." );
	}

	epics_timer = (MX_EPICS_TIMER *)
				malloc( sizeof(MX_EPICS_TIMER) );

	if ( epics_timer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_TIMER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = timer;
	record->record_type_struct = epics_timer;
	record->class_specific_function_list
			= &mxd_epics_timer_timer_function_list;

	timer->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_finish_record_initialization( MX_RECORD *record )
{
	return mx_timer_finish_record_initialization( record );
}

MX_EXPORT mx_status_type
mxd_epics_timer_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_read_parms_from_hardware( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxd_epics_timer_read_parms_from_hardware() called."));

	/* No module state to save, so there is nothing to read. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_write_parms_to_hardware( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxd_epics_timer_write_parms_to_hardware() called."));

	/* No module state to restore. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_open( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_timer_open()";

	MX_TIMER *timer;
	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	timer = (MX_TIMER *) (record->record_class_struct);

	if ( timer == (MX_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_TIMER pointer for record '%s' is NULL.",
			record->name);
	}

	epics_timer = (MX_EPICS_TIMER *) (record->record_type_struct);

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_EPICS_TIMER pointer for timer record '%s' is NULL.",
			record->name );
	}

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	/* Get the clock frequency for the timer.  This has the side effect
	 * of verifying that the timer is present.
	 */

	sprintf( pvname, "%s.FREQ", epics_timer->epics_record_name );

	ezca_status = ezcaGet( pvname, ezcaDouble, 1,
					&(epics_timer->clock_frequency) );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to get the clock frequency for EPICS timer '%s' failed.  "
"EZCA status = %d (%s)",
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		epics_timer->clock_frequency = 1.0;

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_is_busy( MX_TIMER *timer )
{
	const char fname[] = "mxd_epics_timer_is_busy()";

	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	long count_field;
	int ezca_status;

	epics_timer = (MX_EPICS_TIMER *) (timer->record->record_type_struct);

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_TIMER structure pointer is NULL.");
	}

	sprintf( pvname, "%s.CNT", epics_timer->epics_record_name );

	ezca_status = ezcaGet( pvname, ezcaLong, 1, &count_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to read busy status of EPICS timer '%s' (%s) failed. "
"EZCA status = %d (%s)",
			timer->record->name,
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	switch ( count_field ) {
	case 0:
		timer->busy = FALSE;
		break;
	case 1:
		timer->busy = TRUE;
		break;
	default:
		timer->busy = FALSE;

		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Count field (.CNT) for scaler '%s' had an unexpected value of %d",
			epics_timer->epics_record_name, count_field );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_start( MX_TIMER *timer )
{
	const char fname[] = "mxd_epics_timer_start()";

	MX_EPICS_TIMER *epics_timer;
	double seconds;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int ezca_status;
	long count_field;

	seconds = timer->value;

	epics_timer = (MX_EPICS_TIMER *) ( timer->record->record_type_struct );

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_TIMER structure pointer is NULL.");
	}

	/* Set the number of seconds to count for.  This assumes that
	 * the .FREQ field has already been set up correctly.
	 */

	sprintf( pvname, "%s.TP", epics_timer->epics_record_name );

	ezca_status = ezcaPut( pvname, ezcaDouble, 1, &seconds );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to set counting time for EPICS timer '%s' (%s) failed. "
"EZCA status = %d (%s)",
			timer->record->name,
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Start the timer counting. */

	sprintf( pvname, "%s.CNT", epics_timer->epics_record_name );

	count_field = 1;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &count_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to start EPICS timer '%s' (%s) failed. EZCA status = %d (%s)",
			timer->record->name,
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_stop( MX_TIMER *timer )
{
	const char fname[] = "mxd_epics_timer_stop()";

	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int ezca_status;
	long count_field;

	epics_timer = (MX_EPICS_TIMER *) ( timer->record->record_type_struct );

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_TIMER structure pointer is NULL.");
	}

	/* Stop the timer from counting. */

	sprintf( pvname, "%s.CNT", epics_timer->epics_record_name );

	count_field = 0;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &count_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to stop EPICS timer '%s' (%s) failed. EZCA status = %d (%s)",
			timer->record->name,
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	timer->value = 0.0;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_clear( MX_TIMER *timer )
{
	/* It appears that it is not possible to directly clear
	 * an EPICS scaler channel.  So all we can do is return
	 * a success code.
	 */

	timer->value = 0.0;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_read( MX_TIMER *timer )
{
	const char fname[] = "mxd_epics_timer_read()";

	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	double seconds;
	int ezca_status;

	epics_timer = (MX_EPICS_TIMER *)(timer->record->record_type_struct);

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"EPICS_TIMER pointer is NULL.");
	}

	sprintf( pvname, "%s.T", epics_timer->epics_record_name );

	ezca_status = ezcaGet( pvname, ezcaDouble, 1, &seconds );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to read from EPICS timer '%s' failed.  EZCA status = %d (%s)",
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		timer->value = 0.0;

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	timer->value = seconds;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_get_mode( MX_TIMER *timer )
{
	const char fname[] = "mxd_epics_timer_get_mode()";

	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int ezca_status;
	long gate_control;

	epics_timer = (MX_EPICS_TIMER *)(timer->record->record_type_struct);

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"EPICS_TIMER pointer is NULL.");
	}

	/* Get the gate control field G1. */

	sprintf( pvname, "%s.G1", epics_timer->epics_record_name );

	ezca_status = ezcaGet( pvname, ezcaLong, 1, &gate_control );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to get the gate control field for EPICS counter '%s' (%s) failed.  "
"EZCA status = %d (%s)",
			timer->record->name,
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	switch( gate_control ) {
	case 0:
		timer->mode = MXCM_COUNTER_MODE;
		break;

	case 1:
		timer->mode = MXCM_PRESET_MODE;
		break;

	default:
		return mx_error( MXE_CONTROLLER_INTERNAL_ERROR, fname,
		"The gate control field for EPICS timer '%s' (%s) "
		"has the unexpected value %ld.",
			timer->record->name,
			epics_timer->epics_record_name,
			gate_control );
		break;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_set_mode( MX_TIMER *timer )
{
	const char fname[] = "mxd_epics_timer_set_mode()";

	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int ezca_status;
	long gate_control, counter_mode;
	mx_status_type mx_status;

	epics_timer = (MX_EPICS_TIMER *) ( timer->record->record_type_struct );

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_TIMER structure pointer is NULL.");
	}

	/* Choose the correct gate control mode. */

	switch( timer->mode ) {
	case MXCM_PRESET_MODE:
		gate_control = 1;
		break;

	case MXCM_COUNTER_MODE:
		gate_control = 0;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The counter mode %ld requested for timer '%s' is not supported.",
			timer->mode, timer->record->name );
		break;
	}

	/* Make sure the counter is _not_ in continuous count mode. */

	sprintf( pvname, "%s_mode.VAL", epics_timer->epics_record_name );

	counter_mode = 0;	/* One-shot mode. */

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &counter_mode );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to set one-shot mode for EPICS timer '%s' (%s) failed.  "
	"EZCA status = %d (%s)",
			timer->record->name,
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Tell the timer to stop counting. */

	mx_status = mxd_epics_timer_stop( timer );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Set the gate control field G1. */

	sprintf( pvname, "%s.G1", epics_timer->epics_record_name );

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &gate_control );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to set the gate control field for EPICS counter '%s' (%s) failed.  "
"EZCA status = %d (%s)",
			timer->record->name,
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_set_modes_of_associated_counters( MX_TIMER *timer )
{
	const char fname[]
		= "mxd_epics_timer_set_modes_of_associated_counters()";

	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int i, ezca_status;
	long gate_control;

	epics_timer = (MX_EPICS_TIMER *) ( timer->record->record_type_struct );

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_TIMER structure pointer is NULL.");
	}

	/* Start an EPICS synchronous group. */

	ezca_status = ezcaStartGroup();

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
		"Attempt to start an EPICS synchronous group for record '%s' "
		"failed.  EZCA status = %d (%s)",
			timer->record->name,
			ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR,
			fname, mx_error_message);
	}

	/* Set the gate control to 0 for fields for channels 2 through 16.
	 * This is equivalent to setting them all to counter mode.
	 */

	for ( i = 2; i <= 16; i++ ) {

		sprintf( pvname, "%s.G%d", epics_timer->epics_record_name, i );

		gate_control = 0;

		ezca_status = ezcaPut( pvname, ezcaLong, 1, &gate_control );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
			    "Attempt to set the gate control field for the "
			    "associated channel %d of EPICS counter '%s' (%s) "
			    "failed.  EZCA status = %d (%s)",
			    i, timer->record->name,
			    epics_timer->epics_record_name,
			    ezca_status, ezca_error_message );
			    ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR,
				fname, mx_error_message);
		}
	}

	/* Send the synchronous group to EPICS. */

	ezca_status = ezcaEndGroup();

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
		"Execution of EPICS synchronous group for record '%s' "
		"failed.  EZCA status = %d (%s)",
			timer->record->name,
			ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR,
			fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

