/*
 * Name:    d_mmx000.c
 *
 * Purpose: MX driver for the Newport MM3000 & MM4000 4-axis motor controllers.
 *
 *-----------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <float.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "d_mmx000.h"

/* ============ Motor channels ============ */

MX_RECORD_FUNCTION_LIST mxd_mmx000_record_function_list = {
	mxd_mmx000_initialize_type,
	mxd_mmx000_create_record_structures,
	mxd_mmx000_finish_record_initialization,
	mxd_mmx000_delete_record,
	mxd_mmx000_print_structure,
	mxd_mmx000_read_parms_from_hardware,
	mxd_mmx000_write_parms_to_hardware,
	mxd_mmx000_open,
	mxd_mmx000_close,
	NULL,
	mxd_mmx000_resynchronize
};

MX_MOTOR_FUNCTION_LIST mxd_mmx000_motor_function_list = {
	mxd_mmx000_motor_is_busy,
	mxd_mmx000_move_absolute,
	mxd_mmx000_get_position,
	mxd_mmx000_set_position,
	mxd_mmx000_soft_abort,
	mxd_mmx000_immediate_abort,
	mxd_mmx000_positive_limit_hit,
	mxd_mmx000_negative_limit_hit,
	mxd_mmx000_find_home_position,
	mxd_mmx000_constant_velocity_move,
	mxd_mmx000_get_parameter,
	mxd_mmx000_set_parameter
};

#define MMXDEBUG	FALSE

/* MM3000 motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_mm3000_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_STEPPER_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_MMX000_STANDARD_FIELDS
};

long mxd_mm3000_num_record_fields
		= sizeof( mxd_mm3000_record_field_defaults )
			/ sizeof( mxd_mm3000_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_mm3000_rfield_def_ptr
			= &mxd_mm3000_record_field_defaults[0];

/* MM4000 motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_mm4000_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_MMX000_STANDARD_FIELDS
};

long mxd_mm4000_num_record_fields
		= sizeof( mxd_mm4000_record_field_defaults )
			/ sizeof( mxd_mm4000_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_mm4000_rfield_def_ptr
			= &mxd_mm4000_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_mmx000_get_pointers( MX_MOTOR *motor,
			MX_MMX000_MOTOR **mmx000_motor,
			MX_MMX000 **mmx000,
			const char *calling_fname )
{
	const char fname[] = "mxd_mmx000_get_pointers()";

	MX_RECORD *mmx000_record;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( mmx000_motor == (MX_MMX000_MOTOR **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_MMX000_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( mmx000 == (MX_MMX000 **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_MMX000 pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( motor->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
    "The MX_RECORD pointer for the MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*mmx000_motor = (MX_MMX000_MOTOR *) motor->record->record_type_struct;

	if ( *mmx000_motor == (MX_MMX000_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MMX000_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	mmx000_record = (*mmx000_motor)->mmx000_record;

	if ( mmx000_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The mmx000_record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	*mmx000 = (MX_MMX000 *) mmx000_record->record_type_struct;

	if ( *mmx000 == (MX_MMX000 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MMX000 pointer for record '%s' is NULL.",
			motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*============================================*/

MX_EXPORT mx_status_type
mxd_mmx000_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_mmx000_create_record_structures()";

	MX_MOTOR *motor;
	MX_MMX000_MOTOR *mmx000_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	mmx000_motor = (MX_MMX000_MOTOR *) malloc( sizeof(MX_MMX000_MOTOR) );

	if ( mmx000_motor == (MX_MMX000_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MMX000_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = mmx000_motor;
	record->class_specific_function_list
				= &mxd_mmx000_motor_function_list;

	motor->record = record;

	/* This record treats MM3000 motors as if they were steppers with
	 * the encoder tick size as the step size.
	 *
	 * MM4000s are treated as analog motors, since only positioning
	 * commands in engineering units are available.
	 */

	if ( record->mx_type == MXT_MTR_MM3000 ) {
		motor->subclass = MXC_MTR_STEPPER;

	} else {	/* MM4000 */

		motor->subclass = MXC_MTR_ANALOG;
	}

	motor->acceleration_type = MXF_MTR_ACCEL_DIST_PER_SECOND_SQUARED;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_mmx000_finish_record_initialization()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Check to see if the axis number is valid. */

	mmx000_motor = (MX_MMX000_MOTOR *) (record->record_type_struct);

	if ( mmx000_motor == (MX_MMX000_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MMX000_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	if ( mmx000_motor->axis_number < 1 || mmx000_motor->axis_number > 4 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Illegal axis number %d for MX_MMX000_MOTOR.  Allowed range from 1 to 4.",
			mmx000_motor->axis_number );
	}

	/* Save a pointer to the motor's MX_RECORD structure in the MX_MMX000
	 * structure, so that the four motors can find each other.
	 */

	if ( mmx000_motor->mmx000_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MMX000 interface record pointer is NULL for motor '%s'.",
			record->name );
	}

	mmx000 = (MX_MMX000 *)
		(mmx000_motor->mmx000_record->record_type_struct);

	if ( mmx000 == (MX_MMX000 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MMX000 pointer for interface record '%s' is NULL.",
			mmx000_motor->mmx000_record->name );
	}

	mmx000->motor_record[ mmx000_motor->axis_number - 1 ] = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_mmx000_print_structure()";

	MX_MOTOR *motor;
	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);

	if ( record->mx_type == MXT_MTR_MM3000 ) {
		fprintf(file, "  Motor type     = MM3000 motor.\n\n");
	} else if ( record->mx_type == MXT_MTR_MM4000 ) {
		fprintf(file, "  Motor type     = MM4000 motor.\n\n");
	} else {
		fprintf(file, "  Motor type     = Unknown (value = %ld)",
			record->mx_type);
	}

	fprintf(file, "  name           = %s\n", record->name);
	fprintf(file, "  port name      = %s\n", mmx000->record->name);
	fprintf(file, "  axis number    = %d\n", mmx000_motor->axis_number);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'", record->name);
	}
	
	if ( record->mx_type == MXT_MTR_MM3000 ) {
		fprintf(file, "  position       = %g %s  (%ld steps)\n",
			motor->position, motor->units,
			motor->raw_position.stepper );
		fprintf(file, "  scale          = %g %s per step.\n",
			motor->scale, motor->units);
		fprintf(file, "  offset         = %g %s.\n",
			motor->offset, motor->units);
	
		fprintf(file, "  backlash       = %g %s  (%ld steps)\n",
			motor->backlash_correction, motor->units,
			motor->raw_backlash_correction.stepper);
	
		fprintf(file, "  negative limit = %g %s  (%ld steps)\n",
			motor->negative_limit, motor->units,
			motor->raw_negative_limit.stepper);

		fprintf(file, "  positive limit = %g %s  (%ld steps)\n",
			motor->positive_limit, motor->units,
			motor->raw_positive_limit.stepper);

		deadband = motor->scale * (double) motor->deadband.stepper;

		fprintf(file, "  deadband       = %g %s  (%ld steps)\n\n",
			deadband, motor->units,
			motor->deadband.stepper );

	} else {	/* MM4000 */
	
		fprintf(file, "  position       = %g %s  (%g)\n",
			motor->position, motor->units,
			motor->raw_position.analog );
		fprintf(file, "  scale          = %g %s per MM4000 unit.\n",
			motor->scale, motor->units);
		fprintf(file, "  offset         = %g %s.\n",
			motor->offset, motor->units);
		fprintf(file, "  backlash       = %g %s  (%g)\n",
			motor->backlash_correction, motor->units,
			motor->raw_backlash_correction.analog );
		fprintf(file, "  negative limit = %g %s  (%g)\n",
			motor->negative_limit, motor->units,
			motor->raw_negative_limit.analog );
		fprintf(file, "  positive limit = %g %s  (%g)\n",
			motor->positive_limit, motor->units,
			motor->raw_positive_limit.analog );

		deadband = motor->scale * motor->deadband.analog;

		fprintf(file, "  deadband       = %g %s  (%g)\n\n",
			deadband, motor->units,
			motor->deadband.analog );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_read_parms_from_hardware( MX_RECORD *record )
{
	const char fname[] = "mxd_mmx000_read_parms_from_hardware()";

	MX_MOTOR *motor;
	MX_MMX000_MOTOR *mmx000_motor;
	MX_MMX000 *mmx000;
	MX_GENERIC *generic;
	double position;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Flush out any unwritten and unread characters before sending
	 * commands to the Newport controller.
	 */

	generic = (MX_GENERIC *) mmx000->record->record_class_struct;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_GENERIC pointer for record '%s' is NULL.",
			mmx000->record->name );
	}

	status = mx_generic_discard_unwritten_output( generic, MMXDEBUG );

	if ( status.code != MXE_SUCCESS && status.code != MXE_UNSUPPORTED )
		return status;

	status = mx_generic_discard_unread_input( generic, MMXDEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the current position.  This has the side effect of
	 * updating the position value in the MX_MOTOR structure.
	 */

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* At present, the position is the only parameter we read out.
	 * We rely on the non-volatile memory of the controller
	 * for the rest of the controller parameters.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_write_parms_to_hardware( MX_RECORD *record )
{
	/* The only thing we might want to restore is the motor position,
	 * but the Newport controllers do not allow us to set the position
	 * to an arbitrary value, so we do nothing here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxd_mmx000_resynchronize()";

	MX_MMX000_MOTOR *mmx000_motor;
	mx_status_type status;

	mmx000_motor = (MX_MMX000_MOTOR *) record->record_type_struct;

	if ( mmx000_motor == (MX_MMX000_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MMX000_MOTOR pointer for motor is NULL." );
	}

	/* Forward the request on to the resynchronize function for
	 * the interface.
	 */

	status = mxi_mmx000_resynchronize( mmx000_motor->mmx000_record );

	return status;
}

/* ============ Motor specific functions ============ */

MX_EXPORT mx_status_type
mxd_mmx000_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_motor_is_busy()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	char command[20];
	char response[20];
	int length;
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( command, "%dMS", mmx000_motor->axis_number );

	status = mxi_mmx000_command( mmx000, command,
				response, sizeof response, MMXDEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to motor status command.");
	}

	if ( (response[0] & 0x1) == 0 ) {
		motor->busy = FALSE;
	} else {
		motor->busy = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_move_absolute()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	double raw_position;
	long motor_steps;
	char command[20];
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->record->mx_type == MXT_MTR_MM3000 ) {

		motor_steps = motor->raw_setpoint.stepper;

		sprintf( command, "%dPA%ld",
				mmx000_motor->axis_number, motor_steps );

	} else {	/* MM4000 */

		raw_position = motor->raw_setpoint.analog;

		sprintf( command, "%dPA%g",
			mmx000_motor->axis_number, raw_position );
	}

	status = mxi_mmx000_command( mmx000, command, NULL, 0, MMXDEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_mmx000_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_get_position()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	double raw_position;
	long motor_steps;
	char command[20];
	char response[50];
	int num_tokens;
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( command, "%dTP", mmx000_motor->axis_number );

	status = mxi_mmx000_command( mmx000, command,
					response, sizeof response, MMXDEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->record->mx_type == MXT_MTR_MM3000 ) {

		num_tokens = sscanf( response, "%ld", &motor_steps );

		if ( num_tokens != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"No position value seen in response to 'TP' command.  "
			"Response seen = '%s'", response );
		}

		motor->raw_position.stepper = motor_steps;

	} else {	/* MM4000 */

		num_tokens = sscanf( response, "%lg", &raw_position );

		if ( num_tokens != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"No position value seen in response to 'TP' command.  "
			"Response seen = '%s'", response );
		}

		motor->raw_position.analog = raw_position;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_set_position()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	char command[20];
	double position;
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->record->mx_type == MXT_MTR_MM3000 ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Newport MM3000s do not support setting the position value "
		"in any way." );
	}

	position = motor->raw_setpoint.analog;

	if ( fabs(position) > MX_MOTOR_ANALOG_FUZZ ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Newport MM4000s do not support setting the current position "
		"to any value other than zero." );
	}

	sprintf( command, "%dZP", mmx000_motor->axis_number );

	status = mxi_mmx000_command( mmx000, command, NULL, 0, MMXDEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_soft_abort()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	char command[20];
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( command, "%dST", mmx000_motor->axis_number );

	status = mxi_mmx000_command( mmx000, command, NULL, 0, MMXDEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_mmx000_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_immediate_abort()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	char command[20];
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	switch( motor->record->mx_type ) {
	case MXT_MTR_MM3000:
		sprintf( command, "%dAB", mmx000_motor->axis_number );
		break;
	case MXT_MTR_MM4000:
		/* The following actually aborts all four motors. */

		sprintf( command, "AB" );
		break;
	default:
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Motor '%s' is not a Newport MM3000 or MM4000 controlled motor.  "
	"Actual motor type = %ld.", motor->record->mx_type );
		break;
	}

	status = mxi_mmx000_command( mmx000, command, NULL, 0, MMXDEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_mmx000_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_positive_limit_hit()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	char command[20];
	char response[20];
	int length;
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( command, "%dMS", mmx000_motor->axis_number );

	status = mxi_mmx000_command( mmx000, command,
					response, sizeof response, MMXDEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to motor status command.");
	}

	if ( (response[0] & 0x8) == 0 ) {
		motor->positive_limit_hit = FALSE;
	} else {
		motor->positive_limit_hit = TRUE;

		if ( motor->record->mx_type == MXT_MTR_MM4000 ) {
			/* Attempt to automatically turn the motors back on. */

			status = mxi_mmx000_turn_motor_drivers_on(
							mmx000, MMXDEBUG );

			if ( status.code != MXE_SUCCESS )
				return status;
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_negative_limit_hit()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	char command[20];
	char response[20];
	int length;
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( command, "%dMS", mmx000_motor->axis_number );

	status = mxi_mmx000_command( mmx000, command,
					response, sizeof response, MMXDEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to motor status command.");
	}

	if ( (response[0] & 0x10) == 0 ) {
		motor->negative_limit_hit = FALSE;
	} else {
		motor->negative_limit_hit = TRUE;

		if ( motor->record->mx_type == MXT_MTR_MM4000 ) {
			/* Attempt to automatically turn the motors back on. */

			status = mxi_mmx000_turn_motor_drivers_on(
							mmx000, MMXDEBUG );

			if ( status.code != MXE_SUCCESS )
				return status;
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_find_home_position()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	char command[20];
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( command, "%dOR", mmx000_motor->axis_number );

	status = mxi_mmx000_command( mmx000, command, NULL, 0, MMXDEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_constant_velocity_move( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_constant_velocity_move()";

	MX_MMX000 *mmx000;
	MX_MMX000_MOTOR *mmx000_motor;
	char command[20];
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->record->mx_type == MXT_MTR_MM3000 ) {

		if ( motor->constant_velocity_move >= 0 ) {
			sprintf( command, "%dMV+", mmx000_motor->axis_number );
		} else {
			sprintf( command, "%dMV-", mmx000_motor->axis_number );
		}

		status = mxi_mmx000_command( mmx000, command,
						NULL, 0, MMXDEBUG );

		if ( status.code != MXE_SUCCESS )
			return status;

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
	"Constant velocity moves are not supported by the MM4000 or MM4005." );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_get_parameter()";

	MX_MMX000_MOTOR *mmx000_motor;
	MX_MMX000 *mmx000;
	char command[80];
	char response[80];
	int num_items;
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->parameter_type == MXLV_MTR_SPEED ) {

		sprintf( command, "%dDV", mmx000_motor->axis_number );

		status = mxi_mmx000_command( mmx000, command,
				response, sizeof(response), MMXDEBUG );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( ( motor->record->mx_type == MXT_MTR_MM3000 )
		  && ( strchr( response, '-' ) != NULL ) )
		{
			/* MM3000 controlled stepper motor. */

			num_items = sscanf( response, "%*s-%lg",
							&(motor->raw_speed) );
		} else {
			/* All others. */

			num_items = sscanf( response, "%lg",
							&(motor->raw_speed) );
		}

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Unable to parse returned speed for motor '%s'.  Response = '%s'",
				motor->record->name, response );
		}

	} else if ( motor->parameter_type == MXLV_MTR_BASE_SPEED ) {

		if ( motor->record->mx_type == MXT_MTR_MM3000 ) {

			sprintf( command, "%dDV", mmx000_motor->axis_number );

			status = mxi_mmx000_command( mmx000, command,
				response, sizeof(response), MMXDEBUG );

			if ( status.code != MXE_SUCCESS )
				return status;

			if ( strchr( response, '-' ) != NULL ) {

				/* MM3000 controlled stepper motor. */

				num_items = sscanf( response, "%lg-%*s",
						&(motor->raw_base_speed) );
			} else {
				/* MM3000 controlled servo motor. */

				motor->raw_base_speed = 0.0;
			}
		} else {				/* MM4000 */
			motor->raw_base_speed = 0.0;
		}

	} else if ( motor->parameter_type == MXLV_MTR_MAXIMUM_SPEED ) {

		motor->raw_maximum_speed = DBL_MAX;

	} else if ( motor->parameter_type == MXLV_MTR_ACCELERATION_PARAMETERS){

		sprintf( command, "%dA", mmx000_motor->axis_number );

		status = mxi_mmx000_command( mmx000, command,
				response, sizeof(response), MMXDEBUG );

		if ( status.code != MXE_SUCCESS )
			return status;

		num_items = sscanf( response, "%lg",
					&(motor->acceleration_parameters[0]) );

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Unable to parse returned speed for motor '%s'.  Response = '%s'",
				motor->record->name, response );
		}

		motor->acceleration_parameters[1] = 0.0;
		motor->acceleration_parameters[2] = 0.0;
		motor->acceleration_parameters[3] = 0.0;

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			motor->parameter_type );
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mmx000_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_mmx000_set_parameter()";

	MX_MMX000_MOTOR *mmx000_motor;
	MX_MMX000 *mmx000;
	char command[80];
	mx_status_type status;

	status = mxd_mmx000_get_pointers( motor, &mmx000_motor,
						&mmx000, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Format the appropriate parameter setting command. */

	if ( motor->parameter_type == MXLV_MTR_SPEED ) {

		sprintf( command, "%dVA%g", mmx000_motor->axis_number,
						motor->raw_speed );

	} else if ( motor->parameter_type == MXLV_MTR_BASE_SPEED ) {

		if ( motor->record->mx_type == MXT_MTR_MM3000 ) {

			sprintf( command, "%dVB%g", mmx000_motor->axis_number,
						motor->raw_base_speed );
		} else {
			motor->raw_base_speed = 0.0;

			return MX_SUCCESSFUL_RESULT;
		}

	} else if ( motor->parameter_type == MXLV_MTR_MAXIMUM_SPEED ) {

		motor->raw_maximum_speed = DBL_MAX;

		return MX_SUCCESSFUL_RESULT;

	} else if ( motor->parameter_type == MXLV_MTR_ACCELERATION_PARAMETERS){

		sprintf( command, "%dAC%g", mmx000_motor->axis_number,
					motor->acceleration_parameters[0] );

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			motor->parameter_type );
	}

	/* Send the command to the Newport controller. */

	status = mxi_mmx000_command( mmx000,
				command, NULL, 0, MMXDEBUG );

	return status;
}

