/*
 * Name:    d_mono.c 
 *
 * Purpose: MX motor driver to control X-ray monochromators.  This is a
 *          pseudomotor that pretends to be the monochromator 'theta' axis.
 *          The value written to this pseudo motor is used to control the
 *          real 'theta' axis and any other MX devices that are tied into
 *          this pseudomotor.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_var.h"
#include "mx_alloc.h"
#include "mx_const.h"
#include "mx_motor.h"
#include "d_mono.h"
#include "d_monop.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_monochromator_record_function_list = {
	mxd_monochromator_initialize_type,
	mxd_monochromator_create_record_structures,
	mxd_monochromator_finish_record_initialization,
	mxd_monochromator_delete_record,
	mxd_monochromator_print_motor_structure,
	mxd_monochromator_read_parms_from_hardware,
	mxd_monochromator_write_parms_to_hardware,
	mxd_monochromator_open,
	mxd_monochromator_close
};

MX_MOTOR_FUNCTION_LIST mxd_monochromator_motor_function_list = {
	mxd_monochromator_motor_is_busy,
	mxd_monochromator_move_absolute,
	mxd_monochromator_get_position,
	mxd_monochromator_set_position,
	mxd_monochromator_soft_abort,
	mxd_monochromator_immediate_abort,
	mxd_monochromator_positive_limit_hit,
	mxd_monochromator_negative_limit_hit,
	mxd_monochromator_find_home_position,
	mxd_monochromator_constant_velocity_move,
	mxd_monochromator_get_parameter,
	mxd_monochromator_set_parameter
};

/* Monochromator motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_monochromator_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_MONOCHROMATOR_STANDARD_FIELDS
};

long mxd_monochromator_num_record_fields
		= sizeof( mxd_monochromator_record_field_defaults )
			/ sizeof( mxd_monochromator_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_monochromator_rfield_def_ptr
			= &mxd_monochromator_record_field_defaults[0];

/* === */

static MX_MONOCHROMATOR_FUNCTION_LIST mxd_monochromator_function_list[] = {
	{ mxd_monochromator_get_theta_position,
	  mxd_monochromator_set_theta_position,
	  mxd_monochromator_move_theta},
	{ mxd_monochromator_get_energy_position,
	  mxd_monochromator_set_energy_position,
	  mxd_monochromator_move_energy},
	{ mxd_monochromator_get_polynomial_position,
	  mxd_monochromator_set_polynomial_position,
	  mxd_monochromator_move_polynomial},
	{ mxd_monochromator_get_insertion_device_energy_position,
	  mxd_monochromator_set_insertion_device_energy_position,
	  mxd_monochromator_move_insertion_device_energy},
	{ mxd_monochromator_get_bragg_normal_position,
	  mxd_monochromator_set_bragg_normal_position,
	  mxd_monochromator_move_bragg_normal},
	{ mxd_monochromator_get_bragg_translation_position,
	  mxd_monochromator_set_bragg_translation_position,
	  mxd_monochromator_move_bragg_translation},
	{ mxd_monochromator_get_table_position,
	  mxd_monochromator_set_table_position,
	  mxd_monochromator_move_table},
	{ mxd_monochromator_get_diffractometer_theta_position,
	  mxd_monochromator_set_diffractometer_theta_position,
	  mxd_monochromator_move_diffractometer_theta}
};

#if 0
static int mx_num_monochromator_functions
			= sizeof(mxd_monochromator_function_list)
				/ sizeof(mxd_monochromator_function_list[0]);
#endif

/************** Internal use only functions. *************/

static mx_status_type
mxd_monochromator_get_pointers( MX_MOTOR *motor,
				MX_MONOCHROMATOR **monochromator,
				const char *calling_fname )
{
	const char fname[] = "mxd_monochromator_get_pointers()";

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_MOTOR pointer passed by '%s' is NULL.",
			calling_fname );
	}
	if ( monochromator == (MX_MONOCHROMATOR **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_MONOCHROMATOR indirect pointer passed by '%s' is NULL.",
			calling_fname);
	}
	if ( motor->record->mx_type != MXT_MTR_MONOCHROMATOR ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"Record '%s' is not a monochromator record.  Instead, it is of type %ld.",
			motor->record->name, motor->record->mx_type );
	}
	*monochromator = (MX_MONOCHROMATOR *)
				(motor->record->record_type_struct);

	if ( *monochromator == (MX_MONOCHROMATOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MONOCHROMATOR pointer for motor '%s' is NULL.",
			motor->record->name );
	}
	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_get_enable_status( MX_RECORD *list_record,
					int *enable_status )
{
	const char fname[] = "mxd_monochromator_get_enable_status()";

	MX_RECORD **record_array;
	MX_RECORD *enable_status_record;
	void *pointer_to_value;
	mx_status_type status;

	status = mx_get_variable_pointer( list_record, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	record_array = (MX_RECORD **) pointer_to_value;

	enable_status_record = record_array[MXFP_MONO_ENABLED];

	status = mx_receive_variable( enable_status_record );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_get_variable_pointer( enable_status_record,
					&pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	*enable_status = *( int * ) pointer_to_value;

	MX_DEBUG( 1,("%s: '%s' enable status = %d",
		fname, list_record->name, *enable_status));

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_get_dependency_type( MX_RECORD *list_record,
					int *dependency_type )
{
	MX_RECORD **record_array;
	MX_RECORD *dependency_type_record;
	void *pointer_to_value;
	mx_status_type status;

	status = mx_get_variable_pointer( list_record, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	record_array = (MX_RECORD **) pointer_to_value;

	dependency_type_record = record_array[MXFP_MONO_DEPENDENCY_TYPE];

	status = mx_receive_variable( dependency_type_record );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_get_variable_pointer( dependency_type_record,
					&pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	*dependency_type = *( int * ) pointer_to_value;

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_get_param_array_from_list( MX_RECORD *list_record,
					int *num_parameters,
					double **parameter_array )
{
	const char fname[] ="mxd_monochromator_get_param_array_from_list()";

	MX_RECORD_FIELD *field;
	MX_RECORD **record_array;
	MX_RECORD *parameter_array_record;
	void *pointer_to_value;
	mx_status_type status;

	status = mx_get_variable_pointer( list_record, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	record_array = (MX_RECORD **) pointer_to_value;

	parameter_array_record = record_array[MXFP_MONO_PARAMETER_ARRAY];

	status = mx_receive_variable( parameter_array_record );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_find_record_field(parameter_array_record, "value", &field);

	if ( status.code != MXE_SUCCESS )
		return status;

	pointer_to_value = mx_get_field_value_pointer( field );

	if ( pointer_to_value == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Field value pointer is NULL for record field '%s' in record '%s'.",
			field->name, list_record->name );
	}

	*num_parameters = (int) (field->dimension[0]);

	*parameter_array = ( double * ) pointer_to_value;

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_get_record_array_from_list( MX_RECORD *list_record,
					int *num_records,
					MX_RECORD ***returned_record_array )
{
	const char fname[] = "mxd_monochromator_get_record_array_from_list()";

	MX_RECORD_FIELD *field;
	MX_RECORD **record_array;
	MX_RECORD *record_array_record;
	void *pointer_to_value;
	mx_status_type status;

	status = mx_get_variable_pointer( list_record, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	record_array = (MX_RECORD **) pointer_to_value;

	record_array_record = record_array[MXFP_MONO_RECORD_ARRAY];

	status = mx_find_record_field(record_array_record, "value", &field);

	if ( status.code != MXE_SUCCESS )
		return status;

	pointer_to_value = mx_get_field_value_pointer( field );

	if ( pointer_to_value == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Field value pointer is NULL for record field '%s' in record '%s'.",
			field->name, list_record->name );
	}

	*num_records = field->dimension[0];

	*returned_record_array = ( MX_RECORD ** ) pointer_to_value;

	return MX_SUCCESSFUL_RESULT;
}

/********************************************************/

#define NUM_MONO_FIELDS   1

MX_EXPORT mx_status_type
mxd_monochromator_initialize_type( long record_type )
{
	const char fname[] = "mxd_monochromator_initialize_type()";

	const char field_name[NUM_MONO_FIELDS][MXU_FIELD_NAME_LENGTH+1]
	   = {
		"list_array"
	     };

	MX_DRIVER *driver;
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
	MX_RECORD_FIELD_DEFAULTS *field;
	int i;
	long num_record_fields;
	long num_dependencies_field_index;
	long num_dependencies_varargs_cookie;
	mx_status_type status;

	driver = mx_get_driver_by_type( mx_type_list, record_type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record type %ld not found in mx_type_list.",
			record_type );
	}

	record_field_defaults_ptr
			= driver->record_field_defaults_ptr;

	if (record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	record_field_defaults = *record_field_defaults_ptr;

	if ( record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	if ( driver->num_record_fields == (long *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'num_record_fields' pointer for record type '%s' is NULL.",
			driver->name );
	}

	num_record_fields = *(driver->num_record_fields);

	status = mx_find_record_field_defaults_index(
			record_field_defaults, num_record_fields,
			"num_dependencies", &num_dependencies_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie( num_dependencies_field_index, 0,
				&num_dependencies_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	for ( i = 0; i < NUM_MONO_FIELDS; i++ ) {
		status = mx_find_record_field_defaults(
			record_field_defaults, num_record_fields,
			field_name[i], &field );

		if ( status.code != MXE_SUCCESS )
			return status;

		field->dimension[0] = num_dependencies_varargs_cookie;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_monochromator_create_record_structures()";

	MX_MOTOR *motor;
	MX_MONOCHROMATOR *monochromator;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	monochromator = (MX_MONOCHROMATOR *) malloc( sizeof(MX_MOTOR) );

	if ( monochromator == (MX_MONOCHROMATOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_MONOCHROMATOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = monochromator;
	record->class_specific_function_list
				= &mxd_monochromator_motor_function_list;

	motor->record = record;

	/* A DAC motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_finish_record_initialization( MX_RECORD *record )
{
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	return status;
}

MX_EXPORT mx_status_type
mxd_monochromator_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_print_motor_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_monochromator_print_motor_structure()";

	MX_MOTOR *motor;
	MX_MONOCHROMATOR *monochromator;
	double position, deadband;
	int i, num_dependencies;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	monochromator = (MX_MONOCHROMATOR *) (record->record_type_struct);

	if ( monochromator == (MX_MONOCHROMATOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MONOCHROMATOR pointer for record '%s' is NULL.",
			record->name );
	}

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type                 = MONOCHROMATOR.\n\n");

	fprintf(file, "  name                       = %s\n", record->name);

	num_dependencies = monochromator->num_dependencies;

	fprintf(file, "  list array                 = (");

	if ( monochromator->list_array == NULL ) {
		fprintf(file, "NULL");
	} else {
		for ( i = 0; i < num_dependencies; i++ ) {
			fprintf(file, "%s",
				monochromator->list_array[i]->name);

			if ( i != (num_dependencies - 1) ) {
				fprintf(file, ", ");
			}
		}
	}
	fprintf(file, ")\n");

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf(file, "  position                   = %g %s  (%g).\n",
		motor->position, motor->units,
		motor->raw_position.analog );
	fprintf(file, "  theta scale                = %g %s "
					"per unscaled theta unit.\n",
		motor->scale, motor->units);
	fprintf(file, "  theta offset               = %g %s.\n",
		motor->offset, motor->units);
	fprintf(file, "  backlash                   = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog );
	fprintf(file, "  negative limit             = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog );
	fprintf(file, "  positive limit             = %g %s  (%g).\n",
		motor->positive_limit, motor->units,
		motor->raw_positive_limit.analog );

	deadband = motor->scale * motor->deadband.analog;

	fprintf(file, "  deadband                   = %g %s  (%g).\n\n",
		deadband, motor->units,
		motor->deadband.analog );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_monochromator_motor_is_busy()";

	MX_MONOCHROMATOR *monochromator;
	MX_RECORD *list_record;
	MX_RECORD **record_array;
	MX_RECORD *record;
	int i, j, num_records, busy, enabled;
	mx_status_type status;

	MX_DEBUG( 1,("%s invoked.", fname));

	status = mxd_monochromator_get_pointers(motor, &monochromator, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	motor->busy = FALSE;

	for ( i = 0; i < monochromator->num_dependencies; i++ ) {
		list_record = monochromator->list_array[i];

		status = mxd_monochromator_get_enable_status(
						list_record, &enabled );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( enabled ) {
			status = mxd_monochromator_get_record_array_from_list(
				list_record, &num_records, &record_array );

			if ( status.code != MXE_SUCCESS )
				return status;

			for ( j = 0; j < num_records; j++ ) {

				record = record_array[j];

				if ( (record->mx_superclass == MXR_DEVICE)
				  && (record->mx_class == MXC_MOTOR) )
				{
					status = mx_motor_is_busy(
							record, &busy );

					if ( status.code != MXE_SUCCESS )
						return status;

					if ( busy ) {
						motor->busy = TRUE;

						/* Exit the for(j) loop. */

						break;
					}
				}
			}
			if ( busy ) {
				break;   /* Exit the for(i) loop. */
			}
		}
	}
	MX_DEBUG( 1,("%s returning.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_monochromator_move_absolute()";

	MX_MONOCHROMATOR *monochromator;
	MX_RECORD *list_record;
	MX_MONOCHROMATOR_FUNCTION_LIST *flist;
	mx_status_type (*fptr)( MX_MONOCHROMATOR *, MX_RECORD *,
					double, double, int );
	double raw_setpoint, old_raw_position, dummy;
	int i, dependency_type;
	mx_status_type status;

	status = mxd_monochromator_get_pointers(motor, &monochromator, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	raw_setpoint = motor->raw_setpoint.analog;

	/* Get the old position. */

	status = mx_motor_get_position( motor->record, &dummy );

	if ( status.code != MXE_SUCCESS )
		return status;

	old_raw_position = motor->raw_position.analog;

	/* Check the software limits for each of the dependencies. */

	MX_DEBUG( 1,("%s: ***** Checking software limits.", fname));

	for ( i = 0; i < monochromator->num_dependencies; i++ ) {

		list_record = monochromator->list_array[i];

		if ( list_record == (MX_RECORD *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"Record pointer for element %d of list_array in monochromator '%s' is NULL.",
				i, motor->record->name );
		}

		status = mxd_monochromator_get_dependency_type( list_record,
							&dependency_type );

		if ( status.code != MXE_SUCCESS )
			return status;

		flist = &mxd_monochromator_function_list[ dependency_type ];

		if ( flist == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Monochromator function list entry for dependency type %d is NULL.",
				dependency_type );
		}

		fptr = flist->move;

		if ( fptr == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"move function pointer for monochromator dependency type %d is NULL.",
				dependency_type );
		}

		status = (*fptr)( monochromator, list_record,
					raw_setpoint, old_raw_position,
					MXF_MTR_ONLY_CHECK_LIMITS );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	/* Now start the move for each of the dependencies. */

	MX_DEBUG( 1,("%s: ***** Now starting the move.", fname));

	for ( i = 0; i < monochromator->num_dependencies; i++ ) {

		/* Having tested for NULL pointers on the first pass, we
		 * do not need to test for them the second time through.
		 */

		list_record = monochromator->list_array[i];

		status = mxd_monochromator_get_dependency_type( list_record,
							&dependency_type );

		if ( status.code != MXE_SUCCESS )
			return status;

		flist = &mxd_monochromator_function_list[ dependency_type ];

		fptr = flist->move;

		status = (*fptr)( monochromator, list_record,
					raw_setpoint, old_raw_position, 0 );

		if ( status.code != MXE_SUCCESS )
			return status;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_monochromator_get_position()";

	MX_MONOCHROMATOR *monochromator;
	MX_RECORD *list_record;
	MX_MONOCHROMATOR_FUNCTION_LIST *flist;
	mx_status_type (*fptr)( MX_MONOCHROMATOR *, MX_RECORD *, double * );
	double position;
	int dependency_type;
	mx_status_type status;

	status = mxd_monochromator_get_pointers(motor, &monochromator, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the position using the first entry in the list array. */

	list_record = monochromator->list_array[0];

	if ( list_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"First record pointer for list_array in monochromator '%s' is NULL.",
			motor->record->name );
	}

	status = mxd_monochromator_get_dependency_type( list_record,
							&dependency_type );

	if ( status.code != MXE_SUCCESS )
		return status;

	flist = &mxd_monochromator_function_list[ dependency_type ];

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Monochromator function list entry for dependency type %d is NULL.",
			dependency_type );
	}

	fptr = flist->get_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"get_position pointer for monochromator dependency type %d is NULL.",
			dependency_type );
	}

	status = (*fptr)( monochromator, list_record, &position );

	if ( status.code != MXE_SUCCESS )
		return status;

	motor->raw_position.analog = position;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_monochromator_set_position()";

	MX_MONOCHROMATOR *monochromator;
	MX_RECORD *list_record;
	MX_MONOCHROMATOR_FUNCTION_LIST *flist;
	mx_status_type (*fptr)( MX_MONOCHROMATOR *, MX_RECORD *, double );
	double raw_setpoint;
	int dependency_type;
	mx_status_type status;

	status = mxd_monochromator_get_pointers(motor, &monochromator, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	raw_setpoint = motor->raw_setpoint.analog;

	/* Set the position using the first entry in the list array. */

	list_record = monochromator->list_array[0];

	if ( list_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"First record pointer for list_array in monochromator '%s' is NULL.",
			motor->record->name );
	}

	status = mxd_monochromator_get_dependency_type( list_record,
							&dependency_type );

	if ( status.code != MXE_SUCCESS )
		return status;

	flist = &mxd_monochromator_function_list[ dependency_type ];

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Monochromator function list entry for dependency type %d is NULL.",
			dependency_type );
	}

	fptr = flist->set_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"get_position pointer for monochromator dependency type %d is NULL.",
			dependency_type );
	}

	status = (*fptr)( monochromator, list_record, raw_setpoint );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_monochromator_soft_abort()";

	MX_MONOCHROMATOR *monochromator;
	MX_RECORD *list_record;
	MX_RECORD **record_array;
	MX_RECORD *record;
	int i, j, num_records;
	mx_status_type status;

	MX_DEBUG( 1,("%s invoked.",fname));

	status = mxd_monochromator_get_pointers(motor, &monochromator, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	for ( i = 0; i < monochromator->num_dependencies; i++ ) {
		list_record = monochromator->list_array[i];

		status = mxd_monochromator_get_record_array_from_list(
				list_record, &num_records, &record_array );

		if ( status.code != MXE_SUCCESS )
			return status;

		for ( j = 0; j < num_records; j++ ) {

			record = record_array[j];

			if ( (record->mx_superclass == MXR_DEVICE)
			  && (record->mx_class == MXC_MOTOR) )
			{
				(void) mx_motor_soft_abort( record );
			}
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_monochromator_immediate_abort()";

	MX_MONOCHROMATOR *monochromator;
	MX_RECORD *list_record;
	MX_RECORD **record_array;
	MX_RECORD *record;
	int i, j, num_records;
	mx_status_type status;

	MX_DEBUG( 1,("%s invoked.",fname));

	status = mxd_monochromator_get_pointers(motor, &monochromator, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	for ( i = 0; i < monochromator->num_dependencies; i++ ) {
		list_record = monochromator->list_array[i];

		status = mxd_monochromator_get_record_array_from_list(
				list_record, &num_records, &record_array );

		if ( status.code != MXE_SUCCESS )
			return status;

		for ( j = 0; j < num_records; j++ ) {

			record = record_array[j];

			if ( (record->mx_superclass == MXR_DEVICE)
			  && (record->mx_class == MXC_MOTOR) )
			{
				(void) mx_motor_immediate_abort( record );
			}
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

/* With such complicated dependencies possible, it is difficult to say
 * when a limit is tripped whether or not it should be regarded as tripping
 * the 'positive' limit or as tripping the 'negative' limit.  Thus, testing
 * of both limits is folded into the same function.
 */

MX_EXPORT mx_status_type
mxd_monochromator_positive_limit_hit( MX_MOTOR *motor )
{
	return mxd_monochromator_any_limit_hit( motor, 1 );
}

MX_EXPORT mx_status_type
mxd_monochromator_negative_limit_hit( MX_MOTOR *motor )
{
	return mxd_monochromator_any_limit_hit( motor, -1 );
}

MX_EXPORT mx_status_type
mxd_monochromator_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_monochromator_find_home_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
	    "Home search is not supported for monochromator pseudomotors." );
}

MX_EXPORT mx_status_type
mxd_monochromator_constant_velocity_move( MX_MOTOR *motor )
{
	const char fname[] = "mxd_monochromator_constant_velocity_move()";

	MX_MONOCHROMATOR *monochromator;
	MX_RECORD *list_record;
	MX_RECORD **record_array;
	MX_RECORD *record;
	int num_records;
	mx_status_type status;

	status = mxd_monochromator_get_pointers(motor, &monochromator, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the first entry in the list array. */

	list_record = monochromator->list_array[0];

	if ( list_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"First record pointer for list_array in monochromator '%s' is NULL.",
			motor->record->name );
	}

	status = mxd_monochromator_get_record_array_from_list(
				list_record, &num_records, &record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_records <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The first dependency for monochromator motor '%s' has no associated records.",
			motor->record->name );
	}

	/* Get the first entry in the record array. */

	record = record_array[0];

	if ( (record->mx_superclass != MXR_DEVICE)
	  || (record->mx_class != MXC_MOTOR) )
	{
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"The first record '%s' in the first dependency list for "
		"monochromator motor '%s' is not a motor.",
			record->name, motor->record->name );
	}

	/* Start the move using this record. */

	status = mx_motor_constant_velocity_move( record,
				motor->constant_velocity_move );

	return status;
}

MX_EXPORT mx_status_type
mxd_monochromator_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_monochromator_get_parameter()";

	MX_MONOCHROMATOR *monochromator;
	MX_RECORD *list_record;
	MX_RECORD **record_array;
	MX_RECORD *associated_record;
	MX_MOTOR *associated_motor;
	int num_records;
	double associated_motor_speed;
	mx_status_type status;

	status = mxd_monochromator_get_pointers(motor, &monochromator, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the first entry in the list array. */

	list_record = monochromator->list_array[0];

	if ( list_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"First record pointer for list_array in monochromator '%s' is NULL.",
			motor->record->name );
	}

	status = mxd_monochromator_get_record_array_from_list(
				list_record, &num_records, &record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_records <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The first dependency for monochromator motor '%s' has no associated records.",
			motor->record->name );
	}

	/* Get the first entry in the record array. */

	associated_record = record_array[0];

	if ( (associated_record->mx_superclass != MXR_DEVICE)
	  || (associated_record->mx_class != MXC_MOTOR) )
	{
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"The first record '%s' in the first dependency list for "
		"monochromator motor '%s' is not a motor.",
			associated_record->name, motor->record->name );
	}

	associated_motor = (MX_MOTOR *) associated_record->record_class_struct;

	if ( associated_motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The MX_MOTOR pointer for associated motor '%s' "
			"of monochromator motor '%s' is NULL.",
			associated_record->name,
			motor->record->name );
	}

	/* Only getting the motor speed is supported. */

	if ( motor->parameter_type != MXLV_MTR_SPEED ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Changing parameters other than the speed is not supported for "
		"monochromator motor '%s'",
			motor->record->name );
	}

	/* Get the speed. */

	status = mx_motor_get_speed( associated_record,
						&associated_motor_speed );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Pass the speed back to the calling routine. */

	motor->raw_speed = associated_motor_speed;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_monochromator_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_monochromator_set_parameter()";

	MX_MONOCHROMATOR *monochromator;
	MX_RECORD *list_record;
	MX_RECORD **record_array;
	MX_RECORD *associated_record;
	MX_MOTOR *associated_motor;
	int num_records;
	mx_status_type status;

	status = mxd_monochromator_get_pointers(motor, &monochromator, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the first entry in the list array. */

	list_record = monochromator->list_array[0];

	if ( list_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"First record pointer for list_array in monochromator '%s' is NULL.",
			motor->record->name );
	}

	status = mxd_monochromator_get_record_array_from_list(
				list_record, &num_records, &record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_records <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The first dependency for monochromator motor '%s' has no associated records.",
			motor->record->name );
	}

	/* Get the first entry in the record array. */

	associated_record = record_array[0];

	if ( (associated_record->mx_superclass != MXR_DEVICE)
	  || (associated_record->mx_class != MXC_MOTOR) )
	{
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"The first record '%s' in the first dependency list for "
		"monochromator motor '%s' is not a motor.",
			associated_record->name, motor->record->name );
	}

	associated_motor = (MX_MOTOR *) associated_record->record_class_struct;

	if ( associated_motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The MX_MOTOR pointer for associated motor '%s' "
			"of monochromator motor '%s' is NULL.",
			associated_record->name,
			motor->record->name );
	}

	/* Only setting the motor speed is supported. */

	if ( motor->parameter_type != MXLV_MTR_SPEED ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Changing parameters other than the speed is not supported for "
		"monochromator motor '%s'",
			motor->record->name );
	}

	/* Set the speed. */

	status = mx_motor_set_speed( associated_record, motor->raw_speed );

	return status;
}

/****************** Private function definitions. *****************/

static mx_status_type
mxd_monochromator_any_limit_hit( MX_MOTOR *motor, int direction )
{
	const char fname[] = "mxd_monochromator_any_limit_hit()";

	MX_MONOCHROMATOR *monochromator;
	MX_RECORD *list_record;
	MX_RECORD **record_array;
	MX_RECORD *record;
	int i, j, limit_hit, num_records, enabled;
	mx_status_type status;

	MX_DEBUG( 1,("%s invoked.", fname));

	status = mxd_monochromator_get_pointers(motor, &monochromator, fname);

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( direction >= 0 ) {
		motor->positive_limit_hit = FALSE;
	} else {
		motor->negative_limit_hit = FALSE;
	}

	for ( i = 0; i < monochromator->num_dependencies; i++ ) {
		list_record = monochromator->list_array[i];

		status = mxd_monochromator_get_enable_status(
						list_record, &enabled );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( enabled ) {
			status = mxd_monochromator_get_record_array_from_list(
				list_record, &num_records, &record_array );

			if ( status.code != MXE_SUCCESS )
				return status;

			for ( j = 0; j < num_records; j++ ) {

				record = record_array[j];

				if ( (record->mx_superclass == MXR_DEVICE)
				  && (record->mx_class == MXC_MOTOR) )
				{
					status = mx_motor_positive_limit_hit(
						record, &limit_hit );

					if ( status.code != MXE_SUCCESS )
						return status;

					if ( limit_hit ) {
						if ( direction >= 0 ) {
						    motor->positive_limit_hit
								= TRUE;
						} else {
						    motor->negative_limit_hit
								= TRUE;
						}
						break;/* Exit the for() loop.*/
					}

					status = mx_motor_negative_limit_hit(
						record, &limit_hit );

					if ( status.code != MXE_SUCCESS )
						return status;

					if ( limit_hit ) {
						if ( direction >= 0 ) {
						    motor->positive_limit_hit
								= TRUE;
						} else {
						    motor->negative_limit_hit
								= TRUE;
						}
						break;/* Exit the for() loop.*/
					}
				}
			}
		}
	}
	MX_DEBUG( 1,("%s returning.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/******************************************************************/

static mx_status_type
mxd_monochromator_get_theta_position( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double *theta )
{
	MX_RECORD **theta_record_array;
	MX_RECORD *theta_record;
	int num_records;
	mx_status_type status;

	status = mxd_monochromator_get_record_array_from_list(
			list_record, &num_records, &theta_record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	theta_record = theta_record_array[0];

	status = mx_motor_get_position( theta_record, theta );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_set_theta_position( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta )
{
	MX_RECORD **theta_record_array;
	MX_RECORD *theta_record;
	int num_records;
	mx_status_type status;

	status = mxd_monochromator_get_record_array_from_list(
			list_record, &num_records, &theta_record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	theta_record = theta_record_array[0];

	status = mx_motor_set_position( theta_record, theta );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_move_theta( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta,
					double old_theta,
					int flags )
{
	const char fname[] = "mxd_monochromator_move_theta()";

	MX_RECORD **theta_record_array;
	MX_RECORD *theta_record;
	int enabled, num_records;
	mx_status_type status;

	/* Is this dependency enabled? */

	status = mxd_monochromator_get_enable_status( list_record, &enabled );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( enabled == FALSE )
		return MX_SUCCESSFUL_RESULT;

	/* We are enabled, so move the motor. */

	status = mxd_monochromator_get_record_array_from_list(
			list_record, &num_records, &theta_record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	theta_record = theta_record_array[0];

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		MX_DEBUG( 1,("%s: Checking '%s' limits for %g",
				fname, theta_record->name, theta));
	} else {
		MX_DEBUG( 1,("%s: Moving '%s' to %g",
				fname, theta_record->name, theta));
	}

	status = mx_motor_move_absolute( theta_record, theta, NULL, flags );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

/******************************************************************/

static mx_status_type
mxd_mono_get_energy_pointers( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					MX_RECORD **energy_record,
					double *radians_per_engineering_unit,
					double *d_spacing )
{
	const char fname[] = "mxd_mono_get_energy_pointers()";

	MX_RECORD **record_array;
	double *parameter_array;
	int num_records, num_parameters;
	long num_elements;
	void *pointer_to_value;
	mx_status_type status;

	status = mxd_monochromator_get_param_array_from_list(
			list_record, &num_parameters, &parameter_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_parameters != 1 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The parameter array for list record '%s' does not have 1 parameter.  "
"Instead, it has %d parameters.", list_record->name, num_parameters );
	}

	*radians_per_engineering_unit = parameter_array[0];

	if ( fabs(*radians_per_engineering_unit) < 1.0e-100 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The first element (radians per engineering unit) of the parameter array for "
"list record '%s' is too small.  Current value = %g",
			list_record->name, *radians_per_engineering_unit );
	}

	status = mxd_monochromator_get_record_array_from_list(
			list_record, &num_records, &record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_records != 2 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The record array for list record '%s' should have 2 records in it.  "
	"Instead there are %d records in it.",
			list_record->name, num_records );
	}

	/* The first record is the energy motor record. */

	*energy_record = record_array[0];

	/* The second record in the record array is the monochromator
	 * crystal d spacing.
	 */

	status = mx_get_variable_from_record( record_array[1], MXFT_DOUBLE,
					&num_elements, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	*d_spacing = *( double *) pointer_to_value;

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_get_energy_position( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double *theta )
{
	MX_RECORD *energy_record;
	double energy, d_spacing, sin_theta, denominator;
	double angle_in_radians, radians_per_engineering_unit;
	mx_status_type status;

	status = mxd_mono_get_energy_pointers( monochromator,
			list_record, &energy_record,
			&radians_per_engineering_unit,
			&d_spacing );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_motor_get_position( energy_record, &energy );

	if ( status.code != MXE_SUCCESS )
		return status;

	denominator = 2.0 * d_spacing * energy;

	if ( fabs(denominator) >= 1.0 ) {
		sin_theta = MX_HC / denominator;
	} else {
		if ( MX_HC < fabs(denominator * DBL_MAX) ) {
			sin_theta = MX_HC / denominator;
		} else {
			sin_theta = DBL_MAX;
		}
	}

	if ( fabs(sin_theta) > 1.0 ) {
		angle_in_radians = MX_PI / 2.0;
	} else {
		angle_in_radians = asin( sin_theta );
	}

	*theta = angle_in_radians / radians_per_engineering_unit;

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_set_energy_position( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta )
{
	MX_RECORD *energy_record;
	double energy, d_spacing, denominator;
	double angle_in_radians, radians_per_engineering_unit;
	mx_status_type status;

	status = mxd_mono_get_energy_pointers( monochromator,
			list_record, &energy_record,
			&radians_per_engineering_unit,
			&d_spacing );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Compute the energy. */

	angle_in_radians = theta * radians_per_engineering_unit;

	denominator = 2.0 * d_spacing * sin( angle_in_radians );

	if ( fabs(denominator) >= 1.0 ) {
		energy = MX_HC / denominator;
	} else {
		if ( MX_HC < fabs(denominator * DBL_MAX) ) {
			energy = MX_HC / denominator;
		} else {
			energy = DBL_MAX;
		}
	}

	/* Reset the monochromator energy. */

	status = mx_motor_set_position( energy_record, energy );

	return status;
}

static mx_status_type
mxd_monochromator_move_energy( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta,
					double old_theta,
					int flags )
{
	const char fname[] = "mxd_monochromator_move_energy()";

	MX_RECORD *energy_record;
	double energy, d_spacing, denominator;
	double angle_in_radians, radians_per_engineering_unit;
	int enabled;
	mx_status_type status;

	status = mxd_monochromator_get_enable_status( list_record, &enabled );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( enabled == FALSE )
		return MX_SUCCESSFUL_RESULT;

	status = mxd_mono_get_energy_pointers( monochromator,
			list_record, &energy_record,
			&radians_per_engineering_unit,
			&d_spacing );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Compute the energy. */

	angle_in_radians = theta * radians_per_engineering_unit;

	denominator = 2.0 * d_spacing * sin( angle_in_radians );

	energy = mx_divide_safely( MX_HC, denominator );

	/* Move the energy motor to that position. */

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		MX_DEBUG( 1,("%s: Checking '%s' limits for %g",
				fname, energy_record->name, energy));
	} else {
		MX_DEBUG( 1,("%s: Moving '%s' to %g",
				fname, energy_record->name, energy));
	}

	status = mx_motor_move_absolute( energy_record, energy, NULL, flags );

	return status;
}

/******************************************************************/

static mx_status_type
mxd_monochromator_get_polynomial_position(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double *theta )
{
	return MX_SUCCESSFUL_RESULT;
}
static mx_status_type
mxd_monochromator_set_polynomial_position(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta )
{
	return MX_SUCCESSFUL_RESULT;
}
static mx_status_type
mxd_monochromator_move_polynomial( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta,
					double old_theta,
					int flags )
{
	const char fname[] = "mxd_monochromator_move_polynomial()";

	MX_RECORD **record_array;
	MX_RECORD *dependent_motor_record;
	double *parameter_array;
	int i, num_records, num_parameters, enabled;
	double dependent_value;
	mx_status_type status;

	status = mxd_monochromator_get_enable_status( list_record, &enabled );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( enabled == FALSE )
		return MX_SUCCESSFUL_RESULT;

	status = mxd_monochromator_get_param_array_from_list(
			list_record, &num_parameters, &parameter_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxd_monochromator_get_record_array_from_list(
			list_record, &num_records, &record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_records != 1 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The record array for list record '%s' should have 1 record in it.  "
	"Instead there are %d records in it.",
			list_record->name, num_records );
	}

	/* This record should be the dependent motor record. */

	if ( (record_array[0]->mx_superclass != MXR_DEVICE)
	  || (record_array[0]->mx_class != MXC_MOTOR) )
	{
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The first record '%s' in the record array for list record '%s' "
	"is not a motor record.",
			record_array[0]->name, list_record->name );
	}

	dependent_motor_record = record_array[0];

	/* Compute the dependent motor position and move the motor there. */

	dependent_value = parameter_array[ num_parameters - 1 ];

	for ( i = num_parameters - 2; i >= 0; i-- ) {
		dependent_value = parameter_array[i] + theta * dependent_value;
	}

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		MX_DEBUG( 1,("%s: Checking '%s' limits for %g",
			fname, dependent_motor_record->name, dependent_value));
	} else {
		MX_DEBUG( 1,("%s: Moving '%s' to %g",
			fname, dependent_motor_record->name, dependent_value));
	}

	status = mx_motor_move_absolute( dependent_motor_record,
						dependent_value, NULL, flags );

	return status;
}

/******************************************************************/

static mx_status_type
mxd_monochromator_get_insertion_device_energy_position(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double *theta_position )
{
	return MX_SUCCESSFUL_RESULT;
}
static mx_status_type
mxd_monochromator_set_insertion_device_energy_position(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta )
{
	return MX_SUCCESSFUL_RESULT;
}
static mx_status_type
mxd_monochromator_move_insertion_device_energy(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta,
					double old_theta,
					int flags )
{
	const char fname[]
		= "mxd_monochromator_move_insertion_device_energy()";

	MX_RECORD **record_array;
	MX_RECORD *id_motor_record;
	double *parameter_array;
	int num_parameters, num_records;
	long num_elements, gap_harmonic;
	double gap_offset, d_spacing, denominator, mono_energy, id_energy;
	void *pointer_to_value;
	int enabled;
	mx_status_type status;

	status = mxd_monochromator_get_enable_status( list_record, &enabled );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( enabled == FALSE )
		return MX_SUCCESSFUL_RESULT;

	status = mxd_monochromator_get_param_array_from_list(
			list_record, &num_parameters, &parameter_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_parameters != 2 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The parameter array for list record '%s' does not have 2 parameters.  "
"Instead, it has %d parameters.", list_record->name, num_parameters );
	}

	gap_harmonic = mx_round( parameter_array[0] );

	if ( gap_harmonic <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal gap harmonic number %ld.  It should be greater than "
		"or equal to 1.", gap_harmonic );
	}

	gap_offset = parameter_array[1];

	status = mxd_monochromator_get_record_array_from_list(
			list_record, &num_records, &record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_records != 2 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The record array for list record '%s' does not have 2 records.  "
	"Instead, it has %d records.", list_record->name, num_records );
	}

	/* The first record is the insertion device motor record. */

	id_motor_record = record_array[0];

	/* The second record contains the value of the monochromator
	 * crystal d-spacing.
	 */

	status = mx_get_variable_from_record( record_array[1], MXFT_DOUBLE,
					&num_elements, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	d_spacing = *( double * ) pointer_to_value;

	/* Compute the insertion device energy to request. */

	denominator = 2.0 * d_spacing * sin( theta * MX_RADIANS_PER_DEGREE );

	mono_energy = mx_divide_safely( MX_HC, denominator );

	id_energy = ( mono_energy + gap_offset ) / (double) gap_harmonic;

	/* Move the insertion device. */

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		MX_DEBUG( 1,("%s: Checking '%s' limits for %g",
				fname, id_motor_record->name, id_energy));
	} else {
		MX_DEBUG( 1,("%s: Moving '%s' to %g",
				fname, id_motor_record->name, id_energy));
	}

	status = mx_motor_move_absolute( id_motor_record, id_energy,
						NULL, flags );

	return status;
}

/******************************************************************/

/* The following function is used by the Bragg normal motor, the Bragg
 * translation motor, and the table motor below.
 */

static mx_status_type
mxd_monochromator_get_double_crystal_parameters(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					MX_RECORD **dependent_motor_record,
					double *tuning_parameter )
{
	const char fname[]
		= "mxd_monochromator_get_double_crystal_parameters()";

	MX_RECORD **record_array;
	int num_records;
	long num_elements;
	void *pointer_to_value;
	mx_status_type status;

	status = mxd_monochromator_get_record_array_from_list(
			list_record, &num_records, &record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_records != 2 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The record array for list record '%s' should have 2 records in it.  "
	"Instead there are %d records in it.",
			list_record->name, num_records );
	}

	/* The first record is the dependent motor record. */

	if ( (record_array[0]->mx_superclass != MXR_DEVICE)
	  || (record_array[0]->mx_class != MXC_MOTOR) )
	{
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The first record '%s' in the record array for list record '%s' "
	"is not a motor record.",
			record_array[0]->name, list_record->name );
	}

	*dependent_motor_record = record_array[0];

	/* The second record should be a variable or a motor whose value
	 * is the desired tuning parameter.
	 */

	switch( record_array[1]->mx_superclass ) {
	case MXR_VARIABLE:
		status = mx_get_variable_from_record(
					record_array[1], MXFT_DOUBLE,
					&num_elements, &pointer_to_value );

		if ( status.code != MXE_SUCCESS )
			return status;

		*tuning_parameter = *( double * ) pointer_to_value;

		break;
	case MXR_DEVICE:
		status = mx_motor_get_position(
				record_array[1], tuning_parameter );

		if ( status.code != MXE_SUCCESS )
			return status;

		break;
	default:
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"The record '%s' referenced by list record '%s' "
			"is not a variable record or a motor record.",
				record_array[1]->name,
				list_record->name );
		break;
	}
	return MX_SUCCESSFUL_RESULT;
}

/******************************************************************/

static mx_status_type
mxd_monochromator_get_bragg_normal_position( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double *theta )
{
	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_set_bragg_normal_position( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta )
{
	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_move_bragg_normal( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta,
					double old_theta,
					int flags )
{
	const char fname[] = "mxd_monochromator_move_bragg_normal()";

	MX_RECORD *normal_record;
	double beam_offset, new_bragg_normal, denominator;
	int enabled;
	mx_status_type status;

	status = mxd_monochromator_get_enable_status( list_record, &enabled );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( enabled == FALSE )
		return MX_SUCCESSFUL_RESULT;

	status = mxd_monochromator_get_double_crystal_parameters(
				monochromator, list_record,
				&normal_record, &beam_offset );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Compute the new monochromator crystal separation and move
	 * the Bragg normal motor there.
	 */

	denominator = 2.0 * cos( theta * MX_RADIANS_PER_DEGREE );

	new_bragg_normal = mx_divide_safely( beam_offset, denominator );

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		MX_DEBUG( 1,("%s: Checking '%s' limits for %g",
				fname, normal_record->name, new_bragg_normal));
	} else {
		MX_DEBUG( 1,("%s: Moving '%s' to %g",
				fname, normal_record->name, new_bragg_normal));
	}

	status = mx_motor_move_absolute( normal_record, new_bragg_normal,
						NULL, flags );

	return status;
}

/******************************************************************/

static mx_status_type
mxd_monochromator_get_bragg_translation_position(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double *theta )
{
	return MX_SUCCESSFUL_RESULT;
}
static mx_status_type
mxd_monochromator_set_bragg_translation_position(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta )
{
	return MX_SUCCESSFUL_RESULT;
}
static mx_status_type
mxd_monochromator_move_bragg_translation( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta,
					double old_theta,
					int flags )
{
	const char fname[] = "mxd_monochromator_move_bragg_translation()";

	MX_RECORD *translation_record;
	double beam_offset, new_bragg_translation, denominator;
	int enabled;
	mx_status_type status;

	status = mxd_monochromator_get_enable_status( list_record, &enabled );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( enabled == FALSE )
		return MX_SUCCESSFUL_RESULT;

	status = mxd_monochromator_get_double_crystal_parameters(
				monochromator, list_record,
				&translation_record, &beam_offset );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Compute the new monochromator crystal separation and move
	 * the Bragg normal motor there.
	 */

	denominator = 2.0 * sin( theta * MX_RADIANS_PER_DEGREE );

	new_bragg_translation = mx_divide_safely( beam_offset, denominator );

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		MX_DEBUG( 1,("%s: Checking '%s' limits for %g",
		    fname, translation_record->name, new_bragg_translation));
	} else {
		MX_DEBUG( 1,("%s: Moving '%s' to %g",
		    fname, translation_record->name, new_bragg_translation));
	}

	status = mx_motor_move_absolute( translation_record,
			new_bragg_translation, NULL, flags );

	return status;
}

/******************************************************************/

static mx_status_type
mxd_monochromator_get_table_position( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double *theta )
{
	return MX_SUCCESSFUL_RESULT;
}
static mx_status_type
mxd_monochromator_set_table_position( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta )
{
	return MX_SUCCESSFUL_RESULT;
}
static mx_status_type
mxd_monochromator_move_table( MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta,
					double old_theta,
					int flags )
{
	const char fname[] = "mxd_monochromator_move_table()";

	MX_RECORD *table_record;
	double crystal_separation, new_table_position;
	int enabled;
	mx_status_type status;

	status = mxd_monochromator_get_enable_status( list_record, &enabled );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( enabled == FALSE )
		return MX_SUCCESSFUL_RESULT;

	status = mxd_monochromator_get_double_crystal_parameters(
				monochromator, list_record,
				&table_record, &crystal_separation );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Compute the new table height and move the table there. */

	new_table_position = 2.0 * crystal_separation
				* cos( theta * MX_RADIANS_PER_DEGREE );

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		MX_DEBUG( 1,("%s: Checking '%s' limits for %g",
			fname, table_record->name, new_table_position));
	} else {
		MX_DEBUG( 1,("%s: Moving '%s' to %g",
			fname, table_record->name, new_table_position));
	}

	status = mx_motor_move_absolute( table_record,
			new_table_position, NULL, flags );

	return status;
}

/******************************************************************/

static mx_status_type
mxd_monochromator_get_diffractometer_theta_position(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double *theta )
{
	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_set_diffractometer_theta_position(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta )
{
	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxd_monochromator_move_diffractometer_theta(
					MX_MONOCHROMATOR *monochromator,
					MX_RECORD *list_record,
					double theta,
					double old_theta,
					int flags )
{
	const char fname[] = "mxd_monochromator_move_diffractometer_theta()";

	MX_RECORD **record_array;
	MX_RECORD *diffractometer_theta_record;
	double *parameter_array;
	double monochromator_d_spacing;
	double diffractometer_d_spacing;
	double diffractometer_scale, diffractometer_offset;
	double numerator, sin_theta, diffractometer_theta;
	void *pointer_to_value;
	int enabled, num_parameters, num_records;
	long num_elements;
	mx_status_type status;

	status = mxd_monochromator_get_enable_status( list_record, &enabled );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( enabled == FALSE )
		return MX_SUCCESSFUL_RESULT;

	status = mxd_monochromator_get_param_array_from_list(
			list_record, &num_parameters, &parameter_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_parameters != 2 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The parameter array for list record '%s' should have 2 parameters in it.  "
"Instead, there are %d parameters in it.", list_record->name, num_parameters );
	}

	diffractometer_scale = parameter_array[0];
	diffractometer_offset = parameter_array[1];

	status = mxd_monochromator_get_record_array_from_list(
			list_record, &num_records, &record_array );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( num_records != 3 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The record array for list record '%s' should have 3 records in it.  "
"Instead, there are %d records in it.", list_record->name, num_records );
	}

	/* The first record in the record array is the motor record
	 * for the diffractometer theta.
	 */

	diffractometer_theta_record = record_array[0];

	/* The second record is the monochromator d-spacing. */

	status = mx_get_variable_from_record( record_array[1], MXFT_DOUBLE,
					&num_elements, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	monochromator_d_spacing = *( double * ) pointer_to_value;

	/* The third record is the diffractometer d-spacing. */

	status = mx_get_variable_from_record( record_array[2], MXFT_DOUBLE,
					&num_elements, &pointer_to_value );

	if ( status.code != MXE_SUCCESS )
		return status;

	diffractometer_d_spacing = *( double * ) pointer_to_value;

	numerator = monochromator_d_spacing
			* sin( theta * MX_RADIANS_PER_DEGREE );

	sin_theta = mx_divide_safely( numerator, diffractometer_d_spacing );

	if ( fabs(sin_theta) > 1.0 ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"sin(diffractometer_theta) = %g", sin_theta );
	}

	diffractometer_theta = asin( sin_theta ) / MX_RADIANS_PER_DEGREE;

	diffractometer_theta = diffractometer_scale * diffractometer_theta
					+ diffractometer_offset;

	/* Now we may move the motor. */

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		MX_DEBUG( 1,("%s: Checking '%s' limits for %g",
				fname, diffractometer_theta_record->name,
				diffractometer_theta));
	} else {
		MX_DEBUG( 1,("%s: Moving '%s' to %g",
				fname, diffractometer_theta_record->name,
				diffractometer_theta));
	}

	status = mx_motor_move_absolute( diffractometer_theta_record,
					diffractometer_theta, NULL, flags );

	return status;
}

