/*
 * Name:    d_ntimer.c
 *
 * Purpose: MX timer driver to control MX network timers.
 *
 *------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_net.h"
#include "mx_timer.h"
#include "d_ntimer.h"

/* Initialize the timer driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_network_timer_record_function_list = {
	mxd_network_timer_initialize_type,
	mxd_network_timer_create_record_structures,
	mxd_network_timer_finish_record_initialization,
	mxd_network_timer_delete_record,
	NULL,
	mxd_network_timer_read_parms_from_hardware,
	mxd_network_timer_write_parms_to_hardware,
	mxd_network_timer_open,
	mxd_network_timer_close
};

MX_TIMER_FUNCTION_LIST mxd_network_timer_timer_function_list = {
	mxd_network_timer_is_busy,
	mxd_network_timer_start,
	mxd_network_timer_stop,
	mxd_network_timer_clear,
	mxd_network_timer_read,
	mxd_network_timer_get_mode,
	mxd_network_timer_set_mode
};

/* MX network timer data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_network_timer_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_TIMER_STANDARD_FIELDS,
	MXD_NETWORK_TIMER_STANDARD_FIELDS
};

long mxd_network_timer_num_record_fields
		= sizeof( mxd_network_timer_record_field_defaults )
		  / sizeof( mxd_network_timer_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_timer_rfield_def_ptr
			= &mxd_network_timer_record_field_defaults[0];

/*=======================================================================*/

static mx_status_type
mxd_network_timer_get_pointers( MX_TIMER *timer,
			MX_NETWORK_TIMER **network_timer,
			const char *calling_fname )
{
	const char fname[] = "mxd_network_timer_get_pointers()";

	if ( timer == (MX_TIMER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The timer pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( timer->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_RECORD pointer for timer pointer passed by '%s' is NULL.",
			calling_fname );
	}

	if ( timer->record->mx_type != MXT_TIM_NETWORK ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The timer '%s' passed by '%s' is not a network timer.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			timer->record->name, calling_fname,
			timer->record->mx_superclass,
			timer->record->mx_class,
			timer->record->mx_type );
	}

	if ( network_timer != (MX_NETWORK_TIMER **) NULL ) {

		*network_timer = (MX_NETWORK_TIMER *)
				(timer->record->record_type_struct);

		if ( *network_timer == (MX_NETWORK_TIMER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_NETWORK_TIMER pointer for timer record '%s' passed by '%s' is NULL",
				timer->record->name, calling_fname );
		}
	}

	if ( (*network_timer)->server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"The MX server record pointer for MX network timer '%s' is NULL.",
			timer->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mxd_network_timer_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_timer_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_network_timer_create_record_structures()";

	MX_TIMER *timer;
	MX_NETWORK_TIMER *network_timer;

	/* Allocate memory for the necessary structures. */

	timer = (MX_TIMER *) malloc( sizeof(MX_TIMER) );

	if ( timer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_TIMER structure." );
	}

	network_timer = (MX_NETWORK_TIMER *)
				malloc( sizeof(MX_NETWORK_TIMER) );

	if ( network_timer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_NETWORK_TIMER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = timer;
	record->record_type_struct = network_timer;
	record->class_specific_function_list
			= &mxd_network_timer_timer_function_list;

	timer->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_timer_finish_record_initialization( MX_RECORD *record )
{
	return mx_timer_finish_record_initialization( record );
}

MX_EXPORT mx_status_type
mxd_network_timer_delete_record( MX_RECORD *record )
{
	MX_NETWORK_TIMER *network_timer;

	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	network_timer = (MX_NETWORK_TIMER *) record->record_type_struct;

	if ( network_timer != NULL ) {

		free( network_timer );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_timer_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_timer_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_timer_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_timer_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_timer_is_busy( MX_TIMER *timer )
{
	const char fname[] = "mxd_network_timer_is_busy()";

	MX_NETWORK_TIMER *network_timer;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH ];
	int busy;
	mx_status_type status;

	status = mxd_network_timer_get_pointers(
				timer, &network_timer, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.busy", network_timer->remote_record_name );

	status = mx_get( network_timer->server_record,
				rfname, MXFT_INT, &busy );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_timer_start( MX_TIMER *timer )
{
	const char fname[] = "mxd_network_timer_start()";

	MX_NETWORK_TIMER *network_timer;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH ];
	double seconds_to_count;
	mx_status_type status;

	status = mxd_network_timer_get_pointers(
				timer, &network_timer, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	seconds_to_count = timer->value;

	sprintf( rfname, "%s.value", network_timer->remote_record_name );

	status = mx_put( network_timer->server_record,
				rfname, MXFT_DOUBLE, &seconds_to_count );

	return status;
}

MX_EXPORT mx_status_type
mxd_network_timer_stop( MX_TIMER *timer )
{
	const char fname[] = "mxd_network_timer_stop()";

	MX_NETWORK_TIMER *network_timer;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH ];
	int stop;
	double value;
	mx_status_type status;

	status = mxd_network_timer_get_pointers(
				timer, &network_timer, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	stop = 1;

	sprintf( rfname, "%s.stop", network_timer->remote_record_name );

	status = mx_put( network_timer->server_record,
				rfname, MXFT_INT, &stop );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.value", network_timer->remote_record_name );

	status = mx_get( network_timer->server_record,
				rfname, MXFT_DOUBLE, &value );

	timer->value = value;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_timer_clear( MX_TIMER *timer )
{
	const char fname[] = "mxd_network_timer_clear()";

	MX_NETWORK_TIMER *network_timer;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH ];
	int clear;
	mx_status_type status;

	status = mxd_network_timer_get_pointers(
				timer, &network_timer, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	clear = 1;

	sprintf( rfname, "%s.clear", network_timer->remote_record_name );

	status = mx_put( network_timer->server_record,
				rfname, MXFT_INT, &clear );

	timer->value = 0.0;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_timer_read( MX_TIMER *timer )
{
	const char fname[] = "mxd_network_timer_read()";

	MX_NETWORK_TIMER *network_timer;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH ];
	double value;
	mx_status_type status;

	status = mxd_network_timer_get_pointers(
				timer, &network_timer, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.value", network_timer->remote_record_name );

	status = mx_get( network_timer->server_record,
				rfname, MXFT_DOUBLE, &value );

	timer->value = value;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_timer_get_mode( MX_TIMER *timer )
{
	const char fname[] = "mxd_network_timer_get_mode()";

	MX_NETWORK_TIMER *network_timer;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH ];
	int mode;
	mx_status_type status;

	status = mxd_network_timer_get_pointers(
				timer, &network_timer, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	sprintf( rfname, "%s.mode", network_timer->remote_record_name );

	status = mx_get( network_timer->server_record,
				rfname, MXFT_INT, &mode );

	timer->mode = mode;

	return status;
}

MX_EXPORT mx_status_type
mxd_network_timer_set_mode( MX_TIMER *timer )
{
	const char fname[] = "mxd_network_timer_set_mode()";

	MX_NETWORK_TIMER *network_timer;
	char rfname[ MXU_RECORD_FIELD_NAME_LENGTH ];
	int mode;
	mx_status_type status;

	status = mxd_network_timer_get_pointers(
				timer, &network_timer, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	mode = timer->mode;

	sprintf( rfname, "%s.mode", network_timer->remote_record_name );

	status = mx_put( network_timer->server_record,
				rfname, MXFT_INT, &mode );

	return status;
}

