/*
 * Name:    d_pmactc.c 
 *
 * Purpose: MX motor driver for controlling a Delta Tau PMAC motor
 *          through Tom Coleman's EPICS PMAC dual-ported RAM interface.
 *
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_motor.h"
#include "mx_aout.h"
#include "d_pmactc.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_pmac_tc_motor_record_function_list = {
	mxd_pmac_tc_motor_initialize_type,
	mxd_pmac_tc_motor_create_record_structures,
	mxd_pmac_tc_motor_finish_record_initialization,
	mxd_pmac_tc_motor_delete_record,
	mxd_pmac_tc_motor_print_structure,
	mxd_pmac_tc_motor_read_parms_from_hardware,
	mxd_pmac_tc_motor_write_parms_to_hardware,
	mxd_pmac_tc_motor_open,
	mxd_pmac_tc_motor_close
};

MX_MOTOR_FUNCTION_LIST mxd_pmac_tc_motor_motor_function_list = {
	mxd_pmac_tc_motor_motor_is_busy,
	mxd_pmac_tc_motor_move_absolute,
	mxd_pmac_tc_motor_get_position,
	mxd_pmac_tc_motor_set_position,
	mxd_pmac_tc_motor_soft_abort,
	mxd_pmac_tc_motor_immediate_abort,
	mxd_pmac_tc_motor_positive_limit_hit,
	mxd_pmac_tc_motor_negative_limit_hit,
	mxd_pmac_tc_motor_find_home_position,
	mxd_pmac_tc_motor_constant_velocity_move,
	mxd_pmac_tc_motor_get_parameter,
	mxd_pmac_tc_motor_set_parameter
};

/* PMAC dual-ported RAM motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_pmac_tc_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_PMAC_TC_MOTOR_STANDARD_FIELDS
};

long mxd_pmac_tc_motor_num_record_fields
		= sizeof( mxd_pmac_tc_motor_record_field_defaults )
			/ sizeof( mxd_pmac_tc_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_pmac_tc_motor_rfield_def_ptr
			= &mxd_pmac_tc_motor_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_pmac_tc_motor_get_pointers( MX_MOTOR *motor,
			MX_PMAC_TC_MOTOR **pmac_tc_motor,
			const char *calling_fname )
{
	const char fname[] = "mxd_pmac_tc_motor_get_pointers()";

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( pmac_tc_motor == (MX_PMAC_TC_MOTOR **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_PMAC_TC_MOTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( motor->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
    "The MX_RECORD pointer for the MX_MOTOR pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*pmac_tc_motor = (MX_PMAC_TC_MOTOR *) motor->record->record_type_struct;

	if ( *pmac_tc_motor == (MX_PMAC_TC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_PMAC_TC_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_initialize_type( long type )
{
		return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_tc_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_PMAC_TC_MOTOR *pmac_tc_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	pmac_tc_motor = (MX_PMAC_TC_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( pmac_tc_motor == (MX_PMAC_TC_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_PMAC_TC_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = pmac_tc_motor;
	record->class_specific_function_list
				= &mxd_pmac_tc_motor_motor_function_list;

	motor->record = record;

	/* A DAC motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_finish_record_initialization( MX_RECORD *record )
{
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	return status;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_tc_motor_print_structure()";

	MX_MOTOR *motor;
	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	pmac_tc_motor = (MX_PMAC_TC_MOTOR *) (record->record_type_struct);

	if ( pmac_tc_motor == (MX_PMAC_TC_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_TC_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type                = PMAC_TC_MOTOR.\n\n");

	fprintf(file, "  name                      = %s\n", record->name);
	fprintf(file, "  actual position record    = %s\n",
			pmac_tc_motor->actual_position_record_name);
	fprintf(file, "  requested position record = %s\n",
			pmac_tc_motor->requested_position_record_name);
	fprintf(file, "  in position record        = %s\n",
			pmac_tc_motor->in_position_record_name);
	fprintf(file, "  abort record              = %s\n",
			pmac_tc_motor->abort_record_name);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf( file, "  position                  = %g %s  (%g).\n",
		motor->position, motor->units,
		motor->raw_position.analog );
	fprintf( file, "  scale                     = %g %s per EPICS EGU.\n",
		motor->scale, motor->units );
	fprintf( file, "  offset                    = %g %s.\n",
		motor->offset, motor->units);
	fprintf( file, "  backlash                  = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog );
	fprintf( file, "  negative limit            = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog );
	fprintf( file, "  positive limit            = %g %s  (%g).\n",
	        motor->positive_limit, motor->units,
	        motor->raw_positive_limit.analog );

	deadband = motor->scale * motor->deadband.analog;

	fprintf( file, "  deadband                  = %g %s  (%g).\n\n",
		deadband, motor->units,
		motor->deadband.analog );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_open( MX_RECORD *record )
{
	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_motor_is_busy()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	short in_position_value;
	int i, max_retries;
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	motor->busy = FALSE;

	strcpy( pvname, pmac_tc_motor->in_position_record_name );

	max_retries = 9;

	for ( i = 0; i < max_retries; i++ ) {
		ezca_status = ezcaGet( pvname,
					ezcaShort, 1, &in_position_value );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Attempt to read motion status of PMAC TC motor '%s' via record '%s' failed. "
" EZCA status = %d (%s)",
				motor->record->name,
				pvname, ezca_status, ezca_error_message );

			ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR,
				fname, mx_error_message);
		}

		switch ( in_position_value ) {
		case 0:
			motor->busy = TRUE;

			return MX_SUCCESSFUL_RESULT;

			break;
		case 1:
			/* Go back and check again. */

			break;
		default:
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"In position record '%s' for PMAC TC motor '%s' "
			"had an unexpected value of %d",
			pvname, motor->record->name, in_position_value );

			break;
		}
	}

	motor->busy = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_move_absolute()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	double new_position;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	new_position = motor->raw_setpoint.analog;

	/* Send the move command. */

	strcpy(pvname, pmac_tc_motor->requested_position_record_name);

	ezca_status = ezcaPut( pvname, ezcaDouble, 1, &new_position );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
			"Attempt to move PMAC TC motor '%s' via '%s' failed.  "
			"EZCA status = %d (%s)",
			motor->record->name,
			pmac_tc_motor->requested_position_record_name,
			ezca_status, ezca_error_message );

		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_get_position()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	double scale, offset, present_value;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	scale = motor->scale;
	offset = motor->offset;

	strcpy(pvname, pmac_tc_motor->actual_position_record_name);

	ezca_status = ezcaGet( pvname, ezcaDouble, 1, &present_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to get position of PMAC TC motor '%s' via '%s' failed.  "
	"EZCA status = %d (%s)",
			motor->record->name,
			pmac_tc_motor->actual_position_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	motor->raw_position.analog = present_value;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_set_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"'set position' is not supported for an PMAC TC motor." );
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_soft_abort()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short stop_field;
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	strcpy( pvname, pmac_tc_motor->abort_record_name );

	/* Stop the motor move. */

	stop_field = 1;		/* Stop */

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &stop_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to abort move for PMAC TC motor '%s' via '%s' failed.  "
	"EZCA status = %d (%s)",
			motor->record->name,
			pmac_tc_motor->abort_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_immediate_abort( MX_MOTOR *motor )
{
	return mxd_pmac_tc_motor_soft_abort( motor );
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_positive_limit_hit( MX_MOTOR *motor )
{
	motor->positive_limit_hit = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_negative_limit_hit( MX_MOTOR *motor )
{
	motor->negative_limit_hit = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_find_home_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
"'find home position' is not supported for an PMAC TC motor." );
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_constant_velocity_move( MX_MOTOR *motor )
{
	mx_status_type mx_status;

	if ( motor->constant_velocity_move >= 0 ) {
		motor->raw_setpoint.analog = 10000;
	} else {
		motor->raw_setpoint.analog = -10000;
	}

	mx_status = mxd_pmac_tc_motor_move_absolute( motor );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_get_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_get_parameter()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	char strcmd_pvname[60], strrsp_pvname[60];
	char command[41], response[41];
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status, num_items;
	unsigned long pmac_speed;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Send the parameter request command to the StrCmd EPICS variable. */

	sprintf( strcmd_pvname, "%s:StrCmd.VAL", pmac_tc_motor->pmac_name );

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( command, "I%d16", pmac_tc_motor->motor_number );
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Parameter type %d is not supported by the driver for motor '%s'",
			motor->parameter_type );
		break;
	}

	MX_DEBUG(-2,("%s: sending command '%s' to '%s'",
		fname, command, strcmd_pvname));

	ezca_status = ezcaPut( strcmd_pvname, ezcaString, 1, command );

	MX_DEBUG(-2,("%s: EZCA status = %d", fname, ezca_status));

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to set the value of EPICS variable '%s' for motor '%s' failed.  "
"EZCA status = %d (%s)",
			strcmd_pvname, motor->record->name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Now read back the result from the StrRsp EPICS variable. */

	sprintf( strrsp_pvname, "%s:StrRsp.VAL", pmac_tc_motor->pmac_name );

	ezca_status = ezcaGet( strrsp_pvname, ezcaString, 1, response );

	MX_DEBUG(-2,("%s: EZCA status = %d", fname, ezca_status));

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to get the value of EPICS variable '%s' for motor '%s' failed.  "
"EZCA status = %d (%s)",
			strrsp_pvname, motor->record->name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	MX_DEBUG(-2,("%s: received response '%s' from '%s'.",
		fname, response, strrsp_pvname));

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		num_items = sscanf( response, "%lu", &pmac_speed );

		if ( num_items != 1 ) {
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"Unable to parse PMAC response '%s' for motor '%s'.",
				response, motor->record->name );
		}

		motor->raw_speed = (double) pmac_speed;
		break;
	}

	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_tc_motor_set_parameter( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_tc_motor_set_parameter()";

	MX_PMAC_TC_MOTOR *pmac_tc_motor;
	char strcmd_pvname[60];
	char command[41];
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;
	mx_status_type mx_status;

	mx_status = mxd_pmac_tc_motor_get_pointers( motor,
						&pmac_tc_motor, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Send the parameter request set to the StrCmd EPICS variable. */

	sprintf( strcmd_pvname, "%s:StrCmd.VAL", pmac_tc_motor->pmac_name );

	switch( motor->parameter_type ) {
	case MXLV_MTR_SPEED:
		sprintf( command, "I%d16=%ld", pmac_tc_motor->motor_number,
						mx_round( motor->raw_speed ) );
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Parameter type %d is not supported by the driver for motor '%s'",
			motor->parameter_type );
		break;
	}

	MX_DEBUG(-2,("%s: sending command '%s' to '%s'",
		fname, command, strcmd_pvname));

	ezca_status = ezcaPut( strcmd_pvname, ezcaString, 1, command );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to set the value of EPICS variable '%s' for motor '%s' failed.  "
"EZCA status = %d (%s)",
			strcmd_pvname, motor->record->name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	MX_DEBUG(-2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

