/*
 * Name:    d_wavnum.c 
 *
 * Purpose: MX motor driver to control an MX motor in units of 
 *          X-ray wavenumber.
 *
 *          This is generally used to control an X-ray monochromator.
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_var.h"
#include "mx_motor.h"
#include "mx_aout.h"
#include "d_wavnum.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_wavenum_motor_record_function_list = {
	mxd_wavenum_motor_initialize_type,
	mxd_wavenum_motor_create_record_structures,
	mxd_wavenum_motor_finish_record_initialization,
	mxd_wavenum_motor_delete_record,
	mxd_wavenum_motor_print_structure,
	mxd_wavenum_motor_read_parms_from_hardware,
	mxd_wavenum_motor_write_parms_to_hardware,
	mxd_wavenum_motor_open,
	mxd_wavenum_motor_close
};

MX_MOTOR_FUNCTION_LIST mxd_wavenum_motor_motor_function_list = {
	mxd_wavenum_motor_motor_is_busy,
	mxd_wavenum_motor_move_absolute,
	mxd_wavenum_motor_get_position,
	mxd_wavenum_motor_set_position,
	mxd_wavenum_motor_soft_abort,
	mxd_wavenum_motor_immediate_abort,
	mxd_wavenum_motor_positive_limit_hit,
	mxd_wavenum_motor_negative_limit_hit,
	mxd_wavenum_motor_find_home_position
};

/* Photon wavenumber motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_wavenum_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_WAVENUM_MOTOR_STANDARD_FIELDS
};

long mxd_wavenum_motor_num_record_fields
		= sizeof( mxd_wavenum_motor_record_field_defaults )
			/ sizeof( mxd_wavenum_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_wavenum_motor_rfield_def_ptr
			= &mxd_wavenum_motor_record_field_defaults[0];

/* === */

MX_EXPORT mx_status_type
mxd_wavenum_motor_initialize_type( long type )
{
		/* Nothing needed here. */

		return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_wavenum_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_WAVENUMBER_MOTOR *wavenum_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	wavenum_motor = (MX_WAVENUMBER_MOTOR *)
				malloc( sizeof(MX_WAVENUMBER_MOTOR) );

	if ( wavenum_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_WAVENUMBER_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = wavenum_motor;
	record->class_specific_function_list
				= &mxd_wavenum_motor_motor_function_list;

	motor->record = record;

	/* A photon wavenumber motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_finish_record_initialization( MX_RECORD *record )
{
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	return status;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_wavenum_motor_print_structure()";

	MX_MOTOR *motor;
	MX_RECORD *dependent_motor_record;
	MX_MOTOR *dependent_motor;
	MX_WAVENUMBER_MOTOR *wavenumber_motor;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	wavenumber_motor = (MX_WAVENUMBER_MOTOR *)
					(record->record_type_struct);

	if ( wavenumber_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_WAVENUMBER_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	dependent_motor_record = wavenumber_motor->dependent_motor_record;

	if ( dependent_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			record->name );
	}

	dependent_motor = (MX_MOTOR *)
				(dependent_motor_record->record_class_struct);

	if ( dependent_motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			dependent_motor_record->name );
	}

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type          = WAVENUMBER_MOTOR.\n\n");

	fprintf(file, "  name                = %s\n", record->name);
	fprintf(file, "  dependent motor     = %s\n",
					dependent_motor_record->name);
	fprintf(file, "  d spacing           = %s angstroms\n",
				wavenumber_motor->d_spacing_record->name);
	fprintf(file, "  angle scale         = %g radians per %s.\n",
					wavenumber_motor->angle_scale,
					dependent_motor->units);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf(file, "  position            = %g %s  (%g)\n",
		motor->position, motor->units,
		motor->raw_position.analog );
	fprintf(file, "  wavenumber scale    = %g %s per unscaled "
						"wavenumber unit.\n",
		motor->scale, motor->units);
	fprintf(file, "  wavenumber offset   = %g %s.\n",
		motor->offset, motor->units);
	fprintf(file, "  backlash            = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog);
	fprintf(file, "  negative limit      = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog);
	fprintf(file, "  positive limit      = %g %s  (%g).\n",
		motor->positive_limit, motor->units,
		motor->raw_positive_limit.analog);

	deadband = motor->scale * motor->deadband.analog;

	fprintf(file, "  deadband            = %g %s  (%g).\n\n",
		deadband, motor->units,
		motor->deadband.analog );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_read_parms_from_hardware( MX_RECORD *record )
{
	/* All saving of parameters is handled by the MX_MOTOR record
	 * this motor is derived from, so we need not do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_write_parms_to_hardware( MX_RECORD *record )
{
	/* All setting of parameters is handled by the MX_MOTOR record
	 * this motor is derived from, so we need not do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_open( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_close( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_wavenum_motor_motor_is_busy()";

	MX_WAVENUMBER_MOTOR *wavenumber_motor;
	MX_RECORD *dependent_motor_record;
	int busy;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	wavenumber_motor = (MX_WAVENUMBER_MOTOR *)
					(motor->record->record_type_struct);

	if ( wavenumber_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_WAVENUMBER_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	dependent_motor_record = wavenumber_motor->dependent_motor_record;

	if ( dependent_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_is_busy( dependent_motor_record, &busy );

	motor->busy = busy;

	return status;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_wavenum_motor_move_absolute()";

	MX_WAVENUMBER_MOTOR *wavenumber_motor;
	MX_RECORD *dependent_motor_record;
	double wavenumber, d_spacing, minimum_wavenumber;
	double angle_scale, angle_in_radians, unscaled_angle, sin_theta;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	wavenumber_motor = (MX_WAVENUMBER_MOTOR *)
					(motor->record->record_type_struct);

	if ( wavenumber_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_WAVENUMBER_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	dependent_motor_record = wavenumber_motor->dependent_motor_record;

	if ( dependent_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	/* The unscaled wavenumber is in inverse angstroms. */

	wavenumber = motor->raw_setpoint.analog;

	/* This device is different in that it has an extra scale factor
	 * for the angle.  The angle scale does whatever is necessary 
	 * to convert the angle in the dependent motor record into radians.
	 * I can't conceive of a use for an angle offset here, so I haven't
	 * included it.  If you can think of one, tell me.
	 */

	angle_scale = wavenumber_motor->angle_scale;

	status = mx_get_double_variable( motor->record->list_head,
				wavenumber_motor->d_spacing_record->name,
				&d_spacing );

	if ( status.code != MXE_SUCCESS )
		return status;

	minimum_wavenumber = MX_PI / d_spacing;

	if ( wavenumber < 1.0e-9 ) {  /* 1 nano-eV */
		sin_theta = DBL_MAX;
	} else {
		sin_theta = MX_PI / ( d_spacing * wavenumber );
	}

	if ( sin_theta <= 1.0 ) {
		angle_in_radians = asin( sin_theta );
	} else {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"Attempted to move '%s' to an unreachable wavenumber of %g 1/angstroms.  "
"Minimum allowed wavenumber for d spacing = %g angstroms is %g 1/angstroms.",
			motor->record->name, wavenumber, 
			d_spacing, minimum_wavenumber );
	}

	if ( fabs(angle_scale) < MX_MOTOR_ANALOG_FUZZ ) {
		unscaled_angle = DBL_MAX;
	} else {
		unscaled_angle = angle_in_radians / angle_scale;
	}

	status = mx_motor_move_absolute( dependent_motor_record,
						unscaled_angle, NULL, 0 );

	return status;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_wavenum_motor_get_position()";

	MX_WAVENUMBER_MOTOR *wavenumber_motor;
	MX_RECORD *dependent_motor_record;
	double unscaled_angle, angle_in_radians, angle_scale, sin_theta;
	double wavenumber, d_spacing, denominator;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	wavenumber_motor = (MX_WAVENUMBER_MOTOR *)
					(motor->record->record_type_struct);

	if ( wavenumber_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_WAVENUMBER_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	dependent_motor_record = wavenumber_motor->dependent_motor_record;

	if ( dependent_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	angle_scale = wavenumber_motor->angle_scale;

	status = mx_get_double_variable( motor->record->list_head,
				wavenumber_motor->d_spacing_record->name,
				&d_spacing );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_motor_get_position(
				dependent_motor_record, &unscaled_angle );

	if ( status.code != MXE_SUCCESS )
		return status;

	angle_in_radians = angle_scale * unscaled_angle;

	sin_theta = sin( angle_in_radians );

	denominator = d_spacing * sin_theta;

	if ( fabs(denominator) >= 1.0 ) {
		wavenumber = MX_PI / denominator;
	} else {
		if ( MX_PI < fabs(denominator * DBL_MAX) ) {
			wavenumber = MX_PI / denominator;
		} else {
			wavenumber = DBL_MAX;
		}
	}

	motor->raw_position.analog = wavenumber;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_wavenum_motor_set_position()";

	MX_WAVENUMBER_MOTOR *wavenumber_motor;
	MX_RECORD *dependent_motor_record;
	double unscaled_angle, angle_in_radians, angle_scale, sin_theta;
	double d_spacing, wavenumber, minimum_wavenumber;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed NULL." );
	}

	wavenumber_motor = (MX_WAVENUMBER_MOTOR *)
					(motor->record->record_type_struct);

	if ( wavenumber_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_WAVENUMBER_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	dependent_motor_record = wavenumber_motor->dependent_motor_record;

	if ( dependent_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	wavenumber = motor->raw_setpoint.analog;

	angle_scale = wavenumber_motor->angle_scale;

	status = mx_get_double_variable( motor->record->list_head,
				wavenumber_motor->d_spacing_record->name,
				&d_spacing );

	if ( status.code != MXE_SUCCESS )
		return status;

	minimum_wavenumber = MX_PI / d_spacing;

	if ( wavenumber < 1.0e-9 ) {  /* 1 nano-eV */
		sin_theta = DBL_MAX;
	} else {
		sin_theta = MX_PI / ( d_spacing * wavenumber );
	}

	if ( sin_theta <= 1.0 ) {
		angle_in_radians = asin( sin_theta );
	} else {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"Attempted to set '%s' to an unreachable wavenumber of %g 1/angstroms.  "
"Minimum allowed wavenumber for d spacing = %g angstroms is %g 1/angstroms.",
			motor->record->name, wavenumber,
			d_spacing, minimum_wavenumber );
	}

	if ( fabs(angle_scale) < MX_MOTOR_ANALOG_FUZZ ) {
		unscaled_angle = DBL_MAX;
	} else {
		unscaled_angle = angle_in_radians / angle_scale;
	}

	status = mx_motor_set_position(
				dependent_motor_record, unscaled_angle );

	return status;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_wavenum_motor_soft_abort()";

	MX_WAVENUMBER_MOTOR *wavenumber_motor;
	MX_RECORD *dependent_motor_record;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	wavenumber_motor = (MX_WAVENUMBER_MOTOR *)
					(motor->record->record_type_struct);

	if ( wavenumber_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_WAVENUMBER_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	dependent_motor_record = wavenumber_motor->dependent_motor_record;

	if ( dependent_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_soft_abort( dependent_motor_record );

	return status;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_wavenum_motor_immediate_abort()";

	MX_WAVENUMBER_MOTOR *wavenumber_motor;
	MX_RECORD *dependent_motor_record;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	wavenumber_motor = (MX_WAVENUMBER_MOTOR *)
					(motor->record->record_type_struct);

	if ( wavenumber_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_WAVENUMBER_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	dependent_motor_record = wavenumber_motor->dependent_motor_record;

	if ( dependent_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	status = mx_motor_immediate_abort( dependent_motor_record );

	return status;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_wavenum_motor_positive_limit_hit()";

	MX_WAVENUMBER_MOTOR *wavenumber_motor;
	MX_RECORD *dependent_motor_record;
	int angle_limit_hit;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	wavenumber_motor = (MX_WAVENUMBER_MOTOR *)
					(motor->record->record_type_struct);

	if ( wavenumber_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_WAVENUMBER_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	dependent_motor_record = wavenumber_motor->dependent_motor_record;

	if ( dependent_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	/* Note that the reversal in limit sense is _not_ a mistake. */

	status = mx_motor_negative_limit_hit(
				dependent_motor_record, &angle_limit_hit );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( angle_limit_hit ) {
		motor->positive_limit_hit = TRUE;
	} else {
		motor->positive_limit_hit = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_wavenum_motor_negative_limit_hit()";

	MX_WAVENUMBER_MOTOR *wavenumber_motor;
	MX_RECORD *dependent_motor_record;
	int angle_limit_hit;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	wavenumber_motor = (MX_WAVENUMBER_MOTOR *)
					(motor->record->record_type_struct);

	if ( wavenumber_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_WAVENUMBER_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	dependent_motor_record = wavenumber_motor->dependent_motor_record;

	if ( dependent_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	/* Note that the reversal in limit sense is _not_ a mistake. */

	status = mx_motor_positive_limit_hit(
				dependent_motor_record, &angle_limit_hit );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( angle_limit_hit ) {
		motor->negative_limit_hit = TRUE;
	} else {
		motor->negative_limit_hit = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_wavenum_motor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_wavenum_motor_find_home_position()";

	MX_WAVENUMBER_MOTOR *wavenumber_motor;
	MX_RECORD *dependent_motor_record;
	int direction;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	wavenumber_motor = (MX_WAVENUMBER_MOTOR *)
					(motor->record->record_type_struct);

	if ( wavenumber_motor == (MX_WAVENUMBER_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_WAVENUMBER_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	dependent_motor_record = wavenumber_motor->dependent_motor_record;

	if ( dependent_motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Dependent motor record pointer for record '%s' is NULL.",
			motor->record->name );
	}

	direction = motor->home_search;

	status = mx_motor_find_home_position( dependent_motor_record,
							direction );

	return status;
}

