/*
 * Name:    i_cmpmtr.c
 *
 * Purpose: MX driver for Compumotor 6000 series motor controllers.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_alloc.h"
#include "mx_gener.h"
#include "mx_rs232.h"
#include "i_cmpmtr.h"

MX_RECORD_FUNCTION_LIST mxi_compumotor_record_function_list = {
	mxi_compumotor_initialize_type,
	mxi_compumotor_create_record_structures,
	mxi_compumotor_finish_record_initialization,
	mxi_compumotor_delete_record,
	NULL,
	mxi_compumotor_read_parms_from_hardware,
	mxi_compumotor_write_parms_to_hardware,
	mxi_compumotor_open,
	mxi_compumotor_close,
	NULL,
	mxi_compumotor_resynchronize,
};

MX_GENERIC_FUNCTION_LIST mxi_compumotor_generic_function_list = {
	mxi_compumotor_getchar,
	mxi_compumotor_putchar,
	mxi_compumotor_read,
	mxi_compumotor_write,
	mxi_compumotor_input_is_available,
	mxi_compumotor_discard_unread_input,
	mxi_compumotor_discard_unwritten_output
};

MX_RECORD_FIELD_DEFAULTS mxi_compumotor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_COMPUMOTOR_INTERFACE_STANDARD_FIELDS
};

long mxi_compumotor_num_record_fields
		= sizeof( mxi_compumotor_record_field_defaults )
			/ sizeof( mxi_compumotor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_compumotor_rfield_def_ptr
			= &mxi_compumotor_record_field_defaults[0];

#define MXI_COMPUMOTOR_INTERFACE_DEBUG		FALSE

/* A private function for the use of the driver. */

static mx_status_type
mxi_compumotor_get_pointers( MX_GENERIC *generic,
			MX_COMPUMOTOR_INTERFACE **compumotor_interface,
			const char *calling_fname )
{
	const char fname[] = "mxi_compumotor_get_pointers()";

	MX_RECORD *compumotor_interface_record;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_GENERIC pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_COMPUMOTOR_INTERFACE pointer passed by '%s' was NULL.",
			calling_fname );
	}

	compumotor_interface_record = generic->record;

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The compumotor_interface_record pointer for the "
			"MX_GENERIC pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			compumotor_interface_record->record_type_struct;

	if ( *compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_COMPUMOTOR_INTERFACE pointer for record '%s' is NULL.",
			compumotor_interface_record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*==========================*/

MX_EXPORT mx_status_type
mxi_compumotor_initialize_type( long type )
{
	const char fname[] = "mxi_compumotor_initialize_type()";

	MX_DRIVER *driver;
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
	MX_RECORD_FIELD_DEFAULTS *field;
	long num_record_fields;
	long referenced_field_index;
	long num_controllers_varargs_cookie;
	mx_status_type status;

	driver = mx_get_driver_by_type( mx_type_list, type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record type %ld not found in driver list.", type );
	}

	record_field_defaults_ptr
			= driver->record_field_defaults_ptr;

	if ( record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	record_field_defaults = *record_field_defaults_ptr;

	if ( record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'*record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	if ( driver->num_record_fields == (long *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'num_record_fields' pointer for record type '%s' is NULL.",
			driver->name );
	}

	num_record_fields = *(driver->num_record_fields);

	status = mx_find_record_field_defaults_index(
			record_field_defaults, num_record_fields,
			"num_controllers", &referenced_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie( referenced_field_index,
				0, &num_controllers_varargs_cookie);

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_find_record_field_defaults(
			record_field_defaults, num_record_fields,
			"num_axes", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = num_controllers_varargs_cookie;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_compumotor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_compumotor_create_record_structures()";

	MX_GENERIC *generic;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;

	/* Allocate memory for the necessary structures. */

	generic = (MX_GENERIC *) malloc( sizeof(MX_GENERIC) );

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_GENERIC structure." );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
				malloc( sizeof(MX_COMPUMOTOR_INTERFACE) );

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_COMPUMOTOR_INTERFACE structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = generic;
	record->record_type_struct = compumotor_interface;
	record->class_specific_function_list
				= &mxi_compumotor_generic_function_list;

	generic->record = record;
	compumotor_interface->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_compumotor_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxi_compumotor_finish_record_initialization()";

	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	long dimension_array[2];
	size_t sizeof_array[2];
	int i, j;
	long num_controllers;

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
					record->record_type_struct;

	num_controllers = compumotor_interface->num_controllers;

	/* Allocate the motor_array data structure. */

	dimension_array[0] = num_controllers;
	dimension_array[1] = MX_MAX_COMPUMOTOR_AXES;

	sizeof_array[0] = sizeof(MX_RECORD *);
	sizeof_array[1] = sizeof(MX_RECORD **);

	compumotor_interface->motor_array
		= ( MX_RECORD *(*)[MX_MAX_COMPUMOTOR_AXES] )
			malloc( num_controllers * MX_MAX_COMPUMOTOR_AXES
			* sizeof( MX_RECORD * ) );

	if ( compumotor_interface->motor_array == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocate motor_array for Compumotor interface '%s'.",
			record->name );
	}

	/* Initialize the array. */

	for ( i = 0; i < compumotor_interface->num_controllers; i++ ) {
		for ( j = 0; j < MX_MAX_COMPUMOTOR_AXES; j++ ) {
			compumotor_interface->motor_array[i][j] = NULL;
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_compumotor_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_compumotor_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_compumotor_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_compumotor_open( MX_RECORD *record )
{
	const char fname[] = "mxi_compumotor_open()";

	MX_GENERIC *generic;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	generic = (MX_GENERIC *) record->record_class_struct;

	status = mxi_compumotor_get_pointers( generic,
				&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Throw away any pending input and output on the 
	 * Compumotor RS-232 port.
	 */

	status = mx_rs232_discard_unwritten_output(
					compumotor_interface->rs232_record,
					MXI_COMPUMOTOR_INTERFACE_DEBUG );

	switch( status.code ) {
	case MXE_SUCCESS:
	case MXE_UNSUPPORTED:
		break;		/* Continue on. */
	default:
		return status;
		break;
	}

	status = mx_rs232_discard_unread_input(
					compumotor_interface->rs232_record,
					MXI_COMPUMOTOR_INTERFACE_DEBUG);

	return status;
}

MX_EXPORT mx_status_type
mxi_compumotor_close( MX_RECORD *record )
{
	const char fname[] = "mxi_compumotor_close()";

	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
					(record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_COMPUMOTOR_INTERFACE pointer for record '%s' is NULL.",
		record->name);
	}

	/* Get rid of any remaining characters in the input buffer and
	 * do it quietly.
	 */

	status = mx_rs232_discard_unread_input(
				compumotor_interface->rs232_record, FALSE );

	return status;
}

MX_EXPORT mx_status_type
mxi_compumotor_resynchronize( MX_RECORD *record )
{
	mx_status_type status;

	status = mxi_compumotor_close( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxi_compumotor_open( record );

	return status;
}

MX_EXPORT mx_status_type
mxi_compumotor_getchar( MX_GENERIC *generic, char *c, int flags )
{
	const char fname[] = "mxi_compumotor_getchar()";

	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	mx_status_type status;

	status = mxi_compumotor_get_pointers( generic,
				&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_rs232_getchar( compumotor_interface->rs232_record,
					c, flags );

	return status;
}

MX_EXPORT mx_status_type
mxi_compumotor_putchar( MX_GENERIC *generic, char c, int flags )
{
	const char fname[] = "mxi_compumotor_putchar()";

	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	mx_status_type status;

	status = mxi_compumotor_get_pointers( generic,
				&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_rs232_putchar( compumotor_interface->rs232_record,
					c, flags );

	return status;
}

MX_EXPORT mx_status_type
mxi_compumotor_read( MX_GENERIC *generic, void *buffer, size_t count )
{
	const char fname[] = "mxi_compumotor_read()";

	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	mx_status_type status;

	status = mxi_compumotor_get_pointers( generic,
				&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_rs232_read( compumotor_interface->rs232_record,
				buffer, count, NULL, 0 );

	return status;
}

MX_EXPORT mx_status_type
mxi_compumotor_write( MX_GENERIC *generic, void *buffer, size_t count )
{
	const char fname[] = "mxi_compumotor_write()";

	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	mx_status_type status;

	status = mxi_compumotor_get_pointers( generic,
				&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_rs232_write( compumotor_interface->rs232_record,
					buffer, count, NULL, 0 );

	return status;
}

MX_EXPORT mx_status_type
mxi_compumotor_input_is_available( MX_GENERIC *generic, int *input_available )
{
	const char fname[] = "mxi_compumotor_input_is_available()";

	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	mx_status_type status;

	status = mxi_compumotor_get_pointers( generic,
				&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_rs232_input_is_available(
			compumotor_interface->rs232_record, input_available );

	return status;
}

MX_EXPORT mx_status_type
mxi_compumotor_discard_unread_input( MX_GENERIC *generic, int debug_flag )
{
	const char fname[] = "mxi_compumotor_discard_unread_input()";

	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	mx_status_type status;

	status = mxi_compumotor_get_pointers( generic,
				&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_rs232_discard_unread_input(
					compumotor_interface->rs232_record,
					debug_flag );

	return status;
}

MX_EXPORT mx_status_type
mxi_compumotor_discard_unwritten_output( MX_GENERIC *generic, int debug_flag )
{
	const char fname[] = "mxi_compumotor_discard_unwritten_output()";

	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	mx_status_type status;

	status = mxi_compumotor_get_pointers( generic,
				&compumotor_interface, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_rs232_discard_unwritten_output(
					compumotor_interface->rs232_record,
					debug_flag );

	return status;
}

/* === Functions specific to this driver. === */

MX_EXPORT mx_status_type
mxi_compumotor_command( MX_COMPUMOTOR_INTERFACE *compumotor_interface,
		char *command, char *response, int response_buffer_length,
		int debug_flag )
{
	const char fname[] = "mxi_compumotor_command()";

	unsigned long sleep_ms;
	int i, max_attempts;
	mx_status_type status;

	MX_DEBUG(2,("%s invoked.", fname));

	if ( compumotor_interface == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_COMPUMOTOR_INTERFACE pointer passed was NULL." );
	}

	if ( command == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"'command' buffer pointer passed was NULL.  No command sent.");
	}

	/* Send the command string. */

	status = mxi_compumotor_putline(
				compumotor_interface, command, debug_flag );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the response, if one is expected. */

	i = 0;

	max_attempts = 1;
	sleep_ms = 0;

	if ( response != NULL ) {
		for ( i=0; i < max_attempts; i++ ) {
			if ( i > 0 ) {
				MX_DEBUG(-2,
			("%s: No response yet to '%s' command.  Retrying...",
				compumotor_interface->record->name, command ));
			}

			status = mxi_compumotor_getline(
					compumotor_interface, response,
					response_buffer_length, debug_flag );

			if ( status.code == MXE_SUCCESS ) {
				break;		/* Exit the for() loop. */

			} else if ( status.code != MXE_NOT_READY ) {
				MX_DEBUG(-2,
				("*** Exiting with status = %ld",status.code));
				return status;
			}
			mx_msleep(sleep_ms);
		}

		if ( i >= max_attempts ) {
			status = mx_rs232_discard_unread_input(
					compumotor_interface->rs232_record,
					debug_flag );

			if ( status.code != MXE_SUCCESS ) {
				mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Failed at attempt to discard unread characters in buffer for record '%s'",
					compumotor_interface->record->name );
			}

			return mx_error( MXE_INTERFACE_IO_ERROR, fname,
				"No response seen to '%s' command", command );
		}
	}

	MX_DEBUG(2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_compumotor_getline( MX_COMPUMOTOR_INTERFACE *compumotor_interface,
		char *buffer, long buffer_length, int debug_flag )
{
	const char fname[] = "mxi_compumotor_getline()";

	char c;
	int i;
	mx_status_type status;

	if ( debug_flag ) {
		MX_DEBUG(-2, ("%s invoked.", fname));
	}

	/* Any text sent by the Compumotor controller should be prefixed
	 * by an asterix character, so to begin with we try to read and
	 * discard characters until we see an asterix.
	 */

	c = '\0';

	while ( c != '*' ) {
		status = mx_rs232_getchar( compumotor_interface->rs232_record,
						&c, MXF_232_WAIT );

		if ( status.code != MXE_SUCCESS ) {
			buffer[0] = '\0';

			if ( debug_flag ) {
				MX_DEBUG(-2,
			("%s failed while waiting for an asterix character.",
					fname));
			}
			return status;
		}
	}

	for ( i = 0; i < (buffer_length - 1) ; i++ ) {
		status = mx_rs232_getchar( compumotor_interface->rs232_record,
						&c, MXF_232_WAIT );

		if ( status.code != MXE_SUCCESS ) {
			/* Make the buffer contents a valid C string
			 * before returning, so that we can at least
			 * see what appeared before the error.
			 */

			buffer[i] = '\0';

			if ( debug_flag ) {
				MX_DEBUG(-2, ("%s failed.\nbuffer = '%s'",
					fname, buffer));

				MX_DEBUG(-2,
				("Failed with status = %d, c = 0x%x '%c'",
				status, c, c));
			}

			return status;
		}

		if ( debug_flag ) {
			MX_DEBUG(-2,("%s: received c = 0x%x '%c'",
						fname, c, c));
		}

		buffer[i] = c;

		/* Responses from a Compumotor controller should always
		 * be terminated by CR, LF sequence, even if this is
		 * the last line in the response.  This means putting
		 * the command 'EOT13,10,0' in the controller's
		 * STARTP script.
		 */

		if ( c == MX_CR ) {
			i++;

			/* There is also a line feed character to throw away.*/

			status = mx_rs232_getchar(
					compumotor_interface->rs232_record,
					&c, MXF_232_WAIT );

			if ( status.code != MXE_SUCCESS ) {
				if ( debug_flag ) {
					MX_DEBUG(-2,
("%s: failed to see any character after a CR character.", fname));
				}

				return status;
			}

			if ( c != MX_LF ) {
				if ( debug_flag ) {
					MX_DEBUG(-2,
("%s: The character after a CR was not an LF.  Instead saw 0%o", fname, c));
				}
			}

			if ( debug_flag ) {
				MX_DEBUG(-2,("%s: received c = 0x%x '%c'",
						fname, c, c));
			}

			break;		/* exit the for() loop */
		}
	}

	/* Check to see if the last character was an end of string
	 * and if it is then overwrite it with a NULL.
	 */

	if ( buffer[i-1] != MX_CR ) {
		status = mx_error( MXE_INTERFACE_IO_ERROR, fname,
			"Warning: Input buffer overrun." );

		buffer[i] = '\0';
	} else {
		status = MX_SUCCESSFUL_RESULT;

		buffer[i-1] = '\0';
	}

#if MXI_COMPUMOTOR_INTERFACE_DEBUG
	if ( 1 ) {
#else
	if ( debug_flag ) {
#endif
		MX_DEBUG(-2,("%s: buffer = '%s'", fname, buffer) );
	}

	return status;
}

MX_EXPORT mx_status_type
mxi_compumotor_putline( MX_COMPUMOTOR_INTERFACE *compumotor_interface,
				char *buffer, int debug_flag )
{
	const char fname[] = "mxi_compumotor_putline()";

	char *ptr;
	char c;
	mx_status_type status;

#if MXI_COMPUMOTOR_INTERFACE_DEBUG
	if ( 1 ) {
#else
	if ( debug_flag ) {
#endif
		MX_DEBUG(-2,("%s: buffer = '%s'", fname, buffer));
	}

	ptr = &buffer[0];

	while ( *ptr != '\0' ) {
		c = *ptr;

		status = mx_rs232_putchar( compumotor_interface->rs232_record,
						c, MXF_232_WAIT );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( debug_flag ) {
			MX_DEBUG(-2,("%s: sent 0x%x '%c'", fname, c, c));
		}

		ptr++;
	}

	/* Send a CR after the end of the command. */

	c = MX_CR;

	status = mx_rs232_putchar( compumotor_interface->rs232_record,
						c, MXF_232_WAIT );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s (eos): sent 0x%x '%c'", fname, c, c));
	}

#if 1
	/* Also send a LF character. */

	c = MX_LF;

	status = mx_rs232_putchar( compumotor_interface->rs232_record,
						c, MXF_232_WAIT );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("%s (eos): sent 0x%x '%c'", fname, c, c));
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

