/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    mx_alloc.c
 *
 * Purpose: MX memory allocation functions.
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_alloc.h"

MX_EXPORT void *
mx_allocate_array( long num_dimensions,
		long *dimension_array,
		size_t *data_element_size_array )
{
	const char fname[] = "mx_allocate_array()";

	void *array_pointer;
	char *array_element_pointer;
	void *subarray_pointer;
	size_t array_size;
	long i, j;
	size_t current_dimension_element_size;
	mx_status_type status;

	if ( num_dimensions <= 0 ) {
		mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal number of dimensions = %ld", num_dimensions );

		return NULL;
	}

	if ( dimension_array == (long *) NULL ) {
		mx_error( MXE_NULL_ARGUMENT, fname,
		"dimension_array pointer passed is NULL." );

		return NULL;
	}

	if ( data_element_size_array == (size_t *) NULL ) {
		mx_error( MXE_NULL_ARGUMENT, fname,
		"data_element_size_array pointer passed is NULL." );

		return NULL;
	}

	MX_DEBUG( 8,("%s: num_dimensions = %ld", fname, num_dimensions));

	for ( i = 0; i < num_dimensions; i++ ) {
		MX_DEBUG( 8,(
			"%s:   dimension[%ld] = %ld, element size[%ld] = %ld",
			fname, i, dimension_array[i],
			i, data_element_size_array[i]));
	}

	current_dimension_element_size
				= data_element_size_array[num_dimensions - 1];

	array_size = dimension_array[0] * current_dimension_element_size;

	MX_DEBUG( 8,("%s: array_size = %ld", fname, (long) array_size));

	if ( array_size <= 0 ) {
		array_pointer = NULL;
	} else {
		array_pointer = (char **) malloc( array_size );
	}

	MX_DEBUG( 8,("%s: array_pointer = %p", fname, array_pointer));

	if ( array_pointer == NULL ) {
		return NULL;
	}

	if ( num_dimensions > 1 ) {
		for ( i = 0; i < dimension_array[0]; i++ ) {
			array_element_pointer = (char *)array_pointer
					+ i * current_dimension_element_size;

			subarray_pointer = mx_allocate_array(
				num_dimensions - 1,
				&dimension_array[1],
				data_element_size_array );

			MX_DEBUG( 8,
		("%s: array_element_pointer = %p, subarray_pointer = %p",
			fname, array_element_pointer, subarray_pointer));

			/* If the allocation of this part of the array
			 * failed, attempt to free all the parts of
			 * the array that we have already allocated.
			 */

			if ( subarray_pointer == NULL ) {
				for ( j = 0; j < i; j++ ) {
					array_element_pointer
						= (char *)array_pointer
					+ j * current_dimension_element_size;

					subarray_pointer =
				mx_read_void_pointer_from_memory_location(
						array_element_pointer );

					status = mx_free_array(
						subarray_pointer,
						num_dimensions - 1,
						&dimension_array[1],
						data_element_size_array );

					/* If we can't even free memory,
					 * then there's not much else we
					 * can do but give up.
					 */

					if ( status.code != MXE_SUCCESS )
						return NULL;
				}

				return NULL;
			}
			mx_write_void_pointer_to_memory_location(
				array_element_pointer, subarray_pointer );
		}
	}

	return array_pointer;
}

MX_EXPORT mx_status_type
mx_free_array( void *array_pointer,
		long num_dimensions,
		long *dimension_array,
		size_t *data_element_size_array )
{
	const char fname[] = "mx_free_array()";

	char *array_element_pointer;
	void *subarray_pointer;
	long i;
	size_t current_dimension_element_size;
	mx_status_type status;

	if ( array_pointer == NULL ) {
		mx_error( MXE_NULL_ARGUMENT, fname,
		"array_pointer passed is NULL." );
	}

	if ( num_dimensions <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal number of dimensions = %ld", num_dimensions );
	}

	if ( dimension_array == (long *) NULL ) {
		mx_error( MXE_NULL_ARGUMENT, fname,
		"dimension_array pointer passed is NULL." );
	}

	if ( data_element_size_array == (size_t *) NULL ) {
		mx_error( MXE_NULL_ARGUMENT, fname,
		"data_element_size_array pointer passed is NULL." );
	}

	MX_DEBUG( 8,("%s: num_dimensions = %ld", fname, num_dimensions));

	MX_DEBUG( 8,("%s: array_pointer = %p", fname, array_pointer));

	for ( i = 0; i < num_dimensions; i++ ) {
		MX_DEBUG( 8,(
			"%s:   dimension[%ld] = %ld, element size[%ld] = %ld",
			fname, i, dimension_array[i],
			i, data_element_size_array[i]));
	}

	current_dimension_element_size
				= data_element_size_array[num_dimensions - 1];

	if ( num_dimensions == 1 ) {
		if ( array_pointer == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
				"array_pointer at address &p is NULL.",
				&array_pointer );
		}

		/* Finally, let's free the memory pointed to by the pointer. */

		MX_DEBUG( 8,( "%s: About to free pointer array_pointer = %p",
			fname, array_pointer ));

		free( array_pointer );

		/* Set the pointer to NULL, so that we will know that it
		 * contains an invalid pointer value.
		 */

		array_pointer = NULL;

		/* free() is defined to have a void return value in ANSI C,
		 * so there really isn't any way to confirm that things
		 * worked correctly.
		 */
	} else {
		for ( i = 0; i < dimension_array[0]; i++ ) {
			array_element_pointer = (char *)array_pointer
					+ i * current_dimension_element_size;

			if ( array_element_pointer == NULL ) {
				return mx_error(
					MXE_CORRUPT_DATA_STRUCTURE, fname, 
				"array_element_pointer at address %p is NULL.",
					&array_element_pointer );
			}

			subarray_pointer
				= mx_read_void_pointer_from_memory_location(
						array_element_pointer );

			if ( subarray_pointer == NULL ) {
				return mx_error(
					MXE_CORRUPT_DATA_STRUCTURE, fname, 
				"subarray_pointer at address %p is NULL.",
					&subarray_pointer );
			}

			MX_DEBUG( 8,
		("%s: array_element_pointer = %p, subarray_pointer = %p",
			fname, array_element_pointer, subarray_pointer ));

			status = mx_free_array( subarray_pointer,
				num_dimensions - 1, &dimension_array[1],
				data_element_size_array );

			/*If freeing the subarray didn't work, then give up.*/

			if ( status.code != MXE_SUCCESS )
				return status;
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/* The purpose of the following function is to interpret the value
 * stored at a memory location pointed to by a void * as if it were
 * itself an address.
 */

MX_EXPORT void *
mx_read_void_pointer_from_memory_location( void *ptr )
{
	/* As far as I know, the only assumption I am making here is
	 * that the internal representation of a void * is the same
	 * as the internal representation of a void **.  This is
	 * reasonably safe (I think).
	 */

	union {
		void *ptr_to_void;
		void **ptr_to_ptr_to_void;
	} u;

	void **result;

	u.ptr_to_void = ptr;

	result = *( u.ptr_to_ptr_to_void );

	MX_DEBUG( 9, ("readvp: read pointer %p from memory location %p",
				result, ptr));

	return result;
}

/* Similarly, the following function writes an address value "ptr"
 * to a memory location pointed to by the void pointer "memory_location".
 */

MX_EXPORT void
mx_write_void_pointer_to_memory_location( void *memory_location, void *ptr )
{
	union {
		void *ptr_to_void;
		void **ptr_to_ptr_to_void;
	} u;

	u.ptr_to_void = memory_location;

	*( u.ptr_to_ptr_to_void ) = ptr;

	MX_DEBUG( 9, ("writevp: wrote pointer %p to memory location %p",
					ptr, memory_location));

	return;
}

