/*
 * Name:     mx_dfile.c
 *
 * Purpose:  Support for writing a data file during a scan.
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>

#include "mx_dfile.h"
#include "f_none.h"
#include "f_child.h"
#include "f_text.h"
#include "f_sff.h"
#include "f_xafs.h"

MX_DATAFILE_TYPE_ENTRY mx_datafile_type_list[] = {
	{ MXDF_NONE,  "none",  &mxdf_none_datafile_function_list },
	{ MXDF_CHILD, "child", &mxdf_child_datafile_function_list },
	{ MXDF_TEXT,  "text",  &mxdf_text_datafile_function_list },
	{ MXDF_SFF,   "sff",   &mxdf_sff_datafile_function_list },
	{ MXDF_XAFS,  "xafs",  &mxdf_xafs_datafile_function_list },
	{ -1, "", NULL }
};

MX_EXPORT mx_status_type
mx_get_datafile_type_by_name( 
	MX_DATAFILE_TYPE_ENTRY *datafile_type_list,
	char *name,
	MX_DATAFILE_TYPE_ENTRY **datafile_type_entry )
{
	const char fname[] = "mx_get_datafile_type_by_name()";

	char *list_name;
	int i;

	if ( datafile_type_list == NULL ) {
		*datafile_type_entry = NULL;

		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Datafile type list passed was NULL." );
	}

	if ( name == NULL ) {
		*datafile_type_entry = NULL;

		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Plot name pointer passed is NULL.");
	}

	for ( i=0; ; i++ ) {
		/* Check for the end of the list. */

		if ( datafile_type_list[i].type < 0 ) {
			*datafile_type_entry = NULL;

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Datafile type '%s' was not found.", name );
		}

		list_name = datafile_type_list[i].name;

		if ( list_name == NULL ) {
			*datafile_type_entry = NULL;

			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
				"NULL name ptr for datafile type %d.",
				datafile_type_list[i].type );
		}

		if ( strcmp( name, list_name ) == 0 ) {
			*datafile_type_entry = &( datafile_type_list[i] );

			MX_DEBUG( 8,
			("mx_get_datafile_type_by_name: ptr = 0x%p, type = %d",
			*datafile_type_entry, (*datafile_type_entry)->type));

			return MX_SUCCESSFUL_RESULT;
		}
	}
}

MX_EXPORT mx_status_type mx_get_datafile_type_by_value( 
	MX_DATAFILE_TYPE_ENTRY *datafile_type_list,
	long datafile_type,
	MX_DATAFILE_TYPE_ENTRY **datafile_type_entry )
{
	const char fname[] = "mx_get_datafile_type_by_value()";

	int i;

	if ( datafile_type_list == NULL ) {
		*datafile_type_entry = NULL;

		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Datafile type list passed was NULL." );
	}

	if ( datafile_type <= 0 ) {
		*datafile_type_entry = NULL;

		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Datafile type %ld is illegal.", datafile_type );
	}

	for ( i=0; ; i++ ) {
		/* Check for the end of the list. */

		if ( datafile_type_list[i].type < 0 ) {
			*datafile_type_entry = NULL;

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Datafile type %ld was not found.", datafile_type );
		}

		if ( datafile_type_list[i].type == datafile_type ) {
			*datafile_type_entry = &( datafile_type_list[i] );

			MX_DEBUG( 8, ("%s: ptr = 0x%p, type = %d", fname,
				*datafile_type_entry,
				(*datafile_type_entry)->type));

			return MX_SUCCESSFUL_RESULT;
		}
	}
}

MX_EXPORT mx_status_type
mx_datafile_open( MX_DATAFILE *datafile )
{
	const char fname[] = "mx_datafile_open()";

	MX_DATAFILE_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_DATAFILE * );
	mx_status_type status;

	if ( datafile == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_DATAFILE pointer passed was NULL.");
	}

	flist
	  = (MX_DATAFILE_FUNCTION_LIST *) (datafile->datafile_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_DATAFILE_FUNCTION_LIST pointer for datafile is NULL");
	}

	fptr = flist->open;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"open function pointer for datafile type %ld is NULL",
			datafile->type );
	}

	status = (*fptr) ( datafile );

	return status;
}

MX_EXPORT mx_status_type
mx_datafile_close( MX_DATAFILE *datafile )
{
	const char fname[] = "mx_datafile_close()";

	MX_DATAFILE_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_DATAFILE * );
	mx_status_type status;

	if ( datafile == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_DATAFILE pointer passed was NULL.");
	}

	flist
	  = (MX_DATAFILE_FUNCTION_LIST *) (datafile->datafile_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_DATAFILE_FUNCTION_LIST pointer for datafile is NULL");
	}

	/* Now invoke the close function. */

	fptr = flist->close;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"close function pointer for datafile type %ld is NULL",
			datafile->type );
	}

	status = (*fptr) ( datafile );

	/* If the datafile close succeeded, transform the datafile filename
	 * to be ready for the next scan.
	 */

	if ( status.code == MXE_SUCCESS ) {
		status = mx_update_datafile_name( datafile );
	}

	return status;
}

MX_EXPORT mx_status_type
mx_datafile_write_main_header( MX_DATAFILE *datafile )
{
	const char fname[] = "mx_datafile_write_main_header()";

	MX_DATAFILE_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_DATAFILE * );
	mx_status_type status;

	if ( datafile == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_DATAFILE pointer passed was NULL.");
	}

	flist
	  = (MX_DATAFILE_FUNCTION_LIST *) (datafile->datafile_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_DATAFILE_FUNCTION_LIST pointer for datafile is NULL");
	}

	/* Now invoke the write_main_header function. */

	fptr = flist->write_main_header;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"write_main_header function pointer for datafile type %ld is NULL",
			datafile->type );
	}

	status = (*fptr) ( datafile );

	return status;
}

MX_EXPORT mx_status_type
mx_datafile_write_segment_header( MX_DATAFILE *datafile )
{
	const char fname[] = "mx_datafile_write_segment_header()";

	MX_DATAFILE_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_DATAFILE * );
	mx_status_type status;

	if ( datafile == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_DATAFILE pointer passed was NULL.");
	}

	flist
	  = (MX_DATAFILE_FUNCTION_LIST *) (datafile->datafile_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_DATAFILE_FUNCTION_LIST pointer for datafile is NULL");
	}

	/* Now invoke the write_segment_header function. */

	fptr = flist->write_segment_header;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"write_segment_header function pointer for datafile type %ld is NULL",
			datafile->type );
	}

	status = (*fptr) ( datafile );

	return status;
}

MX_EXPORT mx_status_type
mx_datafile_write_trailer( MX_DATAFILE *datafile )
{
	const char fname[] = "mx_datafile_write_trailer()";

	MX_DATAFILE_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_DATAFILE * );
	mx_status_type status;

	if ( datafile == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_DATAFILE pointer passed was NULL.");
	}

	flist
	  = (MX_DATAFILE_FUNCTION_LIST *) (datafile->datafile_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_DATAFILE_FUNCTION_LIST pointer for datafile is NULL");
	}

	/* Now invoke the write_trailer function. */

	fptr = flist->write_trailer;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"write_trailer function pointer for datafile type %ld is NULL",
			datafile->type );
	}

	status = (*fptr) ( datafile );

	return status;
}

MX_EXPORT mx_status_type
mx_add_measurement_to_datafile( MX_DATAFILE *datafile )
{
	const char fname[] = "mx_add_measurement_to_datafile()";

	MX_DATAFILE_FUNCTION_LIST *flist;
	mx_status_type ( *fptr ) ( MX_DATAFILE * );
	mx_status_type status;

	if ( datafile == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_DATAFILE pointer passed was NULL.");
	}

	flist
	  = (MX_DATAFILE_FUNCTION_LIST *) (datafile->datafile_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_DATAFILE_FUNCTION_LIST pointer for datafile is NULL");
	}

	/* Now invoke the add_measurement_to_datafile function. */

	fptr = flist->add_measurement_to_datafile;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"add_measurement_to_datafile function pointer for datafile type %ld is NULL",
			datafile->type );
	}

	status = (*fptr) ( datafile );

	return status;
}

MX_EXPORT mx_status_type
mx_add_array_to_datafile( MX_DATAFILE *datafile,
		long position_type, long num_positions, void *position_array,
		long data_type, long num_data_points, void *data_array )
{
	const char fname[] = "mx_add_array_to_datafile()";

	MX_DATAFILE_FUNCTION_LIST *flist;
	mx_status_type (*fptr)(MX_DATAFILE *,
				long, long, void *, long, long, void *);
	mx_status_type status;

	if ( datafile == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_DATAFILE pointer passed was NULL.");
	}

	flist
	  = (MX_DATAFILE_FUNCTION_LIST *) (datafile->datafile_function_list);

	if ( flist == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_DATAFILE_FUNCTION_LIST pointer for datafile is NULL");
	}

	/* Now invoke the add_array_to_datafile function. */

	fptr = flist->add_array_to_datafile;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"add_array_to_datafile function pointer for datafile type %ld is NULL",
			datafile->type );
	}

	status = (*fptr) ( datafile,
			position_type, num_positions, position_array,
			data_type, num_data_points, data_array );

	return status;
}

MX_EXPORT mx_status_type
mx_update_datafile_name( MX_DATAFILE *datafile )
{
	const char fname[] = "mx_update_datafile_name()";

	char *filename;
	char *period_ptr, *number_string_ptr;
	size_t number_length, digits_length;
	unsigned long number_value, old_number_value, divisor;
	int i, max_digits, num_items;

	if ( datafile == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_DATAFILE pointer passed was NULL.");
	}

	/* If the datafile name looks like "arbitrary_string.integer_number",
	 * we increment the integer number field by 1.  Otherwise, we do
	 * nothing to the filename.  For example, "myname.027" will be
	 * incremented to "myname.028".
	 *
	 * The number of digits in the number field is always kept the same.
	 * If the number has a value that would require a larger number
	 * of digits to display than before, the number instead wraps back
	 * to zero.  For example, "myname.999" would wrap back to "myname.000".
	 */

	filename = datafile->filename;

	/* Is there a period in the filename? */

	period_ptr = strrchr( filename, '.' );

	if ( period_ptr == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Is there anything in the string after the period? */

	number_string_ptr = period_ptr + 1;

	if ( *number_string_ptr == '\0' ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Does the part of the string after the period entirely consist
	 * of the digits 0-9 and nothing else?
	 */

	number_length = strlen( number_string_ptr );

	digits_length = strspn( number_string_ptr, "0123456789" );

	if ( number_length != digits_length ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* We have determined that the part of the string after the period 
	 * is composed entirely of the digits 0-9.  Read in the corresponding
	 * number.
	 */

	num_items = sscanf( number_string_ptr, "%lu", &old_number_value );

	if ( num_items != 1 ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
		"Parse error in sscanf() for filename '%s'", filename );
	}

	/* If the string representation of 'number_value' now would require
	 * more characters than before, cause the value to wrap back to zero.
	 * For example, if 'old_number_value' was 999, force 'number_value'
	 * to be 0.  This is done by computing a divisor, which in this case
	 * would be 1000 and then discovering that while 999/1000 == 0,
	 * 1000/1000 == 1 (using integer division).
	 */

	divisor = 1U;
	max_digits = 0;

	for ( i = 0; i < (int) number_length; i++ ) {

		if ( ULONG_MAX / divisor < 10U ) {
			return mx_error( MXE_FUNCTION_FAILED, fname,
"There are too many digits (%d) in the number string '%s' in filename '%s'.  "
"Reduce it to %d digits or less.",
				number_length, number_string_ptr, filename,
				max_digits );
		}

		divisor *= 10U;

		max_digits++;
	}

	number_value = old_number_value + 1U;

	if ( number_value / divisor > 0U ) {
		number_value = 0U;
	}

	/* Now update the filename while keeping the same number of digits
	 * in the number string field by the use of zero fill.
	 */

	sprintf(number_string_ptr, "%0*lu", (int) number_length, number_value);

	return MX_SUCCESSFUL_RESULT;
}

/*
 * Datafile name handling policy:
 *
 * If the datafile name looks like "arbitrary_string.integer_number",
 * we increment the integer number field by 1.  Otherwise, we do
 * nothing to the filename.  For example, "myname.027" will be
 * incremented to "myname.028".
 *
 * The number of digits in the number field is always kept the same.
 * If the number has a value that would require a larger number
 * of digits to display than before, the number instead wraps back
 * to zero.  For example, "myname.999" would wrap back to "myname.000".
 */

/*
 * The function mx_parse_datafile_name() looks for the presence of a version
 * number at the end of the filename.  If the last period character in
 * the filename is followed only by numerical digits, then the end of
 * the filename is interpreted as a version number.  If the filename
 * has a version number in it, a pointer 'version_number_pointer' to the
 * start of the character string representation of the version number
 * in the supplied buffer 'datafile_name_buffer' is returned.  In addition,
 * a numerical value 'version_number' for the version number is returned,
 * along with the number of characters 'version_number_length' in the
 * string representation of the number.  If the filename has no version
 * number, then 'version_number_pointer' is NULL, while 'version_number'
 * and 'version_number_length' are set to -1.
 */

MX_EXPORT mx_status_type
mx_parse_datafile_name( char *datafile_name, char **version_number_pointer,
			long *version_number, long *version_number_length )
{
	const char fname[] = "mx_parse_datafile_name()";

	char *period_ptr, *number_string_ptr;
	size_t number_length, digits_length;
	unsigned long old_number_value;
	int num_items;

	*version_number_pointer = NULL;
	*version_number = -1;
	*version_number_length = -1;

	/* Is there a period in the filename? */

	period_ptr = strrchr( datafile_name, '.' );

	if ( period_ptr == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Is there anything in the string after the period? */

	number_string_ptr = period_ptr + 1;

	if ( *number_string_ptr == '\0' ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Does the part of the string after the period entirely consist
	 * of the digits 0-9 and nothing else?
	 */

	number_length = strlen( number_string_ptr );

	digits_length = strspn( number_string_ptr, "0123456789" );

	if ( number_length != digits_length ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* We have determined that the part of the string after the period 
	 * is composed entirely of the digits 0-9.  Read in the corresponding
	 * number.
	 */

	num_items = sscanf( number_string_ptr, "%lu", &old_number_value );

	if ( num_items != 1 ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Parse error in sscanf() for filename '%s'",
			datafile_name );
	}

	*version_number_pointer = number_string_ptr;
	*version_number         = old_number_value;
	*version_number_length  = number_length;

	return MX_SUCCESSFUL_RESULT;
}

/* If the string representation of 'version_number' now would require
 * more characters than before, cause the value to wrap back to zero.
 * For example, if 'old_version_number' was 999, force 'version_number'
 * to be 0.  This is done by computing a divisor, which in this case
 * would be 1000 and then discovering that while 999/1000 == 0,
 * 1000/1000 == 1 (using integer division).
 */

MX_EXPORT long
mx_construct_datafile_version_number( long test_version_number,
					long version_number_length )
{
	const char fname[] = "mx_construct_datafile_version_number()";

	unsigned long version_number, divisor;
	int i, max_digits;

	divisor = 1U;
	max_digits = 0;

	for ( i = 0; i < (int) version_number_length; i++ ) {

		if ( ULONG_MAX / divisor < 10U ) {
			(void) mx_error( MXE_FUNCTION_FAILED, fname,
"There are too many digits (%d) in the file version number %ld.  "
"Reduce it to %d digits or less.",
				version_number_length,
				test_version_number,
				max_digits );

			return -1;
		}

		divisor *= 10U;

		max_digits++;
	}

	version_number = test_version_number;

	if ( version_number / divisor > 0U ) {
		version_number = 0U;
	}

	return version_number;
}

