/*
 * Name:    mx_motor.c
 *
 * Purpose: MX function library of generic motor operations.
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_key.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_motor.h"

/*=======================================================================*/

/* This function is used by a motor's finish_record_initialization
 * driver entry point to initialize several fields based on the 
 * values in the record description.
 */

MX_EXPORT mx_status_type
mx_motor_finish_record_initialization( MX_RECORD *motor_record )
{
	const char fname[] = "mx_motor_finish_record_initialization()";

	MX_MOTOR *motor;

	if ( motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}
	motor = (MX_MOTOR *) (motor_record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			motor_record->name );
	}
	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:
		motor->position = motor->offset + motor->scale
				* (double) motor->raw_position.stepper;

		motor->backlash_correction = motor->offset + motor->scale
			* (double) motor->raw_backlash_correction.stepper;
		
		motor->negative_limit = motor->offset + motor->scale
				* (double) motor->raw_negative_limit.stepper;

		motor->positive_limit = motor->offset + motor->scale
				* (double) motor->raw_positive_limit.stepper;

		break;
	case MXC_MTR_ANALOG:
		motor->position = motor->offset + motor->scale
					* motor->raw_position.analog;

		motor->backlash_correction = motor->offset + motor->scale
				* motor->raw_backlash_correction.analog;
		
		motor->negative_limit = motor->offset + motor->scale
					* motor->raw_negative_limit.analog;

		motor->positive_limit = motor->offset + motor->scale
					* motor->raw_positive_limit.analog;

		break;
	}

	motor->setpoint = motor->position;

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

/* This function is a utility function to consolidate all of the pointer
 * mangling that often has to happen at the beginning of an mx_motor_...
 * function.  The parameter 'calling_fname' is passed so that error messages
 * will appear with the name of the calling function.
 */

MX_EXPORT mx_status_type
mx_motor_get_pointers( MX_RECORD *motor_record,
			MX_MOTOR **motor,
			MX_MOTOR_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_motor_get_pointers()";

	if ( motor_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The motor_record pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( motor_record->mx_class != MXC_MOTOR ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The record '%s' passed by '%s' is not a motor.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			motor_record->name, calling_fname,
			motor_record->mx_superclass,
			motor_record->mx_class,
			motor_record->mx_type );
	}

	if ( motor != (MX_MOTOR **) NULL ) {
		*motor = (MX_MOTOR *) (motor_record->record_class_struct);

		if ( *motor == (MX_MOTOR *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MOTOR pointer for record '%s' passed by '%s' is NULL",
				motor_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_MOTOR_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_MOTOR_FUNCTION_LIST *)
				(motor_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_MOTOR_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_MOTOR_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				motor_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mx_motor_is_busy( MX_RECORD *motor_record, int *busy )
{
	const char fname[] = "mx_motor_is_busy()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->motor_is_busy;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"motor_is_busy function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	*busy = motor->busy;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_soft_abort( MX_RECORD *motor_record )
{
	const char fname[] = "mx_motor_soft_abort()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->soft_abort;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"soft_abort function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_immediate_abort( MX_RECORD *motor_record )
{
	const char fname[] = "mx_motor_immediate_abort()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR *);
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->immediate_abort;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"immediate_abort function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	return status;
}

/* --- Move by engineering units functions. --- */

MX_EXPORT mx_status_type
mx_motor_move_relative(MX_RECORD *motor_record,
			double relative_position,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags)
{
	double current_position, new_position;
	mx_status_type status;

	status = mx_motor_get_position( motor_record, &current_position );

	if ( status.code != MXE_SUCCESS )
		return status;

	new_position = current_position + relative_position;

	status = mx_motor_move_absolute( motor_record,
				new_position, display_motor_move_fn, flags );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_move_absolute(MX_RECORD *motor_record,
			double position,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags)
{
	const char fname[] = "mx_motor_move_absolute()";

	MX_MOTOR *motor;
	double raw_setpoint;
	long raw_motor_steps;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 2,("%s: moving '%s' to %g",
			fname, motor_record->name, position ));

	if ( fabs(motor->scale) < MX_MOTOR_ANALOG_FUZZ ) {
		raw_setpoint = 0.0;
	} else {
		raw_setpoint = (position - motor->offset) / motor->scale;
	}

	switch ( motor->subclass ) {
	case MXC_MTR_STEPPER:
		raw_motor_steps = mx_round( raw_setpoint );

		status = mx_motor_move_absolute_steps( motor_record,
			raw_motor_steps, display_motor_move_fn, flags );

		break;

	case MXC_MTR_ANALOG:
		status = mx_motor_move_absolute_analog( motor_record,
			raw_setpoint, display_motor_move_fn, flags );

		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown motor subclass %d.", motor->subclass );
		break;
	}

	return status;
}

MX_EXPORT mx_status_type
mx_motor_array_move_absolute( int num_motors,
			MX_RECORD **motor_record_array,
			double *motor_position,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags )
{
	const char fname[] = "mx_motor_array_move_absolute()";

	MX_MOTOR *motor;
	int i, modified_flags, result_flag;
	int do_backlash, individual_backlash_needed;
	double backlash, backlash_position;
	double present_position, relative_motion;
	mx_status_type status;

	/* See if any of the requested moves will exceed a software limit. */

	MX_DEBUG( 2,("%s: Checking software limits.",fname));

	for ( i = 0; i < num_motors; i++ ) {
		status = mx_is_motor_position_between_software_limits(
				motor_record_array[i], motor_position[i],
				&result_flag, TRUE);

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	/* Move to the backlash_position. */

	MX_DEBUG( 2,("%s: Checking for backlash correction.", fname));

	if ( (flags & MXF_MTR_IGNORE_BACKLASH) == 0 ) {

		/* We must also check to see if the backlash correction
		 * will exceed a limit.
		 */

		do_backlash = FALSE;

		for ( i = 0; i < num_motors; i++ ) {

			motor = (MX_MOTOR *)
				motor_record_array[i]->record_class_struct;

			backlash = motor->backlash_correction;

			status = mx_motor_get_position( motor_record_array[i],
							&present_position );
			if ( status.code != MXE_SUCCESS )
				return status;

			relative_motion = motor_position[i] - present_position;

			/* A given motor needs to do a backlash correction
			 * if relative_motion and backlash have the same signs.
			 */

			individual_backlash_needed = FALSE;

			if (( relative_motion > 0.0 ) && ( backlash > 0.0 )) {
				individual_backlash_needed = TRUE;
			} else
			if (( relative_motion < 0.0 ) && ( backlash < 0.0 )) {
				individual_backlash_needed = TRUE;
			}

			if ( individual_backlash_needed ) {

				do_backlash = TRUE;

				backlash_position
					= motor_position[i] + backlash;

				status =
				  mx_is_motor_position_between_software_limits(
					motor_record_array[i],
					backlash_position,
					&result_flag, TRUE);

				if ( status.code != MXE_SUCCESS )
					return status;
			}
		}

		/* Do the backlash correction. */

		if ( do_backlash ) {
			modified_flags = flags
					   | MXF_MTR_GO_TO_BACKLASH_POSITION;

			modified_flags &= ( ~ MXF_MTR_NOWAIT );

			MX_DEBUG( 2,
("%s: Backlash correction via mx_motor_array_move_segment().", fname));

			status = mx_motor_array_move_segment( num_motors,
					motor_record_array,
					motor_position,
					display_motor_move_fn,
					modified_flags );

			if ( status.code != MXE_SUCCESS )
				return status;
		}
	}

	/* Now move to the final position. */

	modified_flags = flags | MXF_MTR_IGNORE_BACKLASH;

	MX_DEBUG( 2,
("%s: Main move via mx_motor_array_move_segment().", fname));

	status = mx_motor_array_move_segment( num_motors,
					motor_record_array,
					motor_position,
					display_motor_move_fn,
					modified_flags );
	return status;
}

MX_EXPORT mx_status_type
mx_motor_array_move_segment( int num_motors,
			MX_RECORD **motor_record_array,
			double *motor_position,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags )
{
	const char fname[] = "mx_motor_array_move_segment()";

	mx_status_type status;
	int display_move_flag, modified_flags;
	int wait_flag;
	int i, j;

	MX_DEBUG( 2,
		("%s invoked.  flags = %d, display_motor_move_fn = %p",
		fname, flags, display_motor_move_fn));

	display_move_flag = flags & MXF_MTR_SHOW_MOVE;

	modified_flags = flags & ( ~ MXF_MTR_SHOW_MOVE );

	modified_flags |= MXF_MTR_NOWAIT;

	/* Start the motors in sequence.  No simultaneous starts for now. */

	for ( i = 0; i < num_motors; i++ ) {
		MX_DEBUG( 2,("%s: moving %s to %g",
			fname, motor_record_array[i]->name,
			motor_position[i]));

		status = mx_motor_move_absolute(
			motor_record_array[i], motor_position[i],
			NULL, modified_flags );

		if ( status.code != MXE_SUCCESS ) {
			for ( j = 0; j < i-1; j++ ) {
				(void) mx_motor_soft_abort(
						motor_record_array[j] );
			}
			return status;
		}
	}

	MX_DEBUG( 2,("%s: About to display motor moves.", fname));

	/* Display the motor moves if requested. */

	if ( display_move_flag ) {
		status = ( *display_motor_move_fn ) ( flags,
					num_motors, motor_record_array );
	}

	/* Wait for the motor move to complete if requested. */

	wait_flag = ! ( flags & MXF_MTR_NOWAIT );

	if ( wait_flag ) {
		MX_DEBUG( 2,("%s: wait flag is set.",fname));

		status = mx_wait_for_motor_array_stop( num_motors,
						motor_record_array, flags );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_wait_for_motor_stop( MX_RECORD *motor_record, int flags )
{
	const char fname[] = "mx_wait_for_motor_stop()";

	int busy, interrupt, limit_hit;
	mx_status_type status;

	while (1) {
		status = mx_motor_is_busy(motor_record, &busy);

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( busy == FALSE ) {
			/* Exit the while() loop. */

			break;
		}

		if ( (flags & MXF_MTR_IGNORE_KEYBOARD) == 0 ) {

			/* Did someone hit a key? */

			interrupt = mx_user_requested_interrupt();

			switch( interrupt ) {
			case MXF_USER_INT_NONE:
				/* No interrupt ocurred. */

				break;

			case MXF_USER_INT_ABORT:
				mx_motor_soft_abort( motor_record );

				return mx_error( MXE_INTERRUPTED, fname,
					"Move to backlash position aborted.");

				break;

			case MXF_USER_INT_PAUSE:
				return mx_error( MXE_PAUSE_REQUESTED, fname,
					"Pause requested by user.");

				break;
			default:
				return mx_error( MXE_FUNCTION_FAILED, fname,
					"Unexpected value %d returned "
					"by mx_user_requested_interrupt()",
						    interrupt );
				break;
			}
		}

		/* Has the motor hit a limit? */

		status = mx_motor_positive_limit_hit(
				motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( limit_hit ) {
			(void) mx_motor_soft_abort( motor_record );

			return mx_error( MXE_INTERRUPTED, fname,
			"Positive limit switch hit.  Motor move aborted." );
		}

		status = mx_motor_negative_limit_hit(
				motor_record, &limit_hit );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( limit_hit ) {
			(void) mx_motor_soft_abort( motor_record );

			return mx_error( MXE_INTERRUPTED, fname,
			"Negative limit switch hit.  Motor move aborted." );
		}

		mx_msleep(500);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_wait_for_motor_array_stop( int num_motor_records,
			MX_RECORD **motor_record_array,
			int flags )
{
	const char fname[] = "mx_wait_for_motor_array_stop()";

	int i, j, busy, motor_is_moving;
	int limit_hit, any_limit_hit, interrupt;
	mx_status_type status;

	motor_is_moving = TRUE;
	any_limit_hit = FALSE;

	while ( motor_is_moving && ! any_limit_hit ) {

		motor_is_moving = FALSE;

		for ( i = 0; i < num_motor_records; i++ ) {
			/* Are any motors still moving? */

			status = mx_motor_is_busy(
					motor_record_array[i], &busy );

			if ( status.code != MXE_SUCCESS ) {
				for ( j = 0; j < num_motor_records; j++ ) {
					(void) mx_motor_soft_abort(
						motor_record_array[j]);
				}
				return status;
			}

			if ( busy )
				motor_is_moving = TRUE;

			if ( (flags & MXF_MTR_IGNORE_KEYBOARD) == 0 ){
				/* Did someone hit a key? */

				interrupt = mx_user_requested_interrupt();

				switch( interrupt ) {
				case MXF_USER_INT_NONE:
					/* No interrupt occurred. */

					break;

				case MXF_USER_INT_ABORT:

					for (j=0; j < num_motor_records; j++){
						(void) mx_motor_soft_abort(
							motor_record_array[j]);
					}
					return mx_error(
						MXE_INTERRUPTED, fname,
						"Motor moves aborted.");
					break;

				case MXF_USER_INT_PAUSE:
					return mx_error(
						MXE_PAUSE_REQUESTED,
						fname,
					"Pause requested by user." );

					break;

				default:
					return mx_error(
						MXE_FUNCTION_FAILED,
						fname,
				    "Unexpected value %d returned "
				    "by mx_user_requested_interrupt()",
						interrupt );
					break;
				}
			}

			/* Have any motors hit a limit? */

			status = mx_motor_positive_limit_hit(
				motor_record_array[i], &limit_hit );

			if ( status.code != MXE_SUCCESS ) {
				for ( j = 0; j < num_motor_records; j++ ) {
					(void) mx_motor_soft_abort(
						motor_record_array[j]);
				}
				return status;
			}

			if ( limit_hit )
				any_limit_hit = TRUE;

			status = mx_motor_negative_limit_hit(
				motor_record_array[i], &limit_hit );

			if ( status.code != MXE_SUCCESS ) {
				for ( j = 0; j < num_motor_records; j++ ) {
					(void) mx_motor_soft_abort(
						motor_record_array[j]);
				}
				return status;
			}

			if ( limit_hit )
				any_limit_hit = TRUE;
		}

		if ( any_limit_hit ) {
			for ( j = 0; j < num_motor_records; j++ ) {
				(void) mx_motor_soft_abort(
						motor_record_array[j]);
			}
			return mx_error( MXE_INTERRUPTED, fname,
			"Limit switch hit.  Motor moves aborted." );
		}

		mx_msleep(10);
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_motor_get_position( MX_RECORD *motor_record, double *position )
{
	const char fname[] = "mx_motor_get_position()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	double raw_position;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"get_position function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:
		raw_position = (double) ( motor->raw_position.stepper );
		break;
	case MXC_MTR_ANALOG:
		raw_position = motor->raw_position.analog;
		break;
	default:
		raw_position = 0.0;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unrecognized motor subclass %d for motor '%s'.",
			motor->subclass, motor_record->name );
	}

	/* Update position. */

	motor->position = motor->offset + (motor->scale) * raw_position;

	*position = motor->position;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_position( MX_RECORD *motor_record, double setpoint )
{
	const char fname[] = "mx_motor_set_position()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	double raw_setpoint;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"set_position function pointer is NULL.");
	}

	if ( fabs( motor->scale ) < MX_MOTOR_ANALOG_FUZZ ) {
		raw_setpoint = 0.0;
	} else {
		raw_setpoint = ( setpoint - motor->offset ) / motor->scale;
	}

	switch( motor->subclass ) {
	case MXC_MTR_STEPPER:
		motor->raw_setpoint.stepper = mx_round( raw_setpoint );
		break;
	case MXC_MTR_ANALOG:
		motor->raw_setpoint.analog = raw_setpoint;
		break;
	}

	/* Update setpoint. */

	motor->setpoint = motor->offset + motor->scale * raw_setpoint;

	/* Invoke the function. */

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_positive_limit_hit( MX_RECORD *motor_record, int *limit_hit )
{
	const char fname[] = "mx_motor_positive_limit_hit()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->positive_limit_hit;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"positive_limit_hit function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	*limit_hit = motor->positive_limit_hit;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_negative_limit_hit( MX_RECORD *motor_record, int *limit_hit )
{
	const char fname[] = "mx_motor_negative_limit_hit()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->negative_limit_hit;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"negative_limit_hit function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	*limit_hit = motor->negative_limit_hit;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_find_home_position( MX_RECORD *motor_record, int direction )
{
	const char fname[] = "mx_motor_find_home_position()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->find_home_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"find_home_position function pointer is NULL.");
	}

	if ( motor->scale >= 0.0 ) {
		motor->home_search = direction;
	} else {
		motor->home_search = -direction;
	}

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_zero_position_value( MX_RECORD *motor_record )
{
	mx_status_type status;

	status = mx_motor_set_position( motor_record, 0L );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_constant_velocity_move( MX_RECORD *motor_record, int direction )
{
	const char fname[] = "mx_motor_constant_velocity_move()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->constant_velocity_move;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"constant_velocity_move function pointer is NULL.");
	}

	if ( motor->scale >= 0.0 ) {
		motor->constant_velocity_move = direction;
	} else {
		motor->constant_velocity_move = -direction;
	}

	status = ( *fptr ) ( motor );

	return status;
}

/*-----------------------------------------------------------------------*/

/* Setting a speed limit to -1 means ignore that limit, while setting it
 * to any other negative value means the motor speed may not be changed.
 */

static mx_status_type
mx_motor_check_speed_limits( MX_MOTOR *motor, long speed_type )
{
	const char fname[] = "mx_motor_check_speed_limits()";

	char speed_type_name[20];
	double speed, raw_speed;
	long test_value;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_MOTOR pointer passed was NULL." );
	}

	/* Make sure the engineering limit values match the raw limits. */

	motor->maximum_speed_limit = motor->raw_maximum_speed_limit
					* fabs( motor->scale );

	motor->minimum_speed_limit = motor->raw_minimum_speed_limit
					* fabs( motor->scale );

	switch( speed_type ) {
	case MXLV_MTR_SPEED:
		strcpy( speed_type_name, "speed" );

		speed = motor->speed;
		raw_speed = motor->raw_speed;
		break;

	case MXLV_MTR_BASE_SPEED:
		strcpy( speed_type_name, "base speed" );

		speed = motor->base_speed;
		raw_speed = motor->raw_base_speed;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported speed type %ld was passed for motor '%s'",
			speed_type, motor->record->name );
		break;
	}

	/* Check for negative and zero speeds. */

	if ( speed <= 0.0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The %s of %g %s/sec requested for motor '%s' is illegal.  "
		"Motor speeds must be positive numbers.",
			speed_type_name, speed, motor->units,
			motor->record->name );
	}

	/* Check the upper limit. */

	if ( motor->raw_maximum_speed_limit < 0.0 ) {

		test_value = mx_round( motor->raw_maximum_speed_limit );

		/* -1 means ignore the limit. */

		if ( test_value != -1 ) {
			return mx_error( MXE_PERMISSION_DENIED, fname,
			"The %s of motor '%s' is not allowed to be changed.",
				speed_type_name, motor->record->name );
		}
	} else {
		if ( raw_speed > motor->raw_maximum_speed_limit ) {
			return mx_error( MXE_PERMISSION_DENIED, fname,
			"The %s of %g %s/sec requested for '%s' is above the "
			"maximum allowed speed of %g %s/sec.",
				speed_type_name, speed, motor->units,
				motor->record->name,
				motor->maximum_speed_limit, motor->units );
		}
	}

	/* Check the lower limit. */

	if ( motor->raw_minimum_speed_limit < 0.0 ) {

		test_value = mx_round( motor->raw_minimum_speed_limit );

		/* -1 means ignore the limit. */

		if ( test_value != -1 ) {
			return mx_error( MXE_PERMISSION_DENIED, fname,
			"The %s of motor '%s' is not allowed to be changed.",
				speed_type_name, motor->record->name );
		}
	} else {
		if ( raw_speed < motor->raw_minimum_speed_limit ) {
			return mx_error( MXE_PERMISSION_DENIED, fname,
			"The %s of %g %s/sec requested for '%s' is below the "
			"minimum allowed speed of %g %s/sec.",
				speed_type_name, speed, motor->units,
				motor->record->name,
				motor->minimum_speed_limit, motor->units );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/*-----------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mx_motor_get_speed( MX_RECORD *motor_record, double *speed )
{
	const char fname[] = "mx_motor_get_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The get_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_SPEED;

	status = ( *fptr ) ( motor );

	motor->speed = fabs(motor->scale) * motor->raw_speed;

	*speed = motor->speed;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_speed( MX_RECORD *motor_record, double speed )
{
	const char fname[] = "mx_motor_set_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The set_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_SPEED;

	motor->speed = speed;

	motor->raw_speed = mx_divide_safely(motor->speed, fabs(motor->scale));

	status = mx_motor_check_speed_limits( motor, MXLV_MTR_SPEED );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Change the speed. */

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_get_base_speed( MX_RECORD *motor_record, double *base_speed )
{
	const char fname[] = "mx_motor_get_base_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The get_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_BASE_SPEED;

	status = ( *fptr ) ( motor );

	motor->base_speed = fabs(motor->scale) * motor->raw_base_speed;

	*base_speed = motor->base_speed;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_base_speed( MX_RECORD *motor_record, double base_speed )
{
	const char fname[] = "mx_motor_set_base_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The set_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_BASE_SPEED;

	motor->base_speed = base_speed;

	motor->raw_base_speed = mx_divide_safely( motor->base_speed,
							fabs(motor->scale) );

	status = mx_motor_check_speed_limits( motor, MXLV_MTR_BASE_SPEED );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Change the base speed. */

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_get_maximum_speed( MX_RECORD *motor_record, double *maximum_speed )
{
	const char fname[] = "mx_motor_get_maximum_speed()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The get_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_MAXIMUM_SPEED;

	status = ( *fptr ) ( motor );

	motor->maximum_speed = fabs(motor->scale) * motor->raw_maximum_speed;

	*maximum_speed = motor->maximum_speed;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_maximum_speed( MX_RECORD *motor_record, double maximum_speed )
{
	const char fname[] = "mx_motor_set_maximum_speed()";

#if 1
	return mx_error( MXE_UNSUPPORTED, fname,
		"This function is currently disabled." );
#else
	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The set_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_MAXIMUM_SPEED;

	motor->maximum_speed = maximum_speed;

	if ( fabs( motor->scale ) < MX_MOTOR_ANALOG_FUZZ ) {
		motor->raw_maximum_speed = 0.0;
	} else {
		motor->raw_maximum_speed
			= motor->maximum_speed / fabs(motor->scale);
	}

	status = ( *fptr ) ( motor );

	return status;
#endif
}

MX_EXPORT mx_status_type
mx_motor_get_acceleration_type( MX_RECORD *motor_record,
				int *acceleration_type )
{
	const char fname[] = "mx_motor_get_acceleration_type()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The get_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_ACCELERATION_TYPE;

	status = ( *fptr ) ( motor );

	*acceleration_type = motor->acceleration_type;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_get_acceleration_parameters( MX_RECORD *motor_record,
				double *acceleration_parameter_array )
{
	const char fname[] = "mx_motor_get_acceleration_parameters()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;
	int i;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->get_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The get_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_ACCELERATION_PARAMETERS;

	status = ( *fptr ) ( motor );

	for ( i = 0; i < MX_MOTOR_NUM_ACCELERATION_PARAMS; i++ ) {
		acceleration_parameter_array[i] =
				motor->acceleration_parameters[i];
	}

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_acceleration_parameters( MX_RECORD *motor_record,
				double *acceleration_parameter_array )
{
	const char fname[] = "mx_motor_set_acceleration_parameters()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;
	int i;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->set_parameter;

	if ( fptr == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
"The set_parameter function is not supported by the driver for record '%s'.",
			motor_record->name );
	}

	motor->parameter_type = MXLV_MTR_ACCELERATION_PARAMETERS;

	for ( i = 0; i < MX_MOTOR_NUM_ACCELERATION_PARAMS; i++ ) {
		motor->acceleration_parameters[i] =
				acceleration_parameter_array[i];
	}

	status = ( *fptr ) ( motor );

	return status;
}

/* === Move by steps functions. (MXC_MTR_STEPPER) === */

MX_EXPORT mx_status_type
mx_motor_move_relative_steps(MX_RECORD *motor_record,
			long relative_steps,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags)
{
	long current_position, new_position;
	mx_status_type status;

	status = mx_motor_get_position_steps(motor_record, &current_position);

	if ( status.code != MXE_SUCCESS )
		return status;

	new_position = current_position + relative_steps;

	status = mx_motor_move_absolute_steps( motor_record,
			new_position, display_motor_move_fn, flags );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_move_absolute_steps(MX_RECORD *motor_record,
			long motor_steps,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags)
{
	const char fname[] = "mx_motor_move_absolute_steps()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	long positive_limit, negative_limit;
	long position, backlash;
	long relative_motion, backlash_position;
	double dummy_position;
	int mask, do_backlash;
	mx_status_type status, display_move_status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_STEPPER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not a stepper motor.",
			motor_record->name );
	}

	fptr = fl_ptr->move_absolute;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"move_absolute function pointer is NULL.");
	}

	positive_limit = motor->raw_positive_limit.stepper;
	negative_limit = motor->raw_negative_limit.stepper;

	status = mx_motor_get_position_steps(motor_record, &position);

	if ( status.code != MXE_SUCCESS )
		return status;

	relative_motion = motor_steps - position;

	/* Do not do the move if it would be less than the deadband size. */

	if ( labs( relative_motion ) < motor->deadband.stepper ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Do we need to do a backlash correction this time? */

	backlash = motor->raw_backlash_correction.stepper;

	if ( flags & MXF_MTR_IGNORE_BACKLASH ) {
		do_backlash = FALSE;
	} else {
		do_backlash = TRUE;
	}

	if ( do_backlash ) {
		/* Perform the backlash if relative_motion and backlash
		 * are both non-zero and have the same signs.
		 */

		do_backlash = FALSE;

		if ( ( relative_motion > 0 ) && ( backlash > 0 ) ) {
			do_backlash = TRUE;
		} else if ( ( relative_motion < 0 ) && ( backlash < 0 ) ) {
			do_backlash = TRUE;
		}
	}

	/* Do the backlash correction, if necessary. */

	if ( do_backlash ) {
		backlash_position = motor_steps + backlash;

		/* Check software limits. */

		if ( backlash_position > positive_limit ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Backlash correction for '%s' to %d steps would exceed positive limit "
	"at %d steps.",
				motor_record->name,
				backlash_position,
				positive_limit);
		}

		if ( backlash_position < negative_limit ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Backlash correction for '%s' to %d steps would exceed negative limit "
	"at %d steps.",
				motor_record->name,
				backlash_position,
				negative_limit);
		}

		/* Do not do the backlash correction if we were only asked
		 * to check limits.
		 */

		if ( (flags & MXF_MTR_ONLY_CHECK_LIMITS) == 0 ) {

			/* Update setpoint. */

			motor->setpoint = motor->offset
				+ motor->scale * (double) backlash_position;

			/* Do the move. */

			motor->raw_setpoint.stepper = backlash_position;

			status = ( *fptr ) ( motor );

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Leave now if we were not asked to wait for the
			 * completion of the backlash correction.
			 */

			mask = MXF_MTR_NOWAIT | MXF_MTR_GO_TO_BACKLASH_POSITION;

			if ( (flags & mask) == mask ) {
				return MX_SUCCESSFUL_RESULT;
			}

			/* Otherwise, display our progress on the way to
			 * the backlash position.
			 */

			if ( display_motor_move_fn != NULL ) {
				display_move_status = (*display_motor_move_fn)
						( flags, 1, &motor_record );
			} else {
				display_move_status = MX_SUCCESSFUL_RESULT;

				status = mx_wait_for_motor_stop(
						motor_record, flags );

				if ( status.code != MXE_SUCCESS )
					return status;
			}

			/* Update copy of motor position in MOTOR structure. */

			status = mx_motor_get_position(
					motor_record, &dummy_position);

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Return if move was aborted. */

			if ( display_move_status.code != MXE_SUCCESS )
				return display_move_status;
		}
	}

	/* Were we asked to only go to the backlash position? */

	if ( flags & MXF_MTR_GO_TO_BACKLASH_POSITION ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Check software limits. */

	if ( motor_steps > positive_limit ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"Error: move of '%s' to %d steps would exceed positive limit at %d steps.",
			motor_record->name,
			motor_steps,
			positive_limit );
	}

	if ( motor_steps < negative_limit ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"Error: move of '%s' to %d steps would exceed negative limit at %d steps.",
			motor_record->name,
			motor_steps,
			negative_limit );
	}

	/* Return now if we are only checking limits. */

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Update setpoint. */

	motor->setpoint = motor->offset
				+ motor->scale * (double) motor_steps;

	/* Do the move. */

	motor->raw_setpoint.stepper = motor_steps;

	status = ( *fptr ) ( motor );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( display_motor_move_fn != NULL ) {
		display_move_status = display_motor_move_fn(
						flags, 1, &motor_record);
	} else {
		display_move_status = MX_SUCCESSFUL_RESULT;
	}

	/* Update copy of motor position in MX_MOTOR structure, unless
	 * MXF_MTR_NOWAIT is set in 'flags'.
	 */

	if ( (flags & MXF_MTR_NOWAIT) == 0 ) {
		status = mx_motor_get_position(motor_record, &dummy_position);

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	return display_move_status;
}

MX_EXPORT mx_status_type
mx_motor_get_position_steps( MX_RECORD *motor_record, long *motor_steps )
{
	const char fname[] = "mx_motor_get_position_steps()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_STEPPER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not a stepper motor.",
			motor_record->name );
	}

	fptr = fl_ptr->get_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"get_position_steps function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	/* Update position. */

	motor->position = motor->offset
			+ (motor->scale) * motor->raw_position.stepper;

	*motor_steps = motor->raw_position.stepper;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_position_steps( MX_RECORD *motor_record, long motor_steps )
{
	const char fname[] = "mx_motor_set_position_steps()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_STEPPER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not a stepper motor.",
			motor_record->name );
	}

	fptr = fl_ptr->set_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"set_position_steps function pointer is NULL.");
	}

	motor->raw_setpoint.stepper = motor_steps;

	/* Update setpoint. */

	motor->setpoint = motor->offset
			+ motor->scale * (double) motor_steps;

	/* Invoke the function. */

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_motor_steps_to_go( MX_RECORD *motor_record, long *steps_to_go )
{
	const char fname[] = "mx_motor_steps_to_go()";

	MX_MOTOR *motor;
	long step_position, step_setpoint;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_STEPPER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not a stepper motor.",
			motor_record->name );
	}

	status = mx_motor_get_position_steps( motor_record, &step_position );

	if ( status.code != MXE_SUCCESS )
		return status;

	step_setpoint = motor->raw_setpoint.stepper;

	*steps_to_go = step_setpoint - step_position;

	return status;
}

/* === Move analog functions (MXC_MTR_ANALOG) === */

MX_EXPORT mx_status_type
mx_motor_move_absolute_analog(MX_RECORD *motor_record,
			double motor_position,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags)
{
	const char fname[] = "mx_motor_move_absolute_analog()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	double positive_limit, negative_limit;
	double present_position, backlash;
	double relative_motion, backlash_position;
	double test_var, dummy_position;
	int mask, do_backlash;
	char units[80];
	mx_status_type status, display_move_status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_ANALOG ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not an analog motor.",
			motor->record->name );
	}

	fptr = fl_ptr->move_absolute;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"move_absolute function pointer is NULL.");
	}

	positive_limit = motor->raw_positive_limit.analog;
	negative_limit = motor->raw_negative_limit.analog;

	status = mx_motor_get_position_analog(motor_record, &present_position);

	if ( status.code != MXE_SUCCESS )
		return status;

	relative_motion = motor_position - present_position;

	/* Do not do the move if it would be less than the deadband size. */

	if ( fabs( relative_motion ) < motor->deadband.analog ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Do we need to do a backlash correction this time? */

	backlash = motor->raw_backlash_correction.analog;

	if ( flags & MXF_MTR_IGNORE_BACKLASH ) {
		do_backlash = FALSE;
	} else {
		do_backlash = TRUE;
	}

	if ( do_backlash ) {
		/* Perform the backlash if relative_motion and backlash
		 * are both non-zero and have the same signs.
		 */

		do_backlash = FALSE;

		if ( ( relative_motion > 0.0 ) && ( backlash > 0.0 ) ) {
			do_backlash = TRUE;
		} else if ( ( relative_motion < 0.0 ) && ( backlash < 0.0 ) ) {
			do_backlash = TRUE;
		}
	}

	/* Do the backlash correction, if necessary. */

	/* MX currently doesn't record the raw units for analog motors.
	 * If the absolute value of the scale factor for this motor is
	 * very close to 1, I can assume that the raw units are the 
	 * same as the user units.  Otherwise, I don't know what the
	 * units are.  This is a kludge until I replace this with 
	 * something better.
	 */

	if ( do_backlash ) {
		backlash_position = motor_position + backlash;

		/* Check software limits. */

		if ( backlash_position > positive_limit ) {
			test_var = fabs( fabs( motor->scale ) - 1.0 );

			if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
				strcpy( units, " " );
				strncat( units, motor->units,
							sizeof(units) );
			} else {
				strcpy( units, "" );
			}

			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"Backlash correction for '%s' to %.*g%s would exceed positive limit at %.*g%s.",
				motor->record->name,
				motor->record->precision,
				backlash_position, units,
				motor->record->precision,
				positive_limit, units );
		}

		if ( backlash_position < negative_limit ) {
			test_var = fabs( fabs( motor->scale ) - 1.0 );

			if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
				strcpy( units, " " );
				strncat( units, motor->units,
							sizeof(units) );
			} else {
				strcpy( units, "" );
			}

			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"Backlash correction for '%s' to %.*g%s would exceed negative limit at %.*g%s.",
				motor->record->name,
				motor->record->precision,
				backlash_position, units,
				motor->record->precision,
				negative_limit, units );
		}

		/* Do not do the backlash correction if we were only asked
		 * to check limits.
		 */

		if ( (flags & MXF_MTR_ONLY_CHECK_LIMITS) == 0 ) {

			/* Update setpoint. */

			motor->setpoint = motor->offset
				+ motor->scale * backlash_position;

			/* Do the move. */

			motor->raw_setpoint.analog = backlash_position;

			status = ( *fptr ) ( motor );

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Leave now if we were not asked to wait for the
			 * completion of the backlash correction.
			 */

			mask = MXF_MTR_NOWAIT | MXF_MTR_GO_TO_BACKLASH_POSITION;

			if ( (flags & mask) == mask ) {
				return MX_SUCCESSFUL_RESULT;
			}

			/* Otherwise, display our progress on the way to
			 * the backlash position.
			 */

			if ( display_motor_move_fn != NULL ) {
				display_move_status = (*display_motor_move_fn)
						( flags, 1, &motor_record );
			} else {
				display_move_status = MX_SUCCESSFUL_RESULT;

				status = mx_wait_for_motor_stop(
						motor_record, flags );

				if ( status.code != MXE_SUCCESS )
					return status;
			}

			/* Update copy of motor position in MOTOR structure. */

			status = mx_motor_get_position(
					motor_record, &dummy_position);

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Return if move was aborted. */

			if ( display_move_status.code != MXE_SUCCESS )
				return display_move_status;
		}
	}

	/* Were we asked to only go to the backlash position? */

	if ( flags & MXF_MTR_GO_TO_BACKLASH_POSITION ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Check software limits. */

	if ( motor_position > positive_limit ) {
		test_var = fabs( fabs( motor->scale ) - 1.0 );

		if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
			strcpy( units, " " );
			strncat( units, motor->units, sizeof(units) );
		} else {
			strcpy( units, "" );
		}

		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Error: move of '%s' to %.*g%s would exceed positive limit at %.*g%s.",
			motor->record->name,
			motor->record->precision,
			motor_position, units,
			motor->record->precision,
			positive_limit, units );
	}

	if ( motor_position < negative_limit ) {
		test_var = fabs( fabs( motor->scale ) - 1.0 );

		if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
			strcpy( units, " " );
			strncat( units, motor->units, sizeof(units) );
		} else {
			strcpy( units, "" );
		}

		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
	"Error: move of '%s' to %.*g%s would exceed negative limit at %.*g%s.",
			motor->record->name,
			motor->record->precision,
			motor_position, units,
			motor->record->precision,
			negative_limit, units );
	}

	/* Return now if we are only checking limits. */

	if ( flags & MXF_MTR_ONLY_CHECK_LIMITS ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Update setpoint. */

	motor->setpoint = motor->offset
				+ motor->scale * motor_position;

	/* Do the move. */

	motor->raw_setpoint.analog = motor_position;

	status = ( *fptr ) ( motor );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( display_motor_move_fn != NULL ) {
		display_move_status = display_motor_move_fn(
						flags, 1, &motor_record);
	} else {
		display_move_status = MX_SUCCESSFUL_RESULT;
	}

	/* Update copy of motor position in MX_MOTOR structure, unless
	 * MXF_MTR_NOWAIT is set in 'flags'.
	 */

	if ( (flags & MXF_MTR_NOWAIT) == 0 ) {
		status = mx_motor_get_position(motor_record, &dummy_position);

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	return display_move_status;
}

MX_EXPORT mx_status_type
mx_motor_get_position_analog( MX_RECORD *motor_record,
				double *raw_position )
{
	const char fname[] = "mx_motor_get_position_analog()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_ANALOG ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not an analog motor.",
			motor_record->name );
	}

	fptr = fl_ptr->get_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"get_position_analog function pointer is NULL.");
	}

	status = ( *fptr ) ( motor );

	/* Update position. */

	motor->position = motor->offset
			+ (motor->scale) * motor->raw_position.analog;

	*raw_position = motor->raw_position.analog;

	return status;
}

MX_EXPORT mx_status_type
mx_motor_set_position_analog( MX_RECORD *motor_record,
						double raw_setpoint )
{
	const char fname[] = "mx_motor_set_position_analog()";

	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_MOTOR * );
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_ANALOG ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not an analog motor.",
			motor_record->name );
	}

	fptr = fl_ptr->set_position;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"set_position_analog function pointer is NULL.");
	}

	motor->raw_setpoint.analog = raw_setpoint;

	/* Update setpoint. */

	motor->setpoint = motor->offset + motor->scale * raw_setpoint;

	/* Invoke the function. */

	status = ( *fptr ) ( motor );

	return status;
}

MX_EXPORT mx_status_type
mx_is_motor_position_between_software_limits(
	MX_RECORD *motor_record, double proposed_position,
	int *result_flag, int generate_error_message )
{
	const char fname[] =
		"mx_is_motor_position_between_software_limits()";

	MX_MOTOR *motor;
	double scale, offset, unscaled_proposed_position;
	long proposed_step_position;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	scale = motor->scale;
	offset = motor->offset;

	if ( fabs(scale) < MX_MOTOR_ANALOG_FUZZ ) {
		unscaled_proposed_position = 0;
	} else {
		unscaled_proposed_position
			= (proposed_position - offset) / scale;
	}

	switch( motor->subclass ) {
	case MXC_MTR_ANALOG:
	    status =
		mx_is_analog_motor_position_between_software_limits(
			motor_record, unscaled_proposed_position,
			result_flag, generate_error_message );
	    break;
	case MXC_MTR_STEPPER:
	    proposed_step_position = mx_round( unscaled_proposed_position );

	    status =
		mx_is_stepper_motor_position_between_software_limits(
			motor_record, proposed_step_position,
			result_flag, generate_error_message );
	    break;
	default:
	    return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Unrecognized motor subclass = %ld", motor->subclass );
	    break;
	}
	return status;
}

MX_EXPORT mx_status_type
mx_is_analog_motor_position_between_software_limits(
	MX_RECORD *motor_record, double proposed_analog_position,
	int *result_flag, int generate_error_message )
{
	const char fname[] =
	    "mx_is_analog_motor_position_between_software_limits()";

	MX_MOTOR *motor;
	char message_buffer[120];
	char units[80];
	double positive_limit, negative_limit, test_var;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_ANALOG ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not an analog motor.",
			motor->record->name );
	}

	if ( result_flag == (int *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"result_flag pointer passed was NULL." );
	}

	message_buffer[0] = '\0';
	*result_flag = 0;

	positive_limit = motor->raw_positive_limit.analog;
	negative_limit = motor->raw_negative_limit.analog;

	/* MX currently doesn't record the raw units for analog motors.
	 * If the absolute value of the scale factor for this motor is
	 * very close to 1, I can assume that the raw units are the 
	 * same as the user units.  Otherwise, I don't know what the
	 * units are.  This is a kludge until I replace this with 
	 * something better.
	 */

	if ( proposed_analog_position > positive_limit ) {
		*result_flag = 1;

		if ( generate_error_message ) {
			test_var = fabs( fabs( motor->scale ) - 1.0 );

			if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
				strcpy( units, " " );
				strncat( units, motor->units,
						sizeof(units) );
			} else {
				strcpy( units, "" );
			}

			sprintf( message_buffer,
		"Move of '%s' to %.*g%s would exceed positive limit at %.*g%s.",
				motor_record->name,
				motor_record->precision,
				proposed_analog_position, units,
				motor_record->precision,
				positive_limit, units );
		}
	}
	if ( proposed_analog_position < negative_limit ) {
		*result_flag = -1;

		if ( generate_error_message ) {
			test_var = fabs( fabs( motor->scale ) - 1.0 );

			if ( test_var < MX_MOTOR_ANALOG_FUZZ ) {
				strcpy( units, " " );
				strncat( units, motor->units,
						sizeof(units) );
			} else {
				strcpy( units, "" );
			}

			sprintf( message_buffer,
		"Move of '%s' to %.*g%s would exceed negative limit at %.*g%s.",
				motor_record->name,
				motor_record->precision,
				proposed_analog_position, units,
				motor_record->precision,
				negative_limit, units );
		}
	}

	if ( (*result_flag != 0) && generate_error_message ) {
		return mx_error(MXE_WOULD_EXCEED_LIMIT, fname, message_buffer);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_is_stepper_motor_position_between_software_limits(
	MX_RECORD *motor_record, long proposed_step_position,
	int *result_flag, int generate_error_message )
{
	const char fname[] =
	    "mx_is_stepper_motor_position_between_software_limits()";

	MX_MOTOR *motor;
	char message_buffer[120];
	long positive_limit, negative_limit;
	mx_status_type status;

	status = mx_motor_get_pointers( motor_record,
					&motor, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( motor->subclass != MXC_MTR_STEPPER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"Motor '%s' is not an stepper motor.",
			motor->record->name );
	}

	if ( result_flag == (int *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"result_flag pointer passed was NULL." );
	}

	message_buffer[0] = '\0';
	*result_flag = 0;

	positive_limit = motor->raw_positive_limit.stepper;
	negative_limit = motor->raw_negative_limit.stepper;

	if ( proposed_step_position > positive_limit ) {
		*result_flag = 1;

		if ( generate_error_message ) {
			sprintf( message_buffer,
	"Move of '%s' to %ld steps would exceed positive limit at %ld steps.",
				motor_record->name,
				proposed_step_position,
				positive_limit );
		}
	}
	if ( proposed_step_position < negative_limit ) {
		*result_flag = -1;

		if ( generate_error_message ) {
			sprintf( message_buffer,
	"Move of '%s' to %ld steps would exceed negative limit at %ld steps.",
				motor_record->name,
				proposed_step_position,
				negative_limit );
		}
	}

	if ( (*result_flag != 0) && generate_error_message ) {
		return mx_error(MXE_WOULD_EXCEED_LIMIT, fname, message_buffer);
	}

	return MX_SUCCESSFUL_RESULT;
}

