/*
 * Name:    ms_prmtr.c
 *
 * Purpose: Functions used to process MX motor record events.
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_rec.h"
#include "mx_motor.h"
#include "mx_sock.h"

#include "ms_serv.h"
#include "ms_prmtr.h"

mx_status_type
mxsrv_setup_motor_process_functions( MX_RECORD *record_list_head )
{
	const char fname[] = "mxsrv_setup_motor_process_functions()";

	MX_RECORD *current_record;
	MX_RECORD_FIELD *record_field;
	MX_RECORD_FIELD *record_field_array;
	long i;

	MX_DEBUG( 2,("%s invoked.", fname));

	/* The following is done just in case we were called with
	 * a record that isn't really the list head.
	 */

	record_list_head = record_list_head->list_head;

	/* Go through the record list looking for motor records. */

	current_record = record_list_head->next_record;

	while ( current_record != record_list_head ) {
		if ( (current_record->mx_superclass == MXR_DEVICE)
		  && (current_record->mx_class == MXC_MOTOR) ) {

			record_field_array
				= current_record->record_field_array;

			for ( i = 0;
				i < current_record->num_record_fields;
				i++ )
			{
				record_field = &record_field_array[i];

				switch( record_field->label_value ) {
				case MXLV_MTR_BUSY:
				case MXLV_MTR_SETPOINT:
				case MXLV_MTR_POSITION:
				case MXLV_MTR_SOFT_ABORT:
				case MXLV_MTR_POSITIVE_LIMIT_HIT:
				case MXLV_MTR_NEGATIVE_LIMIT_HIT:
				case MXLV_MTR_CONSTANT_VELOCITY_MOVE:
				case MXLV_MTR_SPEED:
				case MXLV_MTR_BASE_SPEED:
				case MXLV_MTR_MAXIMUM_SPEED:
				case MXLV_MTR_ACCELERATION_PARAMETERS:
					record_field->process_function
					    = mxsrv_motor_process_function;
					break;
				default:
					break;
				}
			}
		}
		current_record = current_record->next_record;
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_motor_process_function( void *record_ptr,
			void *record_field_ptr, int operation )
{
	const char fname[] = "mxsrv_motor_process_function()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_MOTOR *motor;
	MX_MOTOR_FUNCTION_LIST *motor_function_list;
	char operation_string[30];
	long label_value;
	mx_status_type status;

	double position, speed;
	double acceleration_params[MX_MOTOR_NUM_ACCELERATION_PARAMS];
	int busy, limit_hit, flags;

	record = (MX_RECORD *) record_ptr;
	record_field = (MX_RECORD_FIELD *) record_field_ptr;

	switch ( operation ) {
	case MX_PROCESS_GET:
		strcpy( operation_string, "(MX_PROCESS_GET)" );
		break;
	case MX_PROCESS_PUT:
		strcpy( operation_string, "(MX_PROCESS_PUT)" );
		break;
	default:
		sprintf( operation_string, "(Operation %d)", operation );
		break;
	}

	label_value = record_field->label_value;

	MX_DEBUG(1,("%s: '%s.%s' %s (%ld)",
		fname, record->name, record_field->name,
		operation_string, label_value));

	/*
	 * A label value < 0 tells us that we do not need to do any
	 * further MX record processing. 
	 */

	if ( label_value < 0 ) {
		return MX_SUCCESSFUL_RESULT;
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	motor_function_list = (MX_MOTOR_FUNCTION_LIST *)
				(record->class_specific_function_list);

	status = MX_SUCCESSFUL_RESULT;

	switch( operation ) {
	case MX_PROCESS_GET:
		switch( label_value ) {
		case MXLV_MTR_POSITION:
			status = mx_motor_get_position( record, &position );
			break;
		case MXLV_MTR_BUSY:
			status = mx_motor_is_busy( record, &busy );
			break;
		case MXLV_MTR_NEGATIVE_LIMIT_HIT:
			status = mx_motor_negative_limit_hit(
						record, &limit_hit );
			break;
		case MXLV_MTR_POSITIVE_LIMIT_HIT:
			status = mx_motor_positive_limit_hit(
						record, &limit_hit );
			break;
		case MXLV_MTR_SPEED:
			status = mx_motor_get_speed( record, &speed );

			break;
		case MXLV_MTR_BASE_SPEED:
			status = mx_motor_get_base_speed( record, &speed );

			break;
		case MXLV_MTR_MAXIMUM_SPEED:
			status = mx_motor_get_base_speed( record, &speed );

			break;
		case MXLV_MTR_ACCELERATION_PARAMETERS:
			status = mx_motor_get_acceleration_parameters(
						record, acceleration_params );

			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_GET label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	case MX_PROCESS_PUT:
		switch( label_value ) {
		case MXLV_MTR_SETPOINT:
			flags = MXF_MTR_NOWAIT | MXF_MTR_IGNORE_KEYBOARD
					| MXF_MTR_IGNORE_BACKLASH;

			if ( motor->set_mode == FALSE ) {
				status = mx_motor_move_absolute( record,
					motor->setpoint, NULL, flags );
			} else {
				status = mx_motor_set_position( record,
						motor->setpoint );
			}
			break;
		case MXLV_MTR_SOFT_ABORT:
			MX_DEBUG(-2,("%s: SOFT ABORT for motor '%s'",
					fname, record->name));

			status = mx_motor_soft_abort( record );

			/* Set the soft abort flag back to zero. */

			motor->soft_abort = 0;
			break;
		case MXLV_MTR_CONSTANT_VELOCITY_MOVE:
			status = mx_motor_constant_velocity_move( record,
					motor->constant_velocity_move );
			break;
		case MXLV_MTR_SPEED:
			status = mx_motor_set_speed( record, motor->speed );

			break;
		case MXLV_MTR_BASE_SPEED:
			status = mx_motor_set_base_speed( record,
							motor->base_speed );

			break;
		case MXLV_MTR_MAXIMUM_SPEED:
			status = mx_motor_set_base_speed( record,
							motor->maximum_speed );

			break;
		case MXLV_MTR_ACCELERATION_PARAMETERS:
			status = mx_motor_set_acceleration_parameters( record,
					motor->acceleration_parameters );

			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_PUT label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown operation code = %ld", operation );
		break;
	}

	return status;
}

